## @file
# SLEPOS LDAP validation and checking module
# - Attribute Checking Functions Library
# @par Authors
# - Ondrej Holecek
# @copy 2011, SuSE Linux

## @class
# Attribute Checking Functions Library (ACFL) class.
# @par Provides
# - generic attributes checking for validity and consistence
# - per object class checking functions
# @par Requires
# - POS::Syslog
# - Net::LDAP::Entry
# - Exporter
# @par Exports
# ACFL_CHECKLIST hash struct where keys are SLEPOS objectClass and values are references to their checking functions
package POS::Admin::Validator::ACFL;

use strict;
use warnings;

use Net::LDAP::Entry;
use POS::Syslog;
use Exporter;

our $VERSION = 0.1;
our @ISA     = qw( Exporter );

our @EXPORT = qw (
  %ACFL_CHECKLIST
);

our %ACFL_CHECKLIST = (
    'scLocation'     => \&scLocation_check,
    'scHarddisk'     => \&scHarddisk_check,
    'scPartition'    => \&scPartition_check,
    'scCashRegister' => \&scCashRegister_check,
    'scRamDisk'      => \&scRamdisk_check,
    'scPosImage'     => \&scPosImage_check
);

## @cmethod POS::Admin::Validator::ACFL new($LDS, $BASE)
# Create new POS::LDAP::Cache object.
# @param LDS reference to POS::LDAP::Cache
# @param BASE string DN of checking root to limit checking scope
# @return A new object ref
sub new
{
    my $self = shift;
    my $type = ref($self) || $self;
    my $this = bless {}, $type;
    my %opt  = @_;

    $this->{'LDS'} = $opt{'LDS'} || die "Missing LDS in ACFL init";
    $this->{'BASE'} = $opt{'BASE'};
    return $this;
} ## end sub new

## @fn private array _getParams($LDS, $dn)
# get Net::LDAP::Entry object, its DN, children and LDS ref
# @param LDS reference to Net::LDAP::Entry or to POS::LDAP::Cache object
# @param dn string DN addressing LDS entry, needed only if LDS is ref to POS::LDAP::Cache
# @param ldap optional POS::LDAP reference. Used in scLocation check for POS::Admin
# @return Net::LDAP::Entry object, its children if LDS was provided else undef, dn and LDS if provided else undef
sub _getParams
{
    my ($LDS, $dn, $ldap) = @_;
    my ($entry, $children);
    if (ref $LDS eq 'Net::LDAP::Entry') {
        $entry    = $LDS;
        $children = [];
        $dn       = $entry->dn();
        undef $LDS;
    }
    else {
        if (ref $LDS eq 'POS::Admin::Validator::ACFL') {
            $LDS = $LDS->{'LDS'};
        }
        $entry    = $LDS->getLDAPentry($dn);
        $children = $LDS->getChildren($dn);
    }
    return ($entry, $dn, $children, $LDS, $ldap);
} ## end sub _getParams

## @fn bool checkEntry($entryDN)
# will check if entry is to be checked (is under $self->{'BASE'}) and then calls generic_check
# @param entryDN string DN of entry
# @return true on success else undef
sub checkEntry
{
    my ($self, $entryDN) = @_;
    my $base = $self->{'BASE'};
    if ($base && $entryDN =~ /$base/) {
        my $entry = $self->{'LDS'}->getLDAPentry($entryDN);
        #generic check
        return generic_check($entry);
    }
    return;
} ## end sub checkEntry

## @fn bool generic_check($LDS, $dn)
# generic attributes check.
# checks ip addresses, dns names
# @param LDS
# @param dn
# @note
# See _getParams for parameters description
# @return true on success else undef
sub generic_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    if (not defined $entry) {
        log_warn_printf('Entry object not defined for %s, skipping attributes check', $dn);
        return 1;
    }
    my $errors = 0;
    # get object class
    my ($objectClass) = grep { $POS::Admin::LdapStruct::LDAP_STRUCT{$_} } $entry->get_value('objectClass');
    # check must attributes
    foreach my $mustAttr (@{ $POS::Admin::LdapStruct::LDAP_STRUCT{$objectClass}{'attributes'}{'must'} }) {
        if (not defined $entry->get_value($mustAttr)) {
            log_error_printf('Entry is missing must attribute "%s"!', $mustAttr);
            $errors = 1;
        }
    }
    if ($errors) {
        log_error_print('Entry validation failed!');
        return;
    }
    # check ip address if present
    my $ipAddr = $entry->get_value('ipHostNumber');
    if (defined $ipAddr) {
        use Net::IPv4Addr(qw(ipv4_network));
        my ($ip, $cidr);
        eval { ($ip, $cidr) = ipv4_network($ipAddr); };
        if ($@) {
            log_error_printf('%s:ipHostNumber parsing failed. Not an IPv4 address');
            $errors += 1;
        }
        if ($ip eq $ipAddr && $ip ne '0.0.0.0') {
            log_error_printf('%s:ipHostNumber does not represent single ip address', $dn);
            $errors += 1;
        }
    }

    # check dns name
    #following cn/ou names will be checked for dns compatibilty ([a-zA-Z0-9] and '-')
    my %dnslike =
      ('scBranchServer' => 'cn', 'scServerContainer' => 'cn', 'scLocation' => 'cn', 'organizationalUnit' => 'ou');
    if ($dnslike{$objectClass}) {
        my $test = $entry->get_value($dnslike{$objectClass});
        my $err  = 0;
        $err += 1 unless $test =~ /^[a-zA-Z0-9\-]+$/;
        $err += 2 if $test =~ /^-/;
        $err += 4 if $test =~ /-$/;
        if ($err) {
            log_error_printf('%s:%s = "%s" does not conform to dns naming standard (a-z,A-Z,0-9,-, " - " only inside)',
                $dn, $dnslike{$objectClass}, $test);
            $errors += 1;
        }
    }
    return if $errors;
    return 1;
} ## end sub generic_check

## @fn bool scLocation_check($LDS, $dn)
# scLocation attributes check.
# checks ip addresses, dhcp ranges, network validity
# @param LDS
# @param dn
# @note
# See _getParams for parameters description
# @return true on success else undef
sub scLocation_check
{
    my ($entry, $dn, $children, $LDS, $ldap) = _getParams(@_);
    if (not defined $entry) {
        log_warn_printf('Entry object not defined for %s, skipping attributes check', $dn);
        return 1;
    }
    my $errors = 0;
    # error checking
    if (length $entry->get_value('userPassword') < 6) {
        # detect SM setup
        my $sm;
        # get root dn
        my ($rootDN) = $dn =~ /(o=[^,]+,c=.+)$/;
        my $standardsDN = 'cn=standards,' . $rootDN;
        if ($LDS and $LDS->getLDAPentry($standardsDN)) {
            $sm = $LDS->getLDAPentry($standardsDN)->get_value('scSUSEManager');
        }
        elsif ($ldap) {
            my $res = $ldap->search(
                base   => $standardsDN,
                scope  => 'base',
                filter => '(objectClass=scHardware)',
                attrs  => ['scSUSEManager']
            );
            if ($res->is_error()) {
                log_debug_printf('unable to find %s', $standardsDN);
            }
            else {
                $sm = $res->pop_entry()->get_value('scSUSEManager');
            }
        }
        if ($sm) {
            log_error_printf(
                'BranchServer %s password must be longer than 5 characters to be compatible with SUSE Manager', $dn);
            $errors += 1;
        }
        else {
            log_warn_printf(
                'BranchServer %s password should be longer than 5 characters to be compatible with SUSE Manager', $dn);
        }
    }
    # ip addresess range checking
    use Net::IPv4Addr(qw(ipv4_in_network ipv4_msk2cidr ipv4_cidr2msk ipv4_network ipv4_parse));
    # get all stored IPv4 adresses
    my $ipNetworkNumber = $entry->get_value('ipNetworkNumber');
    my $ipNetmaskNumber = $entry->get_value('ipNetmaskNumber');
    my $scDefaultGw     = $entry->get_value('scDefaultGw');
    (my $scDhcpRange      = $entry->get_value('scDhcpRange'))      =~ s/\s//g;
    (my $scDhcpFixedRange = $entry->get_value('scDhcpFixedRange')) =~ s/\s//g;
    # do basic IPv4 format check
    foreach ($ipNetworkNumber, $ipNetmaskNumber, $scDefaultGw) {
        eval { ipv4_parse($_); };
        if ($@) {
            log_error_printf('%s:invalid IPv4 address "%s"', $dn, $_);
            return;
        }
    }
    #check for 0.0.0.0 - LDAP will be configured based upon BS configuration
    if ($ipNetworkNumber eq '0.0.0.0' &&
        $ipNetmaskNumber  eq '0.0.0.0' &&
        $scDefaultGw      eq '0.0.0.0' &&
        $scDhcpFixedRange eq '0.0.0.0,0.0.0.1' &&
        $scDhcpRange      eq '0.0.0.2,0.0.0.3') {
        log_warn_printf('%s:BranchServer network data will be updated during BS initialization', $dn);
        return if ($errors);
        return 1;
    }

    #check if ipNetmaskNumber is mask
    if (ipv4_cidr2msk(ipv4_msk2cidr($ipNetmaskNumber)) ne $ipNetmaskNumber) {
        log_error_printf('%s:ipNetmaskNumber does not represent valid netmask', $dn);
        $errors += 1;
    }
    #check if gateway makes sense
    my ($ip, $cidr) = ipv4_network($scDefaultGw);
    if ($ip eq $scDefaultGw) {
        log_error_printf('%s:scDefaultGw does not represent single ip address', $dn);
        $errors += 1;
    }
    if (not ipv4_in_network($ipNetworkNumber, $ipNetmaskNumber, $scDefaultGw)) {
        log_error_printf('%s:scDefaultGw not in branch server network', $dn);
        $errors += 1;
    }
    #check DHCP settings
    if ($entry->get_value('scDhcpExtern') =~ /FALSE/i) {
        my @ranges           = ();
        my @scDhcpRange      = split(/,/, $scDhcpRange);
        my @scDhcpFixedRange = split(/,/, $scDhcpFixedRange);
        foreach (@scDhcpRange) {
            eval {
                if (not ipv4_in_network($ipNetworkNumber, $ipNetmaskNumber, $_)) {
                    log_error_printf('%s:scDhcpRange boundary "%s" is not in branch server network "%s/%s"',
                        $dn, $_, $ipNetworkNumber, $ipNetmaskNumber);
                    $errors += 1;
                }
            };
            if ($@) {
                log_error_printf('%s:scDhcpRange "%s" is not valid IPv4 address', $dn, $_);
                $errors += 1;
            }
            push @ranges, unpack('N', pack('C4', split(/\./, $_)));
        }
        foreach (@scDhcpFixedRange) {
            eval {
                if (not ipv4_in_network($ipNetworkNumber, $ipNetmaskNumber, $_)) {
                    log_error_printf('%s:scDhcpFixedRange boundary "%s" is not in branch server network "%s/%s"',
                        $dn, $_, $ipNetworkNumber, $ipNetmaskNumber);
                    $errors += 1;
                }
            };
            if ($@) {
                log_error_printf('%s:scDhcpFixedRange "%s" is not valid IPv4 address', $dn, $_);
                $errors += 1;
            }
            push @ranges, unpack('N', pack('C4', split(/\./, $_)));
        }
        if (scalar @ranges != 4) {
            log_error_printf('%s DHCP ranges are malformed', $dn);
            $errors += 1;
        }
        else {
            if ($ranges[1] < $ranges[0]) {
                log_error_printf('%s:scDhcpRange second boundary is smaller then first one.', $dn);
                $errors += 1;
            }
            if ($ranges[3] < $ranges[2]) {
                log_error_printf('%s:scDhcpFixedRange second boundary is smaller then first one.', $dn);
                $errors += 1;
            }
            if (($ranges[0] < $ranges[2] and $ranges[1] > $ranges[2]) or
                ($ranges[2] < $ranges[0] and $ranges[3] > $ranges[0])) {
                log_error_printf('%s DHCP ranges overlap', $dn);
                $errors += 1;
            }
        }
    }
    #do scService and scNetworkcard ip adresses check (if present)
    #at least one card must be defined to be in BS range
    my $foundNC = 0;
    while (@$children) {
        my $child = shift @$children;
        my $class = $LDS->getClass($child);
        if ($class =~ /scNetworkcard|scService/) {
            my $ipaddr = $LDS->getLDAPentry($child)->get_value('ipHostNumber');
            if (defined $ipaddr) {
                # if ipHostNumber is defined, check if it is in BS range and valid ipv4 address
                eval {
                    if (ipv4_in_network($ipNetworkNumber, $ipNetmaskNumber, $ipaddr)) {
                        $foundNC = 1;
                    }
                    elsif ($class eq 'scService') {
                        log_error_printf('%s:ipHostNumber of service is not in branch server network', $child);
                        $errors += 1;
                    }
                };
                if ($@) {
                    log_error_printf('%s:ipHostNumber is not valid IPv4 address', $child);
                    $errors += 1;
                }
            }
        }
        if ($class eq 'scServerContainer' or $class eq 'scBranchServer') {
            push @$children, @{ $LDS->getChildren($child) };
        }
    }
    if (@$children and not $foundNC) {
        log_error_printf('No configured NIC for %s found!', $dn);
        $errors += 1;
    }
    return if ($errors);
    return 1;
} ## end sub scLocation_check

## @fn bool scHarddisk_check($LDS, $dn)
# scHarddisk attributes check.
# checks scPartitionTable definition
# @param LDS
# @param dn
# @note
# See _getParams for parameters description
# @return true on success else undef
sub scHarddisk_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    if (not defined $entry) {
        log_warn_printf('Entry object not defined for %s, skipping attributes check', $dn);
        return 1;
    }
    my $scPartitionsTable = 'scPartitionsTable';
    my $partTable         = $entry->get_value($scPartitionsTable);
    my $err               = 0;
    my $hasroot           = 0;
    my $msg               = undef;
    my $np                = 0;

    #check $partTable|scPartition presence
    if ((not defined $LDS) && (not defined $partTable)) {
        log_debug_print('scPartitionsTable not present and in signle entry add mode, skipping check');
    }
    else {
        if (not defined $partTable) {
            #check for scPartition children
            if (scalar @$children == 0) {
                log_error_printf('Neither scPartitionsTable attribute nor scPartition children are defined in %s', $dn);
                return;
            }
            # check if we have root mount
            my $mps = $LDS->getFlag($dn, 'mountPoints');
            foreach my $mp (keys %$mps) {
                if ($mps->{$mp} > 1) {
                    log_error_print('Conflicting root mounts for scHarddisk (%s)', $dn);
                    return;
                }
                elsif ($mp eq '/') {
                    $hasroot = 1;
                }
            }
            if (not $hasroot) {
                log_error_printf('No root mount defined for scHarddisk (%s)', $dn);
                return;
            }
        }
        else {
            my @parts = split(';', $partTable);
            #need format like: "1000 82 x;8000 83 /;2500 S /tmp ext3"
            foreach my $part (@parts) {
                $np++;
                my @pparts = split(' ', $part);
                if (@pparts < 3) {
                    $err = 1;
                    $msg =
                      "$scPartitionsTable format error, not enough data in (\'$part\') specified for partition no. $np";
                    last;
                }
                if (@pparts > 4) {
                    $err = 1;
                    $msg =
                      "$scPartitionsTable format error, too much data in (\'$part\') specified for partition no. $np";
                    last;
                }
                my $psize  = $pparts[0];
                my $ptype  = $pparts[1];
                my $pmount = $pparts[2];
                if ($psize !~ /^\d+$/ && $psize ne 'x') {    #x for rest of free space, see also #550433
                    $err = 2;
                    $msg =
"$scPartitionsTable format error, size must be an integer or 'x' in (\'$part\') in partition no. $np";
                    last;
                }
                if ($ptype !~ /^\d{1,2}$/ && $ptype !~ /^[A-Z]$/ && $ptype ne 'swap')
                {    #letters used L,S, (E,X - nlpos9?), swap seems only in the nlpos9 manual
                    $err = 3;
                    $msg =
"$scPartitionsTable format error, invalid type in (\'$part\') in partition no. $np.\n  Use number, 'L' (=83), 'S' (=82) or 'swap' (=82).";
                    last;
                }
                $hasroot++ if ($pmount eq '/');
            }
            unless ($err) {
                if ($hasroot != 1) {
                    $err = 4;
                    $msg = sprintf(
                        "$scPartitionsTable format error, %s root%s specified in \'%s\'.",
                        $hasroot ? $hasroot : 'no',
                        $hasroot ? 's'      : '', $partTable
                    );
                }
            }
        }
        if (defined $LDS) {
            # mark on parent (scCashRegister) we exists
            my $parent = $LDS->getParentDN($dn);
            $LDS->setFlag($parent, 'disk_ok');
        }
    }
    if ($err) {
        log_error_print($msg);
        return;
    }
    return 1;
} ## end sub scHarddisk_check

sub scPartition_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    my $partNum   = $entry->get_value('scPartNum');
    my $partType  = $entry->get_value('scPartType');
    my $partSize  = $entry->get_value('scPartSize');
    my $partMount = $entry->get_value('scPartMount');

    # check partition type
    if ($partType !~ /^\d{1,2}$/ && $partType !~ /^[A-Z]$/ && $partType ne 'swap') {
        #letters used L,S, (E,X - nlpos9?), swap seems only in the nlpos9 manual
        log_error_printf(
            'scPartition format error, invalid type in (%s)\nUse number, \'L\' (=83), \'S\' (=82) or \'swap\' (=82).',
            $dn);
        return;
    }
    # check partition size
    if ($partSize !~ /^\d+$/ && $partSize ne 'x') {    #x for rest of free space, see also #550433
        log_error_printf('scPartition format error, size must be an integer or \'x\' in (%s)', $dn);
        return;
    }
    # check mount points
    if (defined $LDS) {
        my $parent = $LDS->getParentDN($dn);
        my $mp = $LDS->getFlag($parent, 'mountPoints');
        if (not defined $mp) {
            $mp = { $partMount => 1 };
        }
        else {
            if (not defined $mp->{$partMount}) {
                $mp->{$partMount} = 1;
            }
            else {
                $mp->{$partMount} = $mp->{$partMount} + 1;
            }
        }
        $LDS->setFlag($parent, 'mountPoints', $mp);
    }
    return 1;
} ## end sub scPartition_check

sub scRamdisk_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    # do nothing in single-entry-add mode
    return 1 unless (defined $LDS);
    # else mark on parent (scCashRegister) we exists
    my $parent = $LDS->getParentDN($dn);
    $LDS->setFlag($parent, 'disk_ok');
    return 1;
} ## end sub scRamdisk_check

sub scCashRegister_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    # do nothing in single-entry-add mode
    return 1 unless (defined $LDS);
    # else check we have at least one ramdisk or harddisk
    if (not $LDS->getFlag($dn, 'disk_ok')) {
        log_error_printf('scCashRegister (%s) does not have any disk device configured', $dn);
        return;
    }
    return 1;
} ## end sub scCashRegister_check

sub scPosImage_check
{
    my ($entry, $dn, $children, $LDS) = _getParams(@_);
    my $imageFile = $entry->get_value('scImageFile');
    if ((not defined $imageFile) || (not $imageFile)) {
        log_error_printf('scPosImage (%s) has empty scImageFile attribute', $dn);
        return;
    }
    return 1;
} ## end sub scPosImage_check
