## @file
# SLEPOS LDAP and other data cache module
# - provides data to registration procedures
# - minimizes LDAP access
# - Attribute Checking Functions Library

package POS::posDataCache;

use strict;
use warnings;

use Exporter;
use POS::Config;
use POS::Syslog;
use POS::LDAP;
use POS::posDns;
use POS::LdapUtil;
use POS::LDAP::Cache;
use POS::MAC;
use File::Basename;
use Digest::MD5 qw(md5_hex);


our $VERSION = 0.1;

sub new
{
    my $package = shift;    ## packagename

    # Definition of object data, should be only general data common to all workstation instances
    my $data = {

        'STATIC'             => undef,    ## POS::Config, data from branchserver configuraton file  #todo
        'SCRIPTLIST'         => [],       ## actual array of scripts  #todo
        'SCRIPTLIST_EXTDHCP' => [],       ## actual array of scripts for ext dhcp  #todo

#        'OBJC'                   => undef,                               ## subtree to be filled with selected objects and lists of objects of ldap_entry type

        'LDAP'      => undef,             ## Ldap Connection
                                          #        'LDAPBASE' => undef,        ## LDAP-global-Base - o=...,c=..
                                          #        'BRANCHBASE' => undef,      ## LDAP-Base - dn of used scLocation
        'LDAPCACHE' => undef,             ## POS::LDAP::Cache object reference, to load ldap cache
        'LDAPTREE'  => undef,

## link to LDAPCACHE->LDS hash, which is keyed by dn names, values are arrays containing attributes [0] and links [1] on dependent objects
        'BRANCH' => undef,                ## hash of scLocation relevant data
                                          ## identified by the BRANCHBASE
                                          ## refering to an array with the following
                                          ## information:
                                          ## 1: LDAP-Container of the BRANCH
                                          ## 2: ipNetmaskNumber
                                          ## 3: scDhcpRange
                                          ## 4: scDhcpFixedRange
                                          ## 'ALLOWROLEBASED' is rolebased
                                          ## 'ALLOWGLOBALROLES' are global roles allowed

    };
    $data->{'WS'}{'ACT'}   = undef;       #list of active workstations (hashed by cn name)
    $data->{'WS'}{'DEACT'} = undef;       #list of deactivated workstations (hashed by cn name)
    $data->{'WS'}{'ALL'}   = undef;       #list of all workstations (hashed by cn name)
    $data->{'WS'}{'BY-MAC'} =
      undef;    #list of all (active only, effectively, since empty MAC == deactivated) workstations (hashed by MAC)
    $data->{'IDHASH'}{'LDAP_LIST'} = undef;    #hash of all ids in ldap
    $data->{'IDHASH'}{'LDAP-WS'}   = undef;    #hash of ids in ldap not used by active workstations
    $data->{'IDHASH'}{'WS'}        = undef;    #hash of ids in ldap used by active workstations
    $data->{'IDHASH'}{'CLAIMED'}   = undef;    #hash of ids being claimed in the current cycle
    $data->{'IDHASH'}{'PATHNAME'}  = undef;    #general pathname for id list

    $data->{'ROLEHASH'}{'LDAP_LIST'} = undef;  #hash of all, local, global roles in ldap

    $data->{'ROLEHASH'}{'USED'}    = undef;  #all (used) roles in ldap (should be set consitently with ALLOWGLOBALROLES)
                                             #data derived from USED hash:
    $data->{'ROLEHASH'}{'MACS'}    = undef;  # roles referenced by specific ws mac
    $data->{'ROLEHASH'}{'CRDEF'}   = undef;  # roles containging defualt cashregister
    $data->{'ROLEHASH'}{'HWTYPES'} = undef;  # roles allowed for a given hwtype
    $data->{'IDHASH'}{'PATHNAME'}  = undef;  #general pathname for role list


    $data->{'ERROR'}{'CODE'}    = 0;
    $data->{'ERROR'}{'MESSAGE'} = '';
    set_var($data, @_);
    #delete( $data->{'CONFIG'} ) #  ;    #delete config REFERENCE (needed only to initialize)
    bless($data);

    return unless ($data->update());
    #update is require for latter:
    return unless $data->_init_ldap_static_data();    #read only once on new, do not reload on update
    return $data;
} ## end sub new

## note: return $self->get_error() for externally accessible methods,
##       return undef and parse error directly for internally used methods


sub update
{
    my ($self) = @_;
    #role and id update done manually via load_role_data, load_id_data, process_id_data, process_role_data
    $self->{'ERROR'}{'CODE'}    = 0;
    $self->{'ERROR'}{'MESSAGE'} = '';
    $self->_get_ldap_conn();
    $self->_init_ldap_data() unless $self->{'ERROR'}{'CODE'};
    if ($self->{'ERROR'}{'CODE'}) {
        log_alert_print(
            "Cache Module init error $self->{'ERROR'}{'CODE'} : $self->{'ERROR'}{'MESSAGE'}, cannot continue");
        return undef;
    }
    return 1;
} ## end sub update


#adds/(re)sets a variable of this object
sub set_var
{
    (my $self, my %data) = @_;
    %$self = (%$self, %data);
} ## end sub set_var

# Preconditions: the 'CONFIG' must already have been
#                loaded
# Postconditions: 'LDAP' contains connection
# Returns: $self->{'LDAP'} (undefined or valid and checked)
## @method $ get_ldap_conn()
# Creates LDAP connection object if not yet exists.
# updates LDAP cache object
# @par Details:
# If $self->{'LDAP'} is not defined, a new POS::LDAP obect is created and
# initialized. I all cases, the LDAP connection is checked
# @return $self->{'LDAP'} object reference or undef if LDAP not accessible or cannot be created.
sub _get_ldap_conn
{
    my $self = shift;
    ## if we don't have a working LDAP connection get one
    if (!$self->{'LDAP'} || !$self->{'LDAP'}->check()) {
        $self->{'LDAP'} = POS::LDAP->new(
            'CONFIG' => $self->{'STATIC'}{'CONFIG'}
              #    'NEEDCAP'    => 'rw',
              #    'LDAPBINDDN' => $self->{'STATIC'}{'LDAP_ACCOUNT'},
              #    'LDAPBINDPW' => $self->{'STATIC'}{'LDAP_PASSWORD'}
        );
    }

    ## check if LDAP connection works
    ## this is meant to catch a connection that dies
    if (!$self->{'LDAP'}) {
        log_alert_print('LDAP connection broken');
        ## destroy POS::LDAP object, will be reinstantiated next time through here
        $self->{'ERROR'}{'CODE'}    = -3;
        $self->{'ERROR'}{'MESSAGE'} = 'LDAP connection broken';
        $self->{'LDAP'}             = undef;
        return undef;
    }

    return $self->{'LDAP'};
} ## end sub _get_ldap_conn


# orignially in ldap workstation
sub _init_ldap_data
{
    (my $self) = @_;

    my $ldap = $self->{'LDAP'};
    if (!defined $ldap) {
        $self->{'ERROR'}{'CODE'}    = -1;
        $self->{'ERROR'}{'MESSAGE'} = "LDAP object not defined in posDataCache module, cannot access LDAP";
        return undef;    #cannot return $self->get_error(), not yet blessed
    }

#fill the LDAP hashes, reads whole tree (global+branch, NOT all branches/ous)
# no filtering, what we get depends on what we are (AS/BS) which depends on how we open POS::Config (POS::posDataStatic)
    my $search = $ldap->search(
        base   => $self->{'STATIC'}{'LDAPBASE'},
        filter => '(objectClass=*)',
        attrs  => ['*', 'createTimestamp', 'modifyTimestamp'],
        scope  => 'sub'
    );
    if (!defined $search) {
        log_error_print("Cannot access LDAP tree on \'%s\' base, error", $self->{'STATIC'}{'LDAPBASE'});
        return undef;
    }
    $self->{'LDAPCACHE'} = POS::LDAP::Cache->new(
        'CONFIG'    => $self->{'STATIC'}{'CONFIG'},
        'BASE'      => $self->{'STATIC'}{'LDAPBASE'},
        'LDAP'      => $self->{'LDAP'},
        'LDAP_DATA' => $search
    );
    if (!defined $self->{'LDAPCACHE'}) {
        $self->{'ERROR'}{'CODE'}    = -2;
        $self->{'ERROR'}{'MESSAGE'} = "Cannot create LDAP CACHE, error";
        return undef;
    }

    # shortcut, LDS should be the only one needed, LDAPCACHE should not be used unless some of its methods are needed
    $self->{'LDAPTREE'} = $self->{'LDAPCACHE'}{'LDS'};

    #global part of DEFAULT_PXE, is defined even if not in branchbase mode
    my $pxe_hr = $self->_get_default_cont();
    if (!defined $pxe_hr) {
        $self->{'ERROR'}{'CODE'}    = -3;
        $self->{'ERROR'}{'MESSAGE'} = "Problem in LDAP cache while searching for default pxe configuration...";
        return undef;
    }
    $self->{'DEFAULT_PXE'}{'LDAP'} = $pxe_hr;

    if ($self->{'STATIC'}{'BRANCHBASE'}) {
        # get branch-specific data
        $self->{'BRANCH'} = $self->get_branch_data();
        if (!defined $self->{'BRANCH'}) {
            $self->{'ERROR'}{'CODE'}    = -5;
            $self->{'ERROR'}{'MESSAGE'} = "BRANCH data init error";
            return undef;
        }
        ($self->{'BRANCH'}{'SERVERCONTAINER'}, $self->{'BRANCH'}{'BRANCHSERVER'}, $self->{'SERVICES'}) =
          $self->get_branch_services();
        if (!defined $self->{'SERVICES'}) {
            $self->{'ERROR'}{'CODE'}    = -6;
            $self->{'ERROR'}{'MESSAGE'} = "branchserver SERVICES data init error";
            return undef;
        }

        #set dhcp specific data
        if (!$self->{'STATIC'}{'EXTDHCP'}) {
            if (!defined $self->{'SERVICES'}{'dhcp'}) {
                $self->{'ERROR'}{'CODE'}    = -7;
                $self->{'ERROR'}{'MESSAGE'} = "No DHCP-Server defined under $self->{'STATIC'}{'BRANCHBASE'}";
                return undef;
            }
            my $dynleasetime = $self->{'SERVICES'}{'dhcp'}{'scDhcpDynLeaseTime'};
            my $fixleasetime = $self->{'SERVICES'}{'dhcp'}{'scDhcpFixedLeaseTime'};
            if (!$dynleasetime || !($dynleasetime =~ /^[0-9]+/g)) {
                $dynleasetime = 300;
                log_warn_printf(
"No scDhcpDynLeaseTime defined under $self->{'SERVICES'}{'dhcp'}{'dn'}, will use hardcoded default ($dynleasetime)."
                );
            }
            if (!$fixleasetime || !($fixleasetime =~ /^[0-9]+/g)) {
                $fixleasetime = 14400;
                log_warn_printf(
"No scDhcpFixedLeaseTime defined under $self->{'SERVICES'}{'dhcp'}{'dn'}, will use hardcoded default ($fixleasetime)."
                );
            }
            $self->{'DHCP_DYNLEASETIME'} = $dynleasetime;    #LDAP2DHCP_DYNLEASETIME=300
            $self->{'DHCP_FIXLEASETIME'} = $fixleasetime;    #LDAP2DHCP_LOCALBOOTLEASETIME=14400
        }

        #get_scNetworkCard(s)
        my $ntw_ar = $self->{'LDAPCACHE'}->getChildrenByClass($self->{'BRANCH'}{'BRANCHSERVER'}, 'scNetworkcard');
        foreach my $netcard_dn (@$ntw_ar) {
            #push @{$self->{'NETWORKCARDS'}}, $self->{'LDAPCACHE'}->getLDAPentry($netwcard);
            my $netcard_entry = $self->{'LDAPCACHE'}->getLDAPentry($netcard_dn);
            $self->{'NETWORKCARDS'}{ $netcard_entry->get_value('scDevice') } = $netcard_entry;
        }

        #get default pxes and make kernel parameters line
        my $defpxe_hr = $self->{'DEFAULT_PXE'};

        $defpxe_hr->{'PATH'}     = trimpath($self->{'STATIC'}{'TFTP_PXE_DIR'});
        $defpxe_hr->{'BOOTPATH'} = trimpath($self->{'STATIC'}{'TFTP_BOOT_DIR'});
        $defpxe_hr->{'NAME'}     = $self->{'STATIC'}{'CONFIG'}{'TFTP_PXECFG_FILE'};
        $defpxe_hr->{'TEMPLATE'} = $self->{'STATIC'}{'CONFIG'}{'TFTP_BOOT_PXELINUXCFG_CONFIG_TEMPLATE'};
        $self->_get_default_pxes($defpxe_hr->{'LDAP'});
        $self->process_pxe($defpxe_hr->{'LDAP'});

        #load workstations
        (   $self->{'WS'}{'ACT'},    $self->{'WS'}{'DEACT'}, $self->{'WS'}{'ALL'},
            $self->{'WS'}{'BY-MAC'}, $self->{'WS'}{'BY-IP-ALL'}
        ) = POS::LdapUtil::get_ws_from_ldap('CACHE' => $self, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});

        #init cashreg list - list of all cashregs not in roles
        $self->{'CR-R'} = $self->load_crs_from_ldap();

        #TODO: some list of common dn?

    }    # end of branch specific data
} ## end sub _init_ldap_data


# used for data need only on object initalization, not to be updated during reload
sub _init_ldap_static_data
{
    my ($self) = @_;


    if ($self->{'STATIC'}{'BRANCHBASE'}) {

        if (!defined $self->{'SERVICES'}{'tftp'}) {
            log_error_print("No TFTP-Server defined under $self->{'STATIC'}{'BRANCHBASE'}");
            return undef;
        }
        my $tftp = $self->{'SERVICES'}{'tftp'}{'ipHostNumber'};

        if (!defined $tftp) {
            #get ip address from scNetworkCard
            my $network = $self->{'BRANCH'}{'NETNUM'} . '/' . $self->{'BRANCH'}{'NETMASK'};
            my @ncards  = values %{ $self->{'NETWORKCARDS'} };
            my $entry   = POS::posDns::get_scNetworkCard(\@ncards, $self->{'STATIC'}{'BRANCHBASE'}, $network);
            $tftp = $entry->get_value('ipHostNumber');
            log_warn_print(
"No ipHostNumber of the TFTP-Server under $self->{'STATIC'}{'BRANCHBASE'} defined, using ip $tftp from scNetworkcard"
            );
        }
        $self->{'TFTP'} = $tftp;


        #init dhcp data
        $self->{'DHCP'}{'LEASES'}{'PATHNAME'} = $self->{'STATIC'}{'CONFIG'}{'LEASES2LDAP_LEASES_FILE'};
        $self->{'DHCP'}{'LEASES'}{'FILETIME'} = 0;
        $self->{'DHCP'}{'CONFIG'}{'PATHNAME'} = $self->{'STATIC'}{'CONFIG'}{'LDAP2DHCP_DHCPCONFFILE'};
        $self->{'DHCP'}{'CONFIG'}{'FILETIME'} = 0;


    }    #end of brach specific data init (not for adminserver)

    return 1;

} ## end sub _init_ldap_static_data


####################################################
# methods


# load roles, parse data and prepare for use in posLeaseMonitor
# and create ROLEHASH hash
sub load_role_data
{
    #init roles
    #gather pathnames
    my ($self) = @_;
    my $rolepathname = $self->{'STATIC'}{'ROLE_PATHNAME'};
    unless ($rolepathname) {    #probably not so neccessary, should be ok
        log_alert_print('rolelist is not defined, error in branchserver.conf/adminserver.conf');
        return undef;
    }
    $self->{'ROLEHASH'}{'PATHNAME'} =
      $rolepathname;            # general base pathname, MUST be always present (or backup will do weird things...)
                                #$self->{'ROLEHASH'}{'ALL'}{'FILETIME'} = 0;
    if ($self->{'BRANCH'}{'ALLOWROLEBASED'}) {    # rolebased allowed, prepare role and id lists
        $self->{'ROLEHASH'}{'LDAP_LIST'}{'ALL'} = POS::LdapUtil::get_roles_from_ldap(
            'CACHE'    => $self,
            'UPTOBASE' => $self->{'STATIC'}{'LDAPBASE'},
            'BASE'     => $self->{'STATIC'}{'BRANCHBASE'},
            'SCLOCDN'  => $self->{'STATIC'}{'BRANCHBASE'}
        );
        #split to local, global
        foreach my $rolekey (keys %{ $self->{'ROLEHASH'}{'LDAP_LIST'}{'ALL'} }) {
            my $rolentry = $self->{'ROLEHASH'}{'LDAP_LIST'}{'ALL'}{$rolekey};
            $self->{'ROLEHASH'}{'LDAP_LIST'}{'LOCAL'}{$rolekey}  = $rolentry if !$rolentry->{'GLOBAL'};
            $self->{'ROLEHASH'}{'LDAP_LIST'}{'GLOBAL'}{$rolekey} = $rolentry if $rolentry->{'GLOBAL'};
        }
    }
} ## end sub load_role_data


# load ids
# and create IDHASH hash
sub load_id_data
{
    #init roles
    #gather pathnames
    my ($self) = @_;
    my $idpathname = $self->{'STATIC'}{'ID_PATHNAME'};
    #    if (defined($self->{'STATIC'}{'CONFIG'}{'POS_IDLIST_NAME'})) {
    #        $idpathname = $self->{'STATIC'}{'ID_PATHNAME'};
    #    }
    unless (defined $idpathname) {    #probably not so neccessary, should be ok
        log_alert_print('idlist pathname is not defined, error in default.sh');
        return undef;
    }
    $self->{'IDHASH'}{'PATHNAME'} = $idpathname;
    #$self->{'IDHASH'}{'FILETIME'}   = 0;
    #substract all scWorkstation ($self->{'CACHE'}{'WS'})from ID list
    $self->{'IDHASH'}{'CLAIMED'} = {};    #remove id list of previously registered workstations
    $self->{'IDHASH'}{'LDAP_LIST'} =
      POS::LdapUtil::get_ids_from_ldap('CACHE' => $self, 'BASE' => $self->{'STATIC'}{'BRANCHBASE'});

} ## end sub load_id_data

sub process_id_data
{
    my ($self) = @_;
    my ($idshashldappurged_ref, $idshashws_ref) =
      $self->_purge_id_pool($self->{'IDHASH'}{'LDAP_LIST'}, $self->{'WS'}{'ACT'})
      if defined $self->{'IDHASH'}{'LDAP_LIST'};
    $self->{'IDHASH'}{'LDAP-WS'} = $idshashldappurged_ref;    #ids in LDAP and NOT used by any active WS.
    $self->{'IDHASH'}{'WS'}      = $idshashws_ref;            #ids used by an active WS.

} ## end sub process_id_data

sub process_role_data
{
    my ($self, $useglobal) = @_;

    if ($useglobal) {
        $self->{'ROLEHASH'}{'USED'}{'LDAP'} = $self->{'ROLEHASH'}{'LDAP_LIST'}{'ALL'};
    }
    else {
        $self->{'ROLEHASH'}{'USED'}{'LDAP'} = $self->{'ROLEHASH'}{'LDAP_LIST'}{'LOCAL'};
    }
    $self->{'ROLEHASH'}{'USED'}{'PATHNAME'} = $self->{'ROLEHASH'}{'PATHNAME'};    #all roles rolelist pathname
         #make sure indices are tightly ordered regardless how USED was selected
    $self->_pack_indices($self->{'ROLEHASH'}{'USED'}{'LDAP'});
    ## find default role and first role of all used
    $self->_find_default_roles($self->{'ROLEHASH'}{'USED'}, $self->{'ROLEHASH'}{'USED'}{'LDAP'});

    ## make specific lists of roles containing default cashregister
    $self->{'ROLEHASH'}{'CRDEF'}{'LDAP'}     = $self->_filter_default_cr_roles($self->{'ROLEHASH'}{'USED'}{'LDAP'});
    $self->{'ROLEHASH'}{'CRDEF'}{'PATHNAME'} = $self->{'ROLEHASH'}{'USED'}{'PATHNAME'} . '.default';
    $self->_find_default_roles($self->{'ROLEHASH'}{'CRDEF'}, $self->{'ROLEHASH'}{'CRDEF'}{'LDAP'});

    ## make specific lists by MAC
    foreach my $mac (keys %{ $self->{'WS'}{'BY-MAC'} }) {
        my $macldaphash = $self->_filter_roles_mac($self->{'ROLEHASH'}{'USED'}{'LDAP'}, $self->{'WS'}{'BY-MAC'}{$mac});
        if (keys %{$macldaphash}) {    #non empty hash
            $self->_pack_indices($macldaphash);    #compress indices
            $self->{'ROLEHASH'}{'MACS'}{$mac}{'LDAP'} = $macldaphash;
            my $outmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($mac);
            $self->{'ROLEHASH'}{'MACS'}{$mac}{'PATHNAME'} = $self->{'ROLEHASH'}{'USED'}{'PATHNAME'} . '.' . $outmac;
            $self->_find_default_roles($self->{'ROLEHASH'}{'MACS'}{$mac}, $self->{'ROLEHASH'}{'MACS'}{$mac}{'LDAP'});
        }
        #filetime and file later
    }

    ## make specific lists by HWTYPE
    my $hwldaphash = $self->_filter_roles_hwtype($self->{'ROLEHASH'}{'USED'}{'LDAP'});    #keys=hwtypes,vals=roles
    foreach my $hwtype (keys %{$hwldaphash}) {    #  hwtype already transliterated for ugly chars
                                                  #add roles containing default cashregisters
        $hwldaphash->{$hwtype} =
          $self->_join_role_lists(($self->{'ROLEHASH'}{'CRDEF'}{'LDAP'}, $hwldaphash->{$hwtype}));
        $self->_pack_indices($hwldaphash->{$hwtype});
        $self->{'ROLEHASH'}{'HWTYPES'}{$hwtype}{'LDAP'}     = $hwldaphash->{$hwtype};
        $self->{'ROLEHASH'}{'HWTYPES'}{$hwtype}{'PATHNAME'} = $self->{'ROLEHASH'}{'USED'}{'PATHNAME'} . '.' . $hwtype;
        $self->_find_default_roles($self->{'ROLEHASH'}{'HWTYPES'}{$hwtype},
            $self->{'ROLEHASH'}{'HWTYPES'}{$hwtype}{'LDAP'});
    }
    #pack indices, can be safely done now
    $self->_pack_indices($self->{'ROLEHASH'}{'CRDEF'}{'LDAP'});


} ## end sub process_role_data


sub get_branch_data
{
    my ($self) = @_;
    my %branch_h;

    # test if BRANCHBASE is defined
    if (!$self->{'STATIC'}{'BRANCHBASE'}) {
        log_alert_print("No BRANCHBASE defined in static configuration, cannot continue!");
        return undef;
    }

    # they should all exist already..maybe except scNetworkcard
    my $scLocation = $self->{'LDAPCACHE'}->getLDAPentry($self->{'STATIC'}{'BRANCHBASE'});
    if (!defined $scLocation) {
        log_error_print("scLocation " . $self->{'STATIC'}{'BRANCHBASE'} . " not found");
        return undef;
    }

    #TODO: make a general parsing function for this (ldap entry obj independence)
    #transfer attributes
    my %att_list = ();
    $att_list{'dn'} = $scLocation->dn();    #should be same as BRANCHBASE
    for my $att ($scLocation->attributes) {
        $att_list{$att} = $scLocation->get_value($att);
    }
    $branch_h{'scLocation'} = \%att_list;


    $branch_h{'ALLOWROLEBASED'}   = POS::LdapUtil::boolean_to_integer($branch_h{'scLocation'}{'scAllowRoles'});
    $branch_h{'ALLOWGLOBALROLES'} = POS::LdapUtil::boolean_to_integer($branch_h{'scLocation'}{'scAllowGlobalRoles'});

    $branch_h{'ALLOWIDS'} = 0;
    $branch_h{'ALLOWIDS'} = 1 if ($branch_h{'scLocation'}{'scIdPool'});

    $branch_h{'DNSMAP'} = 0;
    $branch_h{'DNSMAP'} = 1 if ($branch_h{'scLocation'}{'scDnsMapFunc'} && $branch_h{'ALLOWROLEBASED'});

    $branch_h{'IPMAP'} = 0;
    $branch_h{'IPMAP'} = 1 if ($branch_h{'scLocation'}{'scIpMapFunc'} && $branch_h{'ALLOWROLEBASED'});

    $branch_h{'NETNUM'}  = $branch_h{'scLocation'}{'ipNetworkNumber'};
    $branch_h{'NETMASK'} = $branch_h{'scLocation'}{'ipNetmaskNumber'};

    #some consistency checks
    $branch_h{'EXTDHCP'} = $branch_h{'scLocation'}{'scDhcpExtern'};
    if (!defined $self->{'STATIC'}{'CONFIG'}{'LEASES2LDAP_EXTDHCP'}) {
        log_debug_printf(
            "external dhcp status not saved in STATIC data, branchserver either uninitialized or broken...",
        );
    }
    elsif ($branch_h{'EXTDHCP'} ne $self->{'STATIC'}{'CONFIG'}{'LEASES2LDAP_EXTDHCP'}) {
        log_warn_printf(
"external dhcp state changed in ldap to scDhcpExtern=%s please reinitialize branchserver (posInitBranchserver.sh -r) for the change to take effect, continuing with %s",
            $branch_h{'EXTDHCP'}, $self->{'STATIC'}{'CONFIG'}{'LEASES2LDAP_EXTDHCP'}
        );
    }
    ## TODO: check dhcp service object

    return \%branch_h;

} ## end sub get_branch_data


sub get_branch_services
{
    my ($self) = @_;
    my %services_h;
    my $lc    = $self->{'LDAPCACHE'};
    my $bbase = $self->{'STATIC'}{'BRANCHBASE'};
    #get server container
    my $srvcont_ar = $lc->FilterEntries($lc->getChildren($bbase), '(objectClass=scServerContainer)');
    my $srvcont_dn = $srvcont_ar->[0];
    if (!defined $srvcont_ar || @{$srvcont_ar} == 0) {
        log_crit_print("No scServerContainer defined under \'$bbase\', invalid LDAP configuration.");
        return undef;
    }
    elsif (@{$srvcont_ar} > 1) {
        log_warn_print("More than one scServerContainer defined under \'$bbase\', using $srvcont_dn, ignoring others.");
    }
    #get server
    my $srv_ar = $lc->FilterEntries($lc->getChildren($srvcont_dn), '(objectClass=scBranchServer)');
    my $srv_dn = $srv_ar->[0];
    if (@{$srv_ar} > 1) {
        log_warn_print("More than one scBranchServer defined under \'$srvcont_dn\', using $srv_dn, ignoring others.");
    }
    my $services_ar = $lc->FilterEntries($lc->getChildren($srv_dn), '(objectClass=scService)');
    foreach my $service_dn (@{$services_ar}) {
        my %service_h;
        $service_h{'dn'} = $service_dn;    #should be same as BRANCHBASE
        foreach my $att ($lc->{'LDS'}{$service_dn}[0]->attributes) {
            $service_h{$att} = $lc->{'LDS'}{$service_dn}[0]->get_value($att);
        }
        $services_h{ $service_h{'cn'} } = \%service_h;
    }

    return $srvcont_dn, $srv_dn, \%services_h;
} ## end sub get_branch_services


# load default container data
sub _get_default_cont
{
    my ($self) = @_;
    my %pxe_h;
    my $lc = $self->{'LDAPCACHE'};

    # get default kernel name and initrd name from ldap
    my $global_dn_ar = pop @{ $lc->FilterEntries($lc->getChildren($self->{'STATIC'}{'LDAPBASE'}), '(cn=global)') };
    if (!defined $global_dn_ar) {
        log_error_printf("Cannot find global scRefObjectContainer section! Invalid LDAP configuration, aborting.");
        return undef;
    }
    my $default_dc_dn = pop @{ $lc->FilterEntries($lc->getChildren($global_dn_ar), '(cn=default)') };
    if (!defined $default_dc_dn) {
        log_error_printf("Cannot find default scDistributionContainer section! Invalid LDAP configuration, aborting.");
        return undef;
    }
    my $default_dc_en = $lc->getLDAPentry($default_dc_dn);
    $pxe_h{'DEFAULT_DN'} = $default_dc_dn;
    $pxe_h{'KERNEL'}     = $default_dc_en->get_value('scKernelName');
    $pxe_h{'INITRD'}     = $default_dc_en->get_value('scInitrdName');

    return \%pxe_h;

} ## end sub _get_default_cont

# load default pxe data
sub _get_default_pxes
{
    my ($self, $pxe_hr) = @_;

    my @defpxes_a;
    my $lc = $self->{'LDAPCACHE'};

    #search under global default, sclocation, and tftp service
    my @dns = ($pxe_hr->{'DEFAULT_DN'}, $self->{'STATIC'}{'BRANCHBASE'});
    if (defined $self->{'SERVICES'}{'tftp'}) {
        @dns = (@dns, $self->{'SERVICES'}{'tftp'}{'dn'});
    }
    for my $dn (@dns) {
        my $result_pxe = POS::LdapUtil::ldap_gen_entrycache_successive(
            'BASE'  => $dn,
            'CACHE' => $self,
            'FILTER' =>
'(|(objectClass=scConfigFileTemplate)(objectClass=scConfigFileSyncTemplate)(objectClass=scPxeFileTemplate))'
        );
        @defpxes_a = (@defpxes_a, values(%{$result_pxe}));
    }
    $pxe_hr->{'DEF_OBJS'} = \@defpxes_a;

    return $pxe_hr;
} ## end sub _get_default_pxes

#compose pxe line and hash, add it to
#pxe hash structure
#$myhash use custom hash, do not compute
sub process_pxe
{
    my ($self, $pxe_hr, $myhash) = @_;
    $pxe_hr->{'DEFPARAMS'}  = '';
    $pxe_hr->{'SPECPARAMS'} = '';
    #params
    for my $entry (@{ $pxe_hr->{'DEF_OBJS'} }) {
        $pxe_hr->{'DEFPARAMS'} .= " " . $entry->get_value('scKernelParameters');
    }    #simple add, higher priority is latter. space in first place does not matter
    for my $entry (@{ $pxe_hr->{'SPEC_OBJS'} }) {
        $pxe_hr->{'SPECPARAMS'} .= " " . $entry->get_value('scKernelParameters');
    }    #simple add, higher priority is latter. space in first place does not matter

    #hash
    if (!defined $myhash) {
        my $hashsource =
          $pxe_hr->{'DEFPARAMS'} . $pxe_hr->{'SPECPARAMS'} .
          $pxe_hr->{'INITRD'} . $pxe_hr->{'KERNEL'} . $self->{'STATIC'}{'MACCFG'}{'MACPARAM'};  #no dependence on spaces
        $pxe_hr->{'PXEHASH'} = md5_hex($hashsource);
    }
    else {
        $pxe_hr->{'PXEHASH'} = $myhash;
    }

    my $initrdpart = 'initrd=' . $pxe_hr->{'INITRD'};
    my $macpart =
      $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP_PREFIX'} . '="' . $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP'} . '"';
    $pxe_hr->{'MACPART'} = $macpart;
    my $hashpart = 'POS_KERNEL_PARAMS_HASH=' . $pxe_hr->{'PXEHASH'};

    #params and append line
    $pxe_hr->{'ALLPARAMS'}  = $pxe_hr->{'DEFPARAMS'} . ' ' . $macpart . ' ' . $pxe_hr->{'SPECPARAMS'};
    $pxe_hr->{'APPENDLINE'} = $initrdpart . ' ' . $pxe_hr->{'ALLPARAMS'} . ' ' . $hashpart;

    return $pxe_hr;
} ## end sub process_pxe


#generate cash register hash from cached ldap data
sub load_crs_from_ldap
{
    my ($self) = @_;

    my $crhash_r = POS::LdapUtil::get_crs_from_ldap(
        'CACHE'    => $self,
        'BASE'     => $self->{'STATIC'}{'BRANCHBASE'},
        'UPTOBASE' => $self->{'STATIC'}{'LDAPBASE'},
    );

    #compute ordering ({'INDEX'})
    my %depthhash;    #hash by depth
    foreach my $dn (keys %{$crhash_r}) {    #create intermediate hash with keys being the depth level
        my $depth = $crhash_r->{$dn}{'DEPTH'};
        my %tmphash;
        $tmphash{'dn'} = $dn;
        $tmphash{'cn'} = $crhash_r->{$dn}{'cn'};
        push @{ $depthhash{$depth} }, \%tmphash;
    }

    my $index = 0;
    foreach my $key (sort { $b <=> $a } (keys(%depthhash)))
    {                                       #lowest level to highest, and within each level by alphabet
        my @svallist = sort {               #within each depth lever, sort by cnames by alphabet
            $a->{'cn'} cmp $b->{'cn'}
        } @{ $depthhash{$key} };
        foreach my $item (@svallist) {
            $crhash_r->{ $item->{'dn'} }{'INDEX'} = $index;
            #$crlist[$index]=$crhash{$item->{'dn'}};
            $index++;
        }
    }

    return ($crhash_r);


} ## end sub load_crs_from_ldap


#if an ID is already used in a registered scWorkstation
#also test duplicites (which should not happed, but can be created manually)
sub _purge_id_pool
{
    my ($self, $idshash_ref, $wsent_ref) = @_;
    my %returnhash = %{$idshash_ref};
    my %returnwshash;
    my %idws;    #has of workstations id, keys are ids, value is multiplicity
    foreach my $ws (values(%{$wsent_ref})) {
        my $wsid = $ws->get_value('scId');
        next if !defined $wsid;    #skip this if no Id, it may be workstation not using role based approach
        my $dn = $ws->dn();        # be careful with empty Id
        if (defined $idws{$wsid}) {
            my $dupldn = $idws{$wsid};
            log_warn_printf('Duplicate ID \'%s\' found for workstation \'%s\' and \'%s\' ', $wsid, $dn, $dupldn);
            #log_error_printf('Duplicate ID \'%s\' found for workstation \'%s\' and \'%s\' ', $wsid, $dn, $dupldn);
        }
        else {                     #no duplicate
            $idws{$wsid} = $dn;
            if (defined $returnhash{$wsid}) {    #remove from main pool, not if duplicated since it was already removed
                my $index = $returnhash{$wsid}{'INDEX'};
                delete $returnhash{$wsid};
                #also decrease all indices larger than $index, otherwise we will have empty lines in the new idlist.txt
                foreach my $id (keys(%returnhash)) {
                    $returnhash{$id}{'INDEX'} -= 1 if $returnhash{$id}{'INDEX'} > $index;
                }
                #put in the workstation hash
                $returnwshash{$wsid} = $dn;
            }
            else {                               #not in the main pool!
                log_warn_printf('ID \'%s\' of the   \'%s\' workstation is not in the scIdPool! ', $wsid, $dn);
                #log_error_printf('ID \'%s\' of the  \'%s\' workstation is not in the scIdPool! ', $wsid, $dn);
            }
        }
    }
    return (\%returnhash, \%returnwshash);
} ## end sub _purge_id_pool


#one id on each line
# returns list, key is idshash, values are ROLEDN, ROLENAME multiplicities (just suppl. info)
#undef if error
sub get_ids_from_file()
{
    my ($self, $pathname) = @_;
    my %ideshash;
    $self->{'IDHASH'}{'FILETIME'} = 0 if !defined $self->{'IDHASH'}{'FILETIME'};    #init
    my $idfiletime = (stat($pathname))[9];
    if (defined $idfiletime) {
        if ($idfiletime == $self->{'IDHASH'}{'FILETIME'} && defined $self->{'IDHASH'}{'FILE'}) {
            # if file not changed, jusst return data from memory
            log_debug_printf("reading data for \'%s\' from stored hash", $pathname);
            %ideshash = %{ $self->{'IDHASH'}{'FILE'} };
        }
        else {
            log_debug_printf("reading file \'%s\'", $pathname);
            if (open(FH, '<', $pathname)) {
                my @lines = <FH>;
                my $index = 0;      #line number corrected for repeated entries
                foreach my $line (@lines) {
                    $line =~ s/\n//;      #remove new line
                    $line =~ s/\s*$//;    #remove leading whitespace
                    $line =~ s/^\s*//;    #remove trailing whitespace
                    if (defined $ideshash{$line}) {
                        $ideshash{$line}{'COUNT'}++;
                        log_warn_printf('Multiple occurence (%s times) of the same ID \'%s\' found in %s !',
                            $ideshash{$line}, $line, $pathname);
                    }
                    else {
                        $ideshash{$line}{'INDEX'}        = $index;
                        $ideshash{$line}{'MULTILINEIDX'} = 0;
                        $ideshash{$line}{'COUNT'}        = 1;
                        $index++;
                    }
                }
                close(FH);
                $self->{'IDHASH'}{'FILETIME'} = $idfiletime;
            }
            else {
                log_warn_printf('Can\'t open file \'%s\' ', $pathname);
            }
        }
    }

    return (\%ideshash);
} ## end sub get_ids_from_file()


#correct indices to cover whole range, no unused numbers
sub _pack_indices
{
    my ($self, $rolelist_hash_ref) = @_;    #has which has roles as values that we want to reindex
    my @rolelist = values(%{$rolelist_hash_ref});
    my @rsorted = sort { $a->{'INDEX'} <=> $b->{'INDEX'}; } @rolelist;

    for (my $i = 0; $i < @rsorted; $i++) {
        $rsorted[$i]{'INDEX'} = $i;
    }
    return 1;
} ## end sub _pack_indices

# trims away excessive '/' and """
sub trimpath
{
    my ($self, $path) = @_;
    $path = $self if (!defined $path);
    $path =~ s@"@@g;    #"# fix syntax hilight
    $path =~ s@/+@/@g;
    $path =~ s@/$@@g;
    return $path;
} ## end sub trimpath


#filters roles using scAllowedRolesDn attribute of the given scWorkstation
sub _filter_roles_mac
{
    my ($self, $ldaphash, $wshash) = @_;
    my %allowedhash;
    my @roles = $wshash->get_value('scAllowedRolesDn');
    #if (! @roles) {#empty filter means all - NO!
    #  @roles=keys %{$ldaphash};
    #}
    foreach my $roledn (@roles) {
        if (defined $ldaphash->{$roledn}) {
            my %rolehash =
              %{ $ldaphash->{$roledn} };    #copy values, not just refs (indices must be independently changed)
            $allowedhash{$roledn} = \%rolehash;
        }
        else {
            log_warn_printf('Role \'%s\' allowed for \'%s\' does not exist in LDAP, ignoring...',
                $roledn, $wshash->dn());
        }
    }

    return \%allowedhash;
} ## end sub _filter_roles_mac


sub _filter_default_cr_roles
{
    my ($self, $ldaphash) = @_;    #$ldaphash = all roles hash
    my %crdhwhash;
    foreach my $roledn (keys %{$ldaphash}) {
        my ($cashreg_ref, $errcr) = $self->get_role_cr($roledn, 'default');
        if (!$errcr->{'CODE'} && defined $cashreg_ref) {
            my %tmphash = %{ $ldaphash->{$roledn} }; #copy values, not just refs (indices must be independently changed)
            $crdhwhash{$roledn} = \%tmphash;
        }
    }
    #$self->_pack_indices(\%crdhwhash); #not yet, will be also joined
    return \%crdhwhash;
} ## end sub _filter_default_cr_roles


#TODO: deprecate, unify with 'find_hwtype_cr' ??
#get corresponding cashregister under the role and the default one also
sub get_role_cr
{
    my ($self, $roledn, $hwcode) = @_;
    my $cashreg_ref;
    my $cr_same;
    my %error_hash;
    #my $error_ref=\%error_hash;

    my $lc = $self->{'LDAPCACHE'};
    my @cshrs = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($roledn, 'scCashRegister') });
    #my $result =
    #  $self->{'LDAP'}->search(base => $roledn, scope => 'one', filter => 'objectClass=scCashRegister', attrs => []);
    #find_cashreg($cashreg_refs);
    if (@cshrs) {
        foreach my $cshr (@cshrs) {
            if ($cshr->get_value('scCashRegisterName') eq $hwcode) {
                if (!defined $cashreg_ref) {
                    $cashreg_ref = $cshr;
                }
                else {
                    $error_hash{'MESSAGE'} = sprintf(
"Duplicity in scCashRegister object (under \'%s\' role) exists for HWTYPE=%s : \'%s\' and \'%s\', aborting registration..",
                        $roledn, $hwcode, $cashreg_ref->{'dn'}, $cshr->{'dn'});
                    ($hwcode eq 'default') ? $error_hash{'CODE'} = 33 : $error_hash{'CODE'} =
                      34;    #distinguish default fail from specific fail
                    last;
                }
            }
        }
    }
    else {
        $error_hash{'CODE'} = 32;
        $error_hash{'MESSAGE'} =
          sprintf("No scCashRegister object found under role \'%s\', aborting registration..", $roledn);
    }
    return ($cashreg_ref, \%error_hash);    #can be empty
} ## end sub get_role_cr


#gather create hash with keys being hwtype
# contains roles that: have hwtype allowed by scAllowedHwTypes AND contain scCashregister with this role
#
#note: reads LDAP
sub _filter_roles_hwtype
{
    my ($self, $ldaphash) = @_;             #$ldaphash = all roles hash
    my %allhwhash;
    foreach my $roledn (keys %{$ldaphash}) {
        my @cshrs_hw;
       #my $result =
       #  $self->{'LDAP'}->search(base => $roledn, scope => 'one', filter => 'objectClass=scCashRegister', attrs => []);
       #find_cashreg($cashreg_refs);
        my $lc = $self->{'LDAPCACHE'};
        my @cshrs = map($lc->getLDAPentry($_), @{ $lc->getChildrenByClass($roledn, 'scCashRegister') });

        foreach my $cshr (@cshrs) {
            push @cshrs_hw, $cshr->get_value('scCashRegisterName');
        }
        my @filtrolelist = $ldaphash->{$roledn}{'ENTRY'}->get_value('scAllowedHwTypes');
        my $intersectlist_ref;
        if (@filtrolelist) {
            $intersectlist_ref = $self->_intersect_lists(\@filtrolelist, \@cshrs_hw);
        }
        else {    #empty filter <=> all allowed
            $intersectlist_ref = \@cshrs_hw;
        }

        my @finalhwlist;
        #make sure no defaults are there
        foreach my $item (@{$intersectlist_ref}) {
            push(@finalhwlist, $item) unless ($item eq 'default');
        }
        foreach my $hw (@finalhwlist) {
            my %rolehash =
              %{ $ldaphash->{$roledn} };    #copy values, not just refs (indices must be independently changed)
            my $outhwtype = $self->hw_to_outhw($hw);
            $allhwhash{$outhwtype}{$roledn} = \%rolehash;
        }
    }

    #not yet, will be joined
    #compress indices
    #foreach my $hwhash (keys(%allhwhash)) {
    #    my @rolelist;
    #    @rolelist = values %{ $allhwhash{$hwhash} };
    #    $self->_pack_indices(\@rolelist)
    #      ;    #members of this array still point to main data structure, indices will be changed there
    #}

    return \%allhwhash;
} ## end sub _filter_roles_hwtype

sub _intersect_lists
{
    my ($self, $lra, $lrb) = @_;
    my %ha;
    my %hb;
    my @intersection;

    foreach my $p (@{$lra}) {
        $ha{$p} = 1;
    }
    foreach my $p (@{$lrb}) {
        $hb{$p} = 1;
    }
    foreach my $h (keys %ha) {
        push @intersection, $h if (defined $hb{$h});
    }
    return \@intersection;
} ## end sub _intersect_lists


#joins lists of roles
# keys are dn, vaules are role hashes
sub _join_role_lists
{
    my ($self, @rolelists) = @_;
    my %joinhash;
    foreach my $dnhash_ref (@rolelists) {
        foreach my $dn (keys %{$dnhash_ref}) {
            my %tmphash = %{ $dnhash_ref->{$dn} };
            $joinhash{$dn} = \%tmphash;    #copy key, not ref on key;
        }
    }
    my @rolelist = values %joinhash;
    # $self->_pack_indices(\@rolelist); # do manually later
    return (\%joinhash);
} ## end sub _join_role_lists


#finds and set defualt role if the provided role hash contains default role
# sets default role or first (index=0) if no default found
# if no FIRST returned, no role at al is present
# moved from posleaseMonitor
sub _find_default_roles
{
    my ($self, $storehash, $roleshash) = @_;    #$ldaphash = all roles hash
    foreach my $dn (keys(%{$roleshash})) {
        my $rolehash = $roleshash->{$dn};
        $storehash->{'DEFAULT'} = $dn if ($rolehash->{'scRoleName'} eq 'default');
        $storehash->{'FIRST'}   = $dn if ($rolehash->{'INDEX'} == 0);                #first index is zero
    }
    return 1;

} ## end sub _find_default_roles


sub get_roles_from_file()
{
    my ($self, $pathname, $rolehash_ref) = @_;
    my %froleshash;
    $rolehash_ref->{'FILETIME'} = 0 if !defined $rolehash_ref->{'FILETIME'};         #init
    my $rolefiletime = (stat($pathname))[9];
    if (defined $rolefiletime) {
        if ($rolefiletime == $rolehash_ref->{'FILETIME'} && defined $rolehash_ref->{'FILE'}) {
            # if file not changed, and FILE hash already stored, just return data from memory
            log_debug_printf("reading data for \'%s\' from stored hash", $pathname);
            %froleshash = %{ $rolehash_ref->{'FILE'} };
        }
        else {
            log_debug_printf("reading file \'%s\'", $pathname);
            if (open(FH, '<', $pathname)) {
                my @lines = <FH>;
                my $index = 0;      #line number corrected for repeated entries
                foreach my $line (@lines) {
                    $line =~ s/\n//;      #remove new line
                    $line =~ s/\s*$//;    #remove leading whitespace
                    $line =~ s/^\s*//;    #remove trailing whitespace
                    my @parts      = split('\|', $line);
                    my $dn         = shift @parts;
                    my $scRoleName = shift @parts;
                    my $scRoleDescription =
                      join('|', @parts);    #not really expecting role description to contain '|', but anyway..
                    if (!defined $dn) {
                        log_warn_printf("Dubious line \'%s\' in \'%s\' !", $line, $pathname);
                        $dn = 'UNDEF';
                    }
                    $froleshash{$dn}{'dn'}                = $dn;
                    $froleshash{$dn}{'scRoleName'}        = $scRoleName;
                    $froleshash{$dn}{'scRoleDescription'} = $scRoleDescription;
                    $froleshash{$dn}{'INDEX'}             = $index;
                    $index++;
                }
                close(FH);
                $rolehash_ref->{'FILETIME'} = $rolefiletime;
            }
            else {
                log_warn_printf('Can\'t open file \'%s\' ', $pathname);
            }
        }
    }
    return (\%froleshash);
} ## end sub get_roles_from_file()

#write idfile from hash data
sub write_idfile()
{
    my ($self, $pathname, $idhash_ref) = @_;
    my ($fname, $path) = fileparse($pathname);
    my $tmppathname = $path . '/' . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $fname;
    my @idlist;    #order by ldap multiline indices
    foreach my $id (keys %{$idhash_ref}) {
        my $lineidx = $idhash_ref->{$id}{'INDEX'};
        $idlist[$lineidx] = $id;
        #printf(@tmplist);
        #	  my @tmpids=@{$tmpids} if (defined $tmpids);
        #	  @tmpids=(@tmpids,$id); #in both keys???
        #	  $tmphash{$lineidx}=\@tmpids;
    }
    if (open(FH, '>', $tmppathname)) {
        log_info_printf("writting file \'%s\' ", $pathname);
        foreach my $line (@idlist) {
            print FH "$line\n";
        }
        close(FH);
        system('mv', $tmppathname, $pathname);
        return 0;
    }
    else {
        log_warn_printf('Can\'t open file \'%s\' ', $pathname);
        return 1;
    }
} ## end sub write_idfile()

#write rolefile from hash data
sub write_rolefile()
{
    my ($self, $pathname, $rolehash_ref) = @_;
    my ($fname, $path) = fileparse($pathname);
    my $tmppathname = $path . '/' . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $fname;
    my @rolelist;
    if (open(FH, '>', $tmppathname)) {
        log_info_printf("writting file \'%s\' ", $pathname);
        foreach my $dn (keys %{$rolehash_ref}) {    #create intermediate hash with keys being the depth level
            $rolelist[$rolehash_ref->{$dn}{'INDEX'}] = $rolehash_ref->{$dn};
        }
        foreach my $role (@rolelist) {
            print FH "$role->{'dn'}|$role->{'scRoleName'}|$role->{'scRoleDescription'}\n";
        }

#        my $depth=$rolehash_ref->{$dn}{'DEPTH'};
#	    my @tmplist;
#	    @tmplist=@{$leveltmphash{$depth}} if defined $leveltmphash{$depth};
#	    my %tmphash;
#	    $tmphash{'dn'}=$dn;
#	    $tmphash{'cn'}=$rolehash_ref->{$dn}{'cn'};
#	    $tmphash{'scRoleName'}=$rolehash_ref->{$dn}{'scRoleName'};
#	    $tmphash{'scRoleDescription'}=$rolehash_ref->{$dn}{'scRoleDescription'};
#	    @tmplist=(@tmplist,\%tmphash);
#	    $leveltmphash{$depth}=\@tmplist;
#	  }
#      foreach my $key (sort {$b<=>$a} (keys (%leveltmphash))) {  #lowest level to highest, and within each level by alphabet
#	    my @svallist=sort {
#	  	  $a->{'scRoleName'} cmp $b->{'scRoleName'}
#	    } @{$leveltmphash{$key}};
#	    foreach my $role (@svallist) {
#	      print FH "$role->{'dn'}|$role->{'scRoleName'}|$role->{'scRoleDescription'}\n";
#	    }

        #sort keys by alphabet by cn and write them

        close(FH);
        system('mv', $tmppathname, $pathname);
        return 0;
    }
    else {
        log_warn_printf('Can\'t open file \'%s\' ', $pathname);
        return 1;
    }
} ## end sub write_rolefile()


#TODO: maybe combine those two bellow into one?
#deletes all role list files (rolelist, rolelist.MAC, rolelist.HWTYPE)
sub checkdelete_rolefiles()
{
    my ($self, $basepathname) = @_;
    #my ($basename, $path) = fileparse($basepathname);
    my $dirbk    = $self->{'STATIC'}{'KIWI_BCK_DIR'};
    my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_ROLELIST_NAME'};
    my @cmds;
    my @files =
      (keys %{ $self->get_basefiles($basepathname) });    # perl-style file globbing operator  # use gre to prefilter
    foreach my $file (@files) {
        chomp $file;
        @cmds = (@cmds, $self->{'STATIC'}->backup($file, $dirbk, $archname));
        #maybe more specific filter if rolelist* is not good enough
        #$filename='rolelist.00;11;25;A7;D6;1E.02c425157ecd32f259548b33402ff6d3';
    }
    $self->{'STATIC'}->runcmds(\@cmds);
} ## end sub checkdelete_rolefiles()
#deletes all role list files (rolelist, rolelist.MAC, rolelist.HWTYPE)
sub checkdelete_idfiles()
{
    my ($self, $basepathname) = @_;
    #my ($basename, $path) = fileparse($basepathname);
    my $dirbk    = $self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'} . "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BACKUP_BASE'};
    my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_IDLIST_NAME'};
    my @cmds;
    my @files =
      (keys %{ $self->get_basefiles($basepathname) });    # perl-style file globbing operator  # use gre to prefilter
    foreach my $file (@files) {
        chomp $file;
        @cmds = (@cmds, $self->{'STATIC'}->backup($file, $dirbk, $archname));
        #maybe more specific filter if rolelist* is not good enough
        #$filename='rolelist.00;11;25;A7;D6;1E.02c425157ecd32f259548b33402ff6d3';
    }
    $self->{'STATIC'}->runcmds(\@cmds);
} ## end sub checkdelete_idfiles()


#load pxe data from pxe file
# if file newer than already stored

sub load_pxefile
{
    my ($self, $pxe_hr) = @_;
    my %fpxehash;
    my $pathname = trimpath($pxe_hr->{'PATH'} . "/" . $pxe_hr->{'NAME'});
    if (!-e $pathname) {
        log_warn_printf("invalid pxe pathname \'$pathname\' specified!");
        return undef;
    }
    $pxe_hr->{'FILETIME'} = 0 if !defined $pxe_hr->{'FILETIME'};    #init
    my $pxefiletime = (stat($pathname))[9];
    if ($pxefiletime == $pxe_hr->{'FILETIME'} && defined $pxe_hr->{'FILE'}) {
        # if file not changed, and FILE hash already stored, just return data from memory
        log_debug_printf("no changes, keeping stored data for \'%s\'", $pathname);
        %fpxehash = %{ $pxe_hr->{'FILE'} };
        return \%fpxehash;
    }
    log_debug_printf("reading file \'%s\'", $pathname);
    if (open(FH, '<', $pathname)) {
        my @lines = <FH>;
        close(FH);
        my $appendline;
        my $kernel;
        #        my $initrd;
        for my $line (@lines) {
            if ($line =~ /^\s*kernel\s*(.*)\s*$/) {
                $kernel = $1;
            }
            if ($line =~ /^\s*append\s*(.*)\s*$/) {
                $appendline = $1;
            }
            last if $kernel && $appendline;
        }
        $appendline =~ /^initrd=(.*?)\s+/;
        $fpxehash{'INITRD'}     = $1;
        $fpxehash{'KERNEL'}     = $kernel;
        $fpxehash{'APPENDLINE'} = $appendline;
        $pxe_hr->{'FILETIME'}   = $pxefiletime;    #push new filetime
        $pxe_hr->{'FILE'}       = \%fpxehash;      #push new filedata
    }
    else {
        log_warn_printf('Can\'t open file \'%s\' ', $pathname);
    }
    return \%fpxehash;
} ## end sub load_pxefile

sub write_grub2_pxefile
{
    my ($self, $pxe_hr) = @_;
    my %fpxehash;

    my $name = $self->name_to_grub_name($pxe_hr->{'NAME'});
    my $template = "$pxe_hr->{'TEMPLATE'}";
    $template =~ s/pxelinux/grub2/;

    my $kernel = $pxe_hr->{'LDAP'}{'KERNEL'};
    my $initrd = $pxe_hr->{'LDAP'}{'APPENDLINE'};
    $initrd =~ s/.*initrd=([^\s]*).*/$1/;
    my $cmdline = $pxe_hr->{'LDAP'}{'APPENDLINE'};
    $cmdline =~ s/.*initrd=[^\s]*[\s*](.*)/$1/;

    my $pxe_pathname = trimpath($pxe_hr->{'PATH'} . "/" . $name);
    my $pxe_tmp_pathname =
      trimpath($pxe_hr->{'PATH'} . "/" . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $name);
    open(FH, '<', $template);
    my @pxe_config = <FH>;
    close FH;
    foreach my $line (@pxe_config) {
        $line =~ s/\@LINUXKERNEL\@/$kernel/x;
        $line =~ s/\@INITRD\@/$initrd/x;
        $line =~ s/\@COMMANDLINE\@/$cmdline/x;
    }
    open(FH, '>', $pxe_tmp_pathname);
    print FH join("", @pxe_config);
    close FH;
    my @cmds;

    if (-e $pxe_tmp_pathname) {
        push(@cmds, ['chmod', 0644, $pxe_tmp_pathname]);
        push(@cmds, ['move', $pxe_tmp_pathname, $pxe_pathname]);
    }
    else {
        log_error_printf('Can\'t write new pxe file \'%s\' ', $pxe_tmp_pathname);
    }
    return @cmds;
} ## end sub write_grub2_pxefile

#write new pxe file
sub write_pxefile
{
    my ($self, $pxe_hr) = @_;
    my %fpxehash;
    my $pxe_pathname = trimpath($pxe_hr->{'PATH'} . "/" . $pxe_hr->{'NAME'});
    my $pxe_tmp_pathname =
      trimpath($pxe_hr->{'PATH'} . "/" . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $pxe_hr->{'NAME'});
    open(FH, '<', "$pxe_hr->{'TEMPLATE'}");
    my @pxe_config = <FH>;
    close FH;
    foreach my $line (@pxe_config) {
        $line =~ s/^(\s*)\@LINUXKERNEL\@.*$/$1kernel $pxe_hr->{'LDAP'}{'KERNEL'}/x;
        $line =~ s/^(\s*)\@APPENDLINE\@.*$/$1append $pxe_hr->{'LDAP'}{'APPENDLINE'}/x;
    }
    open(FH, '>', $pxe_tmp_pathname);
    print FH join("", @pxe_config);
    close FH;
    my @cmds;

    if (-e $pxe_tmp_pathname) {
        push(@cmds, ['chmod', 0644, $pxe_tmp_pathname]);
        push(@cmds, ['move', $pxe_tmp_pathname, $pxe_pathname]);
    }
    else {
        log_error_printf('Can\'t write new pxe file \'%s\' ', $pxe_tmp_pathname);
    }

    push(@cmds, $self->write_grub2_pxefile($pxe_hr));
    return $self->{'STATIC'}->runcmds(\@cmds);    #delete and maybe create new pxefile if needed

} ## end sub write_pxefile


## @method $ get_config_list()
# Gathers information on existing config.MAC files
# @par Details:
# Reads list of all config.MAC files located in $basedir,
# gathers name, MAC, hash id, creation time and checks whether
# file is plain file or link.
# @param $basedir directory containing the rollback files
# @return \\%%cfgmachash->{$mac}{$hash} hash reference containing:
# @arg @c 'PATHNAME' image file name
# @arg @c 'MAC' MAC address of the terminal
# @arg @c 'MTIME' time of the rolled back config file creation
# @arg @c 'LINK' where link points for file being link, undef for otherwise.
# @arg @c 'LINKMAC' MAC address from the linked cfg file.
# get list of config. mac files
#TODO: unify with hwtype and bootversion file parsing?
sub get_config_list
{
    my ($self) = @_;
    my %cfgmachash;
    my $basedir = $self->{'STATIC'}{'KIWI_DIR'};
    my $prefix  = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};    #i.e. 'config'
    my $filtmp  = $self->{'STATIC'}{'MACCFG'}{'MACFILTER'};
    my $mp      = $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP'};
    my @lines   = <$basedir/$prefix*>;                                 # perl-style file globbing operator
    foreach my $pathname (@lines) {
        chomp $pathname;
        my $cfgfile = basename($pathname);
        if ($cfgfile =~ /^$prefix\.([\d$filtmp]*)$/) {                 #config.MAC
            my $outmac = $1;
            if ($outmac) {
                my $mac =
                  $self->{'STATIC'}{'MACCFG'}->outmac_to_mac($outmac);    # for internal purpose, always standard format
                $cfgmachash{$mac}{'PATHNAME'} = $pathname;
                $cfgmachash{$mac}{'MAC'}      = $mac;
                $cfgmachash{$mac}{'MTIME'}    = (stat($pathname))[9];
                $cfgmachash{$mac}{'LINK'}     = undef;
                $cfgmachash{$mac}{'LINKMAC'}  = undef;
                if (-l $pathname) {
                    $cfgmachash{$mac}{'LINK'} = readlink($pathname);
                    $cfgmachash{$mac}{'LINK'} =~ /^$prefix\.([\d$filtmp]*)$/;    #TODO: have a common function for this
                    $cfgmachash{$mac}{'LINKMAC'} = $self->{'STATIC'}{'MACCFG'}->outmac_to_mac($1);
                }


            }
        }
    }
    return \%cfgmachash;
} ## end sub get_config_list


#get all files with rolelist basename
# put into hash (better for existence checks)
sub get_basefiles()
{
    my ($self, $basepathname) = @_;
    my $filename = basename($basepathname);
    my $dirname  = dirname($basepathname);
    opendir(DIR, $dirname);
    my @files = grep(/$filename/, readdir(DIR));    # use regex in filename to grab more file at once
    my %filesh;
    foreach my $file (@files) {
        $filesh{ $self->trimpath($dirname . '/' . $file) } = 1;
    }
    return \%filesh;

} ## end sub get_basefiles()


##TODO: improve and read only if changed, etc..
# read dhcp leasesfile
sub get_current_dhcp_leases
{
    my $self = shift;
    ## Reading the dhcpd.leases file;
    my $dhcpleases = $self->{'DHCP'}{'LEASES'}{'PATHNAME'};
    my $leasetime  = (stat($dhcpleases))[9];
    if (!defined $leasetime) {
        $self->{'DHCP'}{'LEASES'}{'HASH'} = undef;
        log_warn_printf('Can\'t stat %s, IPs from leases will not be used', $dhcpleases);
        return 0;
    }

    if ($leasetime == $self->{'DHCP'}{'LEASES'}{'FILETIME'}) {
        log_debug_printf('No change in leases file %s, keeping old data', $dhcpleases);
    }
    else {
        log_debug_printf("reading dhcp leases \'%s\'", $dhcpleases);
        my $dhcpentries = POS::DhcpEntries->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});
        my ($leases_hash_ref) = $dhcpentries->read_dhcp_leases($dhcpleases);
        if (defined $leases_hash_ref) {
            $self->{'DHCP'}{'LEASES'}{'HASH'}     = $leases_hash_ref;
            $self->{'DHCP'}{'LEASES'}{'FILETIME'} = $leasetime;
            log_notice_printf('Leases file %s read ok, IPs in leases can be used during registration.', $dhcpleases)
              ;    #even if empty
        }
        else {
            $self->{'DHCP'}{'LEASES'}{'HASH'} = undef;
            log_warn_printf('Can\'t read from %s, IPs from leases will not be used', $dhcpleases);
            return 0;
        }
    }
    return 1;

} ## end sub get_current_dhcp_leases

##TODO: improve and read only if changed, etc..
# read dhcp leasesfile
sub get_current_dhcp_config
{
    my $self = shift;
    ## Reading the dhcpd.leases file;
    my $dhcpcfg = $self->{'DHCP'}{'CONFIG'}{'PATHNAME'};
    my $cfgtime = (stat($self->{'DHCP'}{'CONFIG'}{'PATHNAME'}))[9];
    if (!defined $cfgtime) {
        $self->{'DHCP'}{'CONFIG'}{'HASH'} = undef;
        log_warn_printf('Can\'t stat %s, IPs from leases will not be used', $dhcpcfg);
        return 0;
    }

    if ($cfgtime == $self->{'DHCP'}{'CONFIG'}{'FILETIME'}) {
        log_debug_printf('No change in leases file %s, keeping old data', $dhcpcfg);
    }
    else {
        log_debug_printf("reading dhcp config \'%s\'", $dhcpcfg);
        my $dhcpentries = POS::DhcpEntries->new('STATIC' => $self->{'STATIC'}, 'CACHE' => $self->{'CACHE'});
        my ($cfg_hash_ref) = $dhcpentries->read_dhcp_config($dhcpcfg);
        if (defined $cfg_hash_ref) {
            $self->{'DHCP'}{'CONFIG'}{'HASH'}     = $cfg_hash_ref;
            $self->{'DHCP'}{'CONFIG'}{'FILETIME'} = $cfgtime;
            log_debug_printf(
                'read %s host entries for: %s',
                int keys %{$cfg_hash_ref},
                join(', ', keys %{$cfg_hash_ref})
            );    #even if empty
            log_notice_printf('Dhcp config file %s read ok.', $dhcpcfg);    #even if empty
        }
        else {
            $self->{'DHCP'}{'CONFIG'}{'HASH'} = undef;
            log_warn_printf('Can\'t read from %s, no dhcp entries were loaded.', $dhcpcfg);
            return 0;
        }
    }
    return 1;

} ## end sub get_current_dhcp_config


#translate HWTYPE string if ugly
sub hw_to_outhw
{
    my ($self, $hwtype) = @_;
    my $outhwtype = $hwtype;
    $outhwtype =~ s/([^-_A-Za-z0-9])/sprintf("%%%02X",ord($1))/egs;

    return $outhwtype;
} ## end sub hw_to_outhw


# reads and resets error of this object
# errdata hash naming must be compatible
# with names used in posLeaseMonitor.pm
# (this object is common for all workstations,
#  specific error data should be removed before
#  another workstation is checked/registered)
sub get_error
{
    (my $self) = @_;
    my %errdata;
    %errdata                    = %{ $self->{'ERROR'} };
    $self->{'ERROR'}{'CODE'}    = 0;
    $self->{'ERROR'}{'MESSAGE'} = '';
    return \%errdata;
} ## end sub get_error

# transform pxe name to grub name
sub name_to_grub_name
{
    my ($self, $name) = @_;
    $name =~ s/-/:/g;
    $name .= '.grub2.cfg';
    return $name;
}
