## @file
# SLEPOS static data cache module
# - provides data that are only actualized at the startup
#   of the posleases2ldap.pl and not to change
#   (or change during run is not supported)
# - mostly derived from admin(branch)server.conf (->CONFIG hash)

package POS::posDataStatic;

use strict;
use warnings;

use Exporter;
use POS::Syslog;
use POS::MAC;
use File::Basename;
use POS::LdapUtil;
use POS::SyncImages;
use File::Path;
use File::Copy;

our $VERSION = 0.1;

sub new
{
    my $package = shift;    ## packagename

    # Definition of object data, should be only general data common to all workstation instances
    my $data = {
        'MODE'       => 'branchservice',    ## 'admin', 'branchadmin', 'branchservice'
        'CONFIG'     => undef,              ## POS::Config, data from branchserver configuraton file  #todo
        'LDAPBASE'   => undef,              ## LDAP-global-Base - o=...,c=..
        'BRANCHBASE' => undef,              ## LDAP-Base - dn of used scLocation
        'EXTDHCP'    => undef,              ## external DHCP usage flag
        'MACCFG'     => undef,              ## MAC object, used to check and transform MAC

        # ldap credentials
        'LDAP_ACCOUNT'  => undef,
        'LDAP_PASSWORD' => undef

    };

    $data->{'ERROR'}{'CODE'}    = 0;
    $data->{'ERROR'}{'MESSAGE'} = '';
    set_var($data, @_);
    _init_config_data($data);
    _init_mac_proc($data) unless $data->{'ERROR'}{'CODE'};
    if ($data->{'ERROR'}{'CODE'}) {
        log_alert_print(
            "Static Module init error $data->{'ERROR'}{'CODE'} : $data->{'ERROR'}{'MESSAGE'}, cannot continue");
        return undef;

    }

    return bless($data);
} ## end sub new

## note: return $self->get_error() for externally accessible methods,
##       return undef and parse error directly for internally used methods

#adds/(re)sets a variable of this object
sub set_var
{
    (my $self, my %data) = @_;
    %$self = (%$self, %data);
} ## end sub set_var

# reads config data structure,
# extract some important specific values
sub _init_config_data
{
    (my $self) = @_;
    #config can be given explicitly
    if (!defined $self->{'CONFIG'}) {
        eval { $self->{'CONFIG'} = POS::Config->new($self->{'MODE'}) };
        if ($@) {
            log_crit_print("No suitable configuration file found!");
        }
    }
    ## TODO: read and (de)activate dynamically
    if (!$self->{'CONFIG'}) {
        $self->{'ERROR'}{'CODE'}    = -1;
        $self->{'ERROR'}{'MESSAGE'} = 'Can\'t read configuration, exit now';
        #log_alert_print('Can\'t read configuration, exit now');
        # Unable to load configuration
        return undef;
    }

    $self->{'EXTDHCP'} = POS::LdapUtil::boolean_to_integer($self->{'CONFIG'}{'LEASES2LDAP_EXTDHCP'});

    $self->{'TFTP_DIR'}     = trimpath($self->{'CONFIG'}{'TFTP_BASE'});
    $self->{'KIWI_DIR'}     = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CR_BASE'});
    $self->{'KIWI_BCK_DIR'} = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CR_BACKUP_BASE'});
    $self->{'ROLLBACK_DIR'} =
      trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CR_ROLLBACK_BASE'});   #/srv/tftpboot / KIWI/rollback
    $self->{'UPLOAD_DIR'}     = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_UPLOAD_BASE'});
    $self->{'UPLOAD_BCK_DIR'} = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_UPLOAD_BACKUP_BASE'});
    $self->{'TFTP_IMAGE_DIR'} = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_IMAGE_BASE'});
    $self->{'TFTP_BOOT_DIR'}  = trimpath($self->{'TFTP_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_BOOT_BASE'});
    $self->{'TFTP_CUSTOM_BOOT_DIR'} =
      trimpath($self->{'TFTP_BOOT_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CUSTOM_BOOT_SUBDIR'});
    $self->{'TFTP_PXE_DIR'} = trimpath($self->{'TFTP_BOOT_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_PXELINUXCFG_BASE'});

    $self->{'ROLE_PATHNAME'} = trimpath($self->{'KIWI_DIR'} . '/' . $self->{'CONFIG'}{'POS_ROLELIST_NAME'});
    $self->{'ID_PATHNAME'}   = trimpath($self->{'KIWI_DIR'} . '/' . $self->{'CONFIG'}{'POS_IDLIST_NAME'});

    $self->{'SYNC_DIR'}       = trimpath($self->{'CONFIG'}{'POS_LOCAL_SYNC_DIRECTORY'});
    $self->{'SYNC_IMAGE_DIR'} = trimpath($self->{'SYNC_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_IMAGE_BASE'});
    $self->{'SYNC_BOOT_DIR'}  = trimpath($self->{'SYNC_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_BOOT_BASE'});

    $self->{'ROLLBACK_PXEPREFIX'} = 'pxe';    #TODO: put under ROLLBACK subhash
    $self->{'ROLLBACK_CONFIGSDIR'} =
      trimpath($self->{'ROLLBACK_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CR_ROLLBACK_CFG'})
      ;                                       #/srv/tftpboot/KIWI/rollback /configs';
    $self->{'ROLLBACK_BOOTDIR'} =
      trimpath($self->{'ROLLBACK_DIR'} . "/" . $self->{'CONFIG'}{'TFTP_CR_ROLLBACK_BOOT'})
      ;                                       #/srv/tftpboot/KIWI/rollback /boot';

    #scConfigFileSyncTemplate commands
    my $himself->{'STATIC'}{'CONFIG'}{'ALL_LDAPHOST'} = $self->{'CONFIG'}{'ALL_LDAPHOST'};
    my $adminserver = POS::SyncImages::adminserver($himself);    #TODO: put in some other common module?
    $self->{'POS_REMOTE_SYNC_CONFIG_COMMANDS'} =
      'rsync -axzv rsync://' .
      $adminserver . '/config  ' . $self->{'CONFIG'}{'POS_LOCAL_SYNC_DIRECTORY'} . '/config --delete';


    #define ldapbase
    $self->{'LDAPBASE'} = $self->{'CONFIG'}{'ALL_LDAPBASE'} unless defined $self->{'LDAPBASE'};
    if (!defined $self->{'LDAPBASE'}) {
        $self->{'ERROR'}{'CODE'}    = -2;
        $self->{'ERROR'}{'MESSAGE'} = 'There is no ALL_LDAPBASE defined in CONFIG, exit now';
        return undef;
    }

    #define branchbase
    $self->{'BRANCHBASE'} = $self->{'CONFIG'}{'BRANCH_LDAPBASE'} unless defined $self->{'BRANCHBASE'};
    if (!defined $self->{'BRANCHBASE'} && $self->{'MODE'} ne 'admin') {
        $self->{'ERROR'}{'CODE'}    = -3;
        $self->{'ERROR'}{'MESSAGE'} = 'There is no BRANCH_LDAPBASE defined in CONFIG, or given on commandline';
        return undef;
    }


    if ($self->{'MODE'} eq 'admin' || $self->{'MODE'} eq 'branchadmin') {
        $self->{'LDAP_ACCOUNT'}  = $self->{'CONFIG'}{'POS_ADMIN_DN'};
        $self->{'LDAP_PASSWORD'} = $self->{'CONFIG'}{'POS_ADMIN_PASSWORD'};
        #these values are read from branch or admin config, according to mode
    }
    elsif ($self->{'MODE'} eq 'branchservice') {
        $self->{'LDAP_ACCOUNT'}  = $self->{'CONFIG'}{'LEASES2LDAP_ADMIN'};
        $self->{'LDAP_PASSWORD'} = $self->{'CONFIG'}{'LEASES2LDAP_ADMIN_PW'};
    }


    #assume some sensible default values if not present
    if (!defined $self->{'CONFIG'}{'MAX_ARCHIVE_SIZE'}) { $self->{'CONFIG'}{'MAX_ARCHIVE_SIZE'} = 1000000 }
    if (!defined $self->{'CONFIG'}{'MAX_TARGZ_FILES'})  { $self->{'CONFIG'}{'MAX_TARGZ_FILES'}  = 20 }


#whitelist setup for key=values pairs in uploaded and api files, coarse grained filter to remove potentialy dangerous content
    my %hwtypelist;
    $hwtypelist{'IPADDR'}       = '^(\d{1,3}\.){3}\d{1,3}$';  #simple sanity IPv4 regexp, more detailed check done later
    $hwtypelist{'HWTYPE'}       = '^\V*$';                    #anything sensible - no vertical whitespace, can be empty
    $hwtypelist{'HWBIOS'}       = '^\V*$';
    $hwtypelist{'POS_ID'}       = '^\V*$';
    $hwtypelist{'POS_ROLE'}     = '^\V*$';
    $hwtypelist{'POS_MAC'}      = '^[0-9a-fA-F,:]+$';
    $hwtypelist{'NCNAME'}       = '^\V*$';
    $hwtypelist{'CRNAME'}       = '^\V*$';
    $hwtypelist{'ARCHITECTURE'} = '^\V*$';
    $self->{'WHITELIST'}{'HWTYPE'} = \%hwtypelist;

    my %bootvlist;
    $bootvlist{'IPADDR'}                = '^(\d{1,3}\.){3}\d{1,3}$';
    $bootvlist{'HWBIOS'}                = '^\V*$';
    $bootvlist{'POS_CFG_HASH'}          = '^[a-f0-9]{32}$';
    $bootvlist{'POS_ROLLBACK'}          = '^\w+$';
    $bootvlist{'POS_MAC'}               = '^[0-9a-fA-F,:]+$';
    $bootvlist{'HWTYPE'}                = '^\V*$';                    #anything sensible - no vertical whitespace, can be empty
    $self->{'WHITELIST'}{'BOOTVERISON'} = \%bootvlist;


} ## end sub _init_config_data


sub _init_mac_proc
{
    (my $self) = @_;
    #temp before posDnsDhscp calling of MAC.pm is changed
    $self->{'MACCFG'} = POS::MAC->new('CONFIG' => $self->{'CONFIG'});
} ## end sub _init_mac_proc


####################################################
# methods


#translate HWTYPE string if ugly
sub hw_to_outhw
{
    my ($self, $hwtype) = @_;
    my $outhwtype = $hwtype;
    $outhwtype =~ s/([^-_A-Za-z0-9])/sprintf("%%%02X",ord($1))/egs;

    return $outhwtype;
} ## end sub hw_to_outhw


#TODO: put into separate module (filetools?) (same code as in posDataCache)
#from posCRConfig.pm
# trims away excessive '/' and """
sub trimpath
{
    my ($self, $path) = @_;
    $path = $self if (!defined $path);
    $path =~ s@"@@g;    #"# fix syntax hilight
    $path =~ s@/+@/@g;
    $path =~ s@/$@@g;
    return $path;
} ## end sub trimpath


# reads and resets error of this object
# errdata hash naming must be compatible
# with names used in posLeaseMonitor.pm
# (this object is common for all workstations,
#  specific error data should be removed before
#  another workstation is checked/registered)
sub get_error
{
    (my $self) = @_;
    my %errdata;
    %errdata                    = %{ $self->{'ERROR'} };
    $self->{'ERROR'}{'CODE'}    = 0;
    $self->{'ERROR'}{'MESSAGE'} = '';
    return \%errdata;
} ## end sub get_error


#TODO: put into another module? (pos file tools...)
#originaly from posCRConfig.pm
# $keeporiginal: 1 = keep orig. file, 0 or undef = delete orig. file
# $filefilter = grep filter to find all uncompressed files kept outside. undef -> ignore limit
# $maxextfiles = maximun number of plain files outside of archive (useful to limit e.g.
#    hwtype.MAC.hash). no check if undef
sub backup
{
    my ($self, $bkfilepathname, $apath, $archbarename, $keeporiginal, $filefilter, $maxextfiles) = @_;

    sub timestamp
    {
        my $timestring;
        my @t = localtime(time);
        my $y = sprintf("%04d", 1900 + $t[5]);
        my $m = sprintf("%02d", 1 + $t[4]);
        my $d = sprintf("%02d", $t[3]);
        my $h = sprintf("%02d", $t[2]);
        my $n = sprintf("%02d", $t[1]);
        my $s = sprintf("%02d", $t[0]);
        $timestring = "$y-$m-$d\_$h;$n;$s";
        return $timestring;
    } ## end sub timestamp

    sub get_oldest_file
    {    # assuming files in array exists
        my ($files_ar) = @_;
        my $ltime;
        my $lindex = 0;
        for (my $i = 0; $i < @{$files_ar}; $i++) {
            my $ftime = (stat($files_ar->[$i]))[9];
            if (!defined $ltime || $ftime < $ltime) {
                $lindex = $i;
                $ltime  = $ftime;
            }
        }
        return $lindex;
    } ## end sub get_oldest_file
        #removes files so there is at most $maxfiles of them

    sub get_remove_cmds
    {    # assuming files in array exists
        my ($files_ar, $maxfiles) = @_;
        my $i = 0;
        my @rcmds;
        my @rfiles = @{$files_ar};    #copy array
        return @rcmds if (!defined $maxfiles);
        $maxfiles = 0 if $maxfiles < 0;
        while (@rfiles > $maxfiles) {
            my $rindex        = get_oldest_file(\@rfiles);
            my $ofilepathname = $rfiles[$rindex];
            splice(@rfiles, $rindex, 1);
            @rcmds = (@rcmds, ['unlink', $ofilepathname]) if defined $ofilepathname;
            $i++;
        }
        return (@rcmds);
    } ## end sub get_remove_cmds

    my @cmds = ();
    my $cmd;
    my $maxsize             = $self->{'CONFIG'}{'MAX_ARCHIVE_SIZE'};
    my $maxtgzfiles         = $self->{'CONFIG'}{'MAX_TARGZ_FILES'};
    my @bkfilepathnameparts = split('/', $bkfilepathname);
    my $bkfile              = pop(@bkfilepathnameparts);
    my $bkpath              = join('/', @bkfilepathnameparts);
    my $archname            = $archbarename . ".tar";
    my $timestr             = timestamp();
    my $archsize            = (stat("$apath/$archname"))[7];
    my $bkfilepath          = dirname($bkfilepathname);
    my $tftpbase            = $self->{'CONFIG'}{'TFTP_BASE'};
    my $cpmv                = 'move';
    $cpmv = 'copy' if $keeporiginal;

    if ($bkfilepath !~ /^$tftpbase/) {
        log_alert_print
"Backup path \'$bkfilepath\' does not contain \'$tftpbase\', aborting backup to prevent data loss (undefined/illegal path?)";
        return (@cmds);    #empty now
    }

    if (!-e $apath) {      #can disable backup by deleting backup directory, not so serious error
        log_warn_print "Backup path \'$apath\' does not exist, cannot backup \'$bkfilepathname\'";
        return (@cmds);    #empty now
    }

    opendir(ARCHDIR, $apath);
    my @afiles = readdir(ARCHDIR);
    closedir(ARCHDIR);
    my @efiles;
    @efiles = map("$apath/$_", grep(/$filefilter/, @afiles)) if defined $filefilter;
    my @tgzfiles = map("$apath/$_", grep(/^$archbarename.*\.tar\.gz/, @afiles));

    if (!defined $archsize) { $archsize = 0 }
    if ($archsize > $maxsize) {
        if ($maxtgzfiles > 0) {    #do not bother create tar.gz if no allowed
                #@cmds = (@cmds, "gzip -c \"$apath/$archname\" > \"$apath/$archbarename.$timestr.tar.gz\"");
            @cmds = (@cmds, ['external', 'gzip', "$apath/$archname"]);
            @cmds = (@cmds, ['move', "$apath/$archname.gz", "$apath/$archbarename.$timestr.tar.gz"]);
            $maxtgzfiles--;    #compensation for newly created one
        }
        @cmds = (@cmds, ['unlink', "$apath/$archname"]);
    }
    @cmds = (@cmds, [$cpmv, $bkfilepathname, "$apath/$bkfile.$timestr"]);
    @cmds = (@cmds, ['external', 'tar', 'uC', $apath, '-f', "$apath/$archname", "$bkfile.$timestr"]);
    if (defined $filefilter && defined $maxextfiles && $maxextfiles == 0) {
        @cmds = (@cmds, ['unlink', "$apath/$bkfile.$timestr"]);
    }
    else {                     #remove if no allowed
        @cmds = (@cmds, ['move', "$apath/$bkfile.$timestr", "$apath/$bkfile"]);
        $maxextfiles-- if defined $maxextfiles;    #compensation for newly created one
    }

    #delete excess plain files, put those command in the beginning to save diskspace
    @cmds = (get_remove_cmds(\@efiles, $maxextfiles), @cmds);
    #delete excess gz files
    @cmds = (get_remove_cmds(\@tgzfiles, $maxtgzfiles), @cmds);

    return (@cmds);


} ## end sub backup


#only instance should be here
## @method void runcmds($cmds_ref)
# Runs supplied system commands
# @par Details:
# A procedure to consecutively run a list of system commands. Commands and their results are prined into log with 'info' priority.
# warning: In the end, all commands are removed from the input array
# @param $cmds_ref reference to an array of system commands references to be executed.
# System command reference is a reference to a list with the first element being the command and the rest being the parameters
# @return 1
sub runcmds
{
    my ($self, $cmds_ref) = @_;

    my $cmd_ar = shift @{$cmds_ref};
    while (defined $cmd_ar) {
        my $res;
        my $cmdstr = join(',', @{$cmd_ar});
        log_info_print "executing: $cmdstr";
        my $cmd = shift @{$cmd_ar};
        if ($cmd eq 'symlink') {    #process perl builtin
            unlink $cmd_ar->[1];    #otherwise it will not update links on directories
            $res = !symlink $cmd_ar->[0], $cmd_ar->[1];
        }
        elsif ($cmd eq 'unlink') {    #only one file
            $res = !unlink $cmd_ar->[0];
        }
        elsif ($cmd eq 'unlinkf') {    #only one file, force
            unlink $cmd_ar->[0];
        }
        elsif ($cmd eq 'chmod') {      #only one file
            $res = !chmod $cmd_ar->[0], $cmd_ar->[1];
        }
        elsif ($cmd eq 'mkpath') {
            mkpath $cmd_ar->[0];       #no error even if existing
        }
        elsif ($cmd eq 'move') {
            $res = !move $cmd_ar->[0], $cmd_ar->[1];
        }
        elsif ($cmd eq 'copy') {
            $res = !copy $cmd_ar->[0], $cmd_ar->[1];
        }
        elsif ($cmd eq 'external') {
            $res = system(@{$cmd_ar});
        }
        else {
            log_error_printf('Unknown command: \'%s\' ', $cmdstr);
        }
        log_info_print "   result: $res" if $res;    #TODO: test
        $cmd_ar = shift @{$cmds_ref};
    }
} ## end sub runcmds


#safely remove directory and its content
# creates command to remove recursively all content, with some
# checks to prevent failures in case of bad input data
sub rmdir_safe
{
    my ($self, $dir, $regexp_ar) = @_;
    #my $cmd = "rm -rf --preserve-root  \"$dir\"";
    my $cmd_ar = ['external', 'rm', '-rf', '--preserve-root', $dir];
    foreach my $regexp (@{$regexp_ar}) {
        if ($dir !~ /$regexp/) {
            my $cmdstr = join(',', @{$cmd_ar});
            log_warn_print
"directory $dir does not conform to  \'$regexp\' regexp, aborting recursive delete to prevent unexpected behaviour.";
            #$cmd = "echo \"ignored command : $cmd\"";
            $cmd_ar = ['external', 'echo', "ignored command : $cmdstr"];
        }
    }

    return $cmd_ar;
} ## end sub rmdir_safe


sub update
{
} ## end sub update

