#!/usr/bin/perl -w
## @file
# Implementation of POS::posDns


## @class
# Generates the DNS configuration and zonedata from LDAP.
# @par
# @note @em non-class perl module
package POS::posDns;

use strict;

# version  of the package
our $VERSION = 0.1;

use Exporter;

our @ISA = qw( Exporter );

use POS::LDAP;
use Net::DNS;
use Net::IP;
use Net::IPv4Addr qw( :all);
use Socket;
use File::Path;
use File::Copy;
use File::Compare;
use IO::File;
use POSIX qw(tmpnam floor ceil);
use Time::localtime;
use URI::Escape;
use POS::Config;
use POS::Syslog;
use POS::Pid;
use POS::IP qw(:DEFAULT is_in_range);
use POS::LdapUtil;
use strict;

## @cmethod object new()
# @return A new object
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ## config data
        'CACHE'  => undef,    ## cached ldap data
    };

    %$self = (%$self, @_);
    $self->{'Zone_Conf'}         = {};
    $self->{'Changed_Zones'}     = [];
    $self->{'Rev_IP_Hash'}       = {};
    $self->{'Modification_Time'} = '20000101010000';

    return if !defined $self->{'STATIC'};
    return if !defined $self->{'CACHE'};

    return bless($self);
} ## end sub new


sub gen_location_zones($)
{
    my $self              = shift;
    my $modification_time = $self->{'Modification_Time'};

    my $base_list = get_location_baselist($self);

    #TODO: use cache here
    foreach my $one_base (keys %{$base_list}) {
        if ($base_list->{$one_base}{'MODIFIED'} == 0) {
            next;
        }
        # search the needed entries in each base
        my $ldap_result = $self->{'CACHE'}{'LDAP'}->search(
            base  => $one_base,
            scope => 'sub',
            filter =>
'(|(objectclass=scLocation)(objectClass=scWorkstation)(objectClass=scPrinter)(objectClass=scBranchServer)(objectClass=scService)(objectClass=scServerContainer)(objectClass=scNetworkcard))',

            attrs => [
                'scDhcpRange', 'scDhcpFixedRange', 'ipNetworkNumber', 'ipNetmaskNumber',
                'scDnsName',   'scServiceStatus',  'ipHostNumber',    'cn',
                'objectClass', 'associatedDomain', 'macAddress'
            ]
        );
        $ldap_result->code && die "gen_location_zones: $ldap_result->error";

        # generate Zone for each base
        gen_zone($self, $ldap_result);
        # generate reverse zone for each base
        gen_location_rev_zone($self, $ldap_result, $base_list->{$one_base}{'CONFIG'});
    }
} ## end sub gen_location_zones($)


sub read_zone_template($$)
{
    my ($self, $template_filename) = @_;


    my @template_lines = ();

    if (!open(FH, '<', $template_filename)) {
        log_alert_printf('Can\'t read %s, aborting', $template_filename);
        die("Can't read $template_filename");
    }
    else {
        @template_lines = <FH>;
        close(FH);
    }
    return @template_lines;
} ## end sub read_zone_template($$)


sub incr_serial($)
{
    my ($old_serial) = @_;

    my $tm = localtime;
    my $new_serial_number;
    my $new_serial_prefix = sprintf("%04d%02d%02d", $tm->year + 1900, $tm->mon + 1, $tm->mday);

    if ($new_serial_prefix eq substr($old_serial, 0, 8)) {
        #$new_serial_number = sprintf("%s%02d", $new_serial_prefix, $substr($old_serial,8,2)+1);
        my $increm = substr($old_serial, 8, 2) + 1;
        $increm = 99 if ($increm > 99);
        $new_serial_number = sprintf("%s%02d", $new_serial_prefix, $increm);
    }
    else {
        $new_serial_number = $new_serial_prefix . "00";
    }

    return $new_serial_number;
} ## end sub incr_serial($)

# this is currently unused, get_serial() is used instead
sub get_old_serial($)
{
    my ($zone_name) = @_;

    my $serial = "0000000000";

    my $res = new Net::DNS::Resolver;
    my $query = $res->query("$zone_name", "SOA");
    if ($query) {
        $serial = ($query->answer)[0]->serial;
    }
    return $serial;
} ## end sub get_old_serial($)

sub get_serial($)
{
    my ($zone_array_ref) = @_;

    my $zone = "";
    my $line;

    # find the line containing the serial number
    # normally it is the same line containing SOA or the next line
    my $line_number = 0;
    foreach $line (@$zone_array_ref) {
        last if $line =~ /SOA/;
        $line_number++;
    }
    # check if the SOA line was really found
    if ($zone_array_ref->[$line_number] =~ /SOA/) {

        # for now I assume the serial number is in the next line behind
        # the SOA line
        if ($zone_array_ref->[$line_number + 1] =~ /(\d{10})/) {
            return $1;
        }
    }
    return "0000000000";
} ## end sub get_serial($)


sub replace_serial($$)
{
    my ($zone_array_ref, $new_serial) = @_;

    my $zone = "";
    my $line;

    # find the line containing the serial number
    # normally it is the same line containing SOA or the next line
    my $line_number = 0;
    foreach $line (@$zone_array_ref) {
        last if $line =~ /SOA/;
        $line_number++;
    }
    # check if the SOA line was really found
    if ($zone_array_ref->[$line_number] =~ /SOA/) {

        # for now I assume the serial number is in the next line behind
        # the SOA line
        $zone_array_ref->[$line_number + 1] =~ s/\d{10}/$new_serial/gs;
    }
} ## end sub replace_serial($$)


sub gen_zone($$)
{
    my ($self, $ldap_result) = @_;
    my $Zone_Template_File = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONETEMPLATE'};

    # if no entries were found, do nothing
    if ($ldap_result->count == 0) {
        return 1;
    }

    my @ldap_entries = $ldap_result->entries;

    # get the zonename by analyzing the dn of the first entry
    my $zone_name = get_zonename($ldap_entries[0]);
    #print "zone name: $zone_name\n";

    # read zone-header file
    my @new_zone_lines = read_zone_template($self, $Zone_Template_File);

    # write informational comment into array
    push(@new_zone_lines, "; Zone $zone_name generated by ldap2dns.pl\n");

    # build GENERATE line for scDhcpRange if a cn=standards entry exists
    push(@new_zone_lines, gen_dhcp_range_rec($ldap_result));

    # build records for each scWorkstation entry
    push(@new_zone_lines, gen_workstation_recs($ldap_result));

    # generate record lines for scServer and scService entries
    push(@new_zone_lines, gen_srv_records($ldap_result, $zone_name));

    # build records for each scPrinter entry
    push(@new_zone_lines, gen_printer_recs($ldap_result));

    # compare new and old zone
    my $zonefile_name = $zone_name;
    if (zone_changed($self, \@new_zone_lines, $zone_name)) {
        # create backup copy of zone file
        backup_zone_file($self, $zonefile_name, $zone_name);
    }

    # write new zone file
    write_zone_file($self, $zonefile_name, $zone_name, \@new_zone_lines);

    return 1;
} ## end sub gen_zone($$)

sub get_scNetworkCard( $$$ )
{
    my ($ldap_entries_ar, $server_dn, $network) = @_;
    my $entry = undef;
    my $dn;
    foreach my $ldap_entry (@$ldap_entries_ar) {
        $dn = $ldap_entry->dn;
        if ($dn and
            ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scNetworkcard') and
            ($dn =~ /.*,$server_dn/) and
            ipv4_in_network($network, $ldap_entry->get_value('ipHostNumber'))) {
            $entry = $ldap_entry;
            last;
        }
    }
    return $entry;
} ## end sub get_scNetworkCard( $$$ )

sub gen_srv_records($$)
{
    my ($ldap_result, $zone_name) = @_;

    my @ldap_entries = $ldap_result->entries;

    my %servers;
    my @dns_records = ();

    # Fix to generate CNAME records correctly.
    # Currently, the reverse lookup supports the ability to specify a
    # canonical entry.  The canonical written into the reverse lookup
    # record in order to give the correct answer on reverse lookup.
    # However, this code must be modified to create an A record rather
    # than a CNAME record for the canonical entry.

    # First, go through the list of services to see if there is a canonical
    # entry. If so, keep its dns and process it with the scLocation.

    # how to generate CNAME records?
    # - create hash based on the scServer entries,
    #    key: ipHostNumber, value: scDnsName
    # - make A records from the hash
    # - for each scService entry:
    #    -if canonical dns generate A record for service, generate CNAME
    #       record for scLocation
    #    -else if not canonical
    #    - if ipHostNumber exists in hash, generate CNAME record,
    #                    otherwise generate A record.

    my $locationDns     = undef;
    my $network         = "";
    my $network_address = get_ldap_value($ldap_result, 'scLocation', 'ipNetworkNumber');
    my $netmask         = get_ldap_value($ldap_result, 'scLocation', 'ipNetmaskNumber');
    if ($network_address && $netmask) {
        $network = $network_address . "/" . $netmask;
    }


    # determine whether there is a canonical dns on one of the services
    foreach my $ldap_entry (@ldap_entries) {
        my $service_status = $ldap_entry->get_value('scServiceStatus');
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scService') and
            defined($service_status) and
            $service_status eq 'TRUE') {
            my $ip_addr = $ldap_entry->get_value('ipHostNumber');
            # scDnsName is a multivalue attribute
            my @dns_names = $ldap_entry->get_value('scDnsName');
            # first dns_name may be CNAME or A record
            my $dns_name = shift(@dns_names);
            $dns_name =~ s/(\;.*)$//;    # multi-value field with semicolon
            my $canonical;
            if ($1) {
                $canonical = ($1 =~ /^\;(c|canon|canonic|canonical)$/i);
            }
            if ($canonical) {
                $locationDns = $dns_name;
            }
        }
    }

    # at this point, if locationDns is undefined, there are not
    # services with a canonical entry. We can now deal with the scLocation
    # and scWorkstation objects

    foreach my $ldap_entry (@ldap_entries) {
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scBranchServer')) {
            my $nc_entry = get_scNetworkCard(\@ldap_entries, $ldap_entry->dn(), $network);
            my $dns_name = $ldap_entry->get_value('cn');
            if ($nc_entry and defined($nc_entry->get_value('ipHostNumber'))) {
                my $ip_addr = $nc_entry->get_value('ipHostNumber');
                if (defined($locationDns)) {
                    # generate A record for canonical service and
                    # generate CNAME record for scLocation
                    $servers{$ip_addr} = $locationDns;

                    # generate A record for canonical service
                    push(@dns_records, sprintf("%s\tIN\tA\t%s\n", $locationDns, $ip_addr));

                    # generate CNAME record for scLocation
                    push(@dns_records, sprintf("%s\tCNAME\t%s.%s.\n", $dns_name, $servers{$ip_addr}, $zone_name));
                }
                else {
                    # generate A record
                    $servers{$ip_addr} = $dns_name;

                    push(@dns_records, sprintf("%s\tIN\tA\t%s\n", $dns_name, $ip_addr));
                }
            }
        }
        # for generating CNAME records for caching servers, scWorkstation
        # entries must be considered, if a scLocation entry does exist
        # in the ldap result. So we just generate hash entries for them
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scWorkstation')) {
            ## getting directly
            #my $ip_addr = ldap_cn2addr($ldap_entry->get_value('cn'),
            #              $network_address);
            my $ip_addr  = $ldap_entry->get_value('ipHostNumber');
            my $dns_name = $ldap_entry->get_value('cn');
            my $mac      = $ldap_entry->get_value('macAddress');
            #print "$dns_name \n";
            $servers{$ip_addr} = $dns_name if $mac;
        }
    }

    # now for each service create A record or CNAME record.
    foreach my $ldap_entry (@ldap_entries) {
        my $service_status = $ldap_entry->get_value('scServiceStatus');
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scService') and
            defined($service_status) and
            $service_status eq 'TRUE') {
            my $ip_addr = $ldap_entry->get_value('ipHostNumber');
            if (not defined $ip_addr) {
                my ($dn) = $ldap_entry->dn() =~ /^[^,]*,(.*)/;
                $ip_addr = get_scNetworkCard(\@ldap_entries, $dn, $network)->get_value('ipHostNumber');
            }
            # scDnsName is a multivalue attribute
            my @dns_names = $ldap_entry->get_value('scDnsName');
            # first dns_name may be CNAME or A record
            my $dns_name = shift(@dns_names);
            $dns_name =~ s/(\;.*)$//;    # multi-value field with semicolon
            if (!defined($locationDns) or $dns_name ne $locationDns) {
                # only write the entry if it hasn't already been written
                if (exists($servers{$ip_addr})) {
                    # generate CNAME record
                    push(@dns_records, sprintf("%s\tCNAME\t%s.%s.\n", $dns_name, $servers{$ip_addr}, $zone_name));
                }
                else {
                    # generate A record
                    push(@dns_records, sprintf("%s\tIN\tA\t%s\n", $dns_name, $ip_addr));
                }

            }
            # remaining dns names must always be CNAME records
            # if @dns_names is empty here (only on value), nothing is done
            foreach $dns_name (@dns_names) {
                push(@dns_records, sprintf("%s\tCNAME\t%s.%s.\n", $dns_name, $servers{$ip_addr}, $zone_name));
            }

        }
    }
    return @dns_records;
} ## end sub gen_srv_records($$)


sub still_in_range
{
    my $from    = shift;
    my $to      = shift;
    my $current = shift;

    my @fromIP    = split(/\./, $from);
    my @currentIP = split(/\./, $current);
    my @toIP      = split(/\./, $to);

    ## Need to make sure that it is equal or above the
    ## low range and equal or below the high range
    my $cmplow  = 0;
    my $cmphigh = 0;

    my $i = 0;
    for ($i = 0; $i < 4; $i++) {
        if ($cmplow == 0) {
            if ($fromIP[$i] > $currentIP[$i]) {
                $cmplow = -1;
            }
            elsif ($fromIP[$i] < $currentIP[$i]) {
                $cmplow = 1;
            }
        }
        if ($cmphigh == 0) {
            if ($toIP[$i] > $currentIP[$i]) {
                $cmphigh = 1;
            }
            elsif ($toIP[$i] < $currentIP[$i]) {
                $cmphigh = -1;
            }
        }

        if ($cmplow != 0 && $cmphigh != 0) {
            last;
        }
    }

    if ($cmplow < 0 || $cmphigh < 0) {
        return 0;
    }

    return 1;
} ## end sub still_in_range

sub get_dhcp_range_pairs($)
{
    # This routine needs to split an ip range
    # into the smaller pieces in order to
    # handle the components separately
    my $ip_range = shift;

    # find where they differ
    # remove all whitespace
    $ip_range =~ s/\s//g;

    # find where they differ
    # remove all whitespace
    $ip_range =~ s/\s//g;

    my @ip_range = split(/,/, $ip_range);

    my @from = split(/\./, $ip_range[0]);
    my @to   = split(/\./, $ip_range[1]);

    my @range_list;
    my $done = 0;

    while (!$done) {
        my $range_bottom = join(".", @from);
        my $range_top;

        if ($from[0] == $to[0] &&
            $from[1] == $to[1] &&
            $from[2] == $to[2]) {
            $range_top = join(".", @to);
        }
        else {
            $range_top = join(".", @from[0 .. 2]) . ".255";
        }
        push(@range_list, "$range_bottom,$range_top");

        #readjust from to move up to the next part of the
        #subnet
        $from[3] = 0;
        ++$from[2];
        if ($from[2] == 256) {
            $from[2] = 0;
            ++$from[1];
            if ($from[1] == 256) {
                $from[1] = 0;
                ++$from[0];
            }
        }

        my $current = join(".", @from);
        if (!still_in_range($ip_range[0], $ip_range[1], $current)) {
            $done = 1;
        }
    }

    return @range_list;
} ## end sub get_dhcp_range_pairs($)

sub gen_dhcp_range_rec($$)
{
    my ($ldap_result, $rev_flag) = @_;
    my $objectclass;
    my @zone_line = ();
    my $ip_prefix;
    #my @ip1;
    #my @ip2;
    my @num_range;
    my $ldap_entry;

    foreach my $entry ($ldap_result->entries) {
        # there should be only one such entry here
        # if one is found, it will be used.
        # sanity checks for LDAP Data are another project
        #last if ( ($objectclass = $ldap_entry->get_value('objectClass'))
        #                                       eq 'scLocation');
        if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => 'scLocation')) {
            $ldap_entry  = $entry;
            $objectclass = 'scLocation';
            last;
        }
    }

    if (defined $objectclass) {
        if ($objectclass eq 'scLocation') {
            my $ip_range = $ldap_entry->get_value('scDhcpRange');


            my @ip_range_pairs = get_dhcp_range_pairs($ip_range);

            foreach my $ip_inner_range (@ip_range_pairs) {
                $ip_prefix = analyze_dhcp_range($ip_inner_range);
                @num_range = analyze_dhcp_range($ip_inner_range);

                if (!defined $ip_prefix) {
                    log_warn_printf('first octets in scDhcpRange %s of %s ' . 'different',
                        $ldap_entry->dn, $ip_inner_range);
                }
                elsif (defined $rev_flag) {
                    my $zone_name = get_zonename($ldap_entry);
                    push(@zone_line,
                        '$GENERATE ' . $num_range[0] . "-" . $num_range[1] . "\t" . '$' . "\tPTR\t" . 'dhcp$' . "." .
                          $zone_name . ".\n");
                }
                else {
                    push(@zone_line,
                        '$GENERATE ' . $num_range[0] . "-" . $num_range[1] . "\t" . 'dhcp$' . "\t\tA\t" . $ip_prefix .
                          '.$' . "\n");

                }
            }
        }
    }
    return @zone_line;
} ## end sub gen_dhcp_range_rec($$)


sub gen_workstation_recs($$)
{
    my ($ldap_result) = @_;
    my @zone_lines = ();
    my $network_address;

    # get network address from scLocation entry
    $network_address = get_ldap_value($ldap_result, 'scLocation', 'ipNetworkNumber');

    #print "gen_workstation_recs\n";
    #if ( defined $network_address )
    #{
    foreach my $entry ($ldap_result->entries) {
        if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => 'scWorkstation')) {
            # calculate IP-Adress from cn and network address
            #my $ws_ip_addr = ldap_cn2addr($entry->get_value('cn'),
            #                       $network_address);
            my $ws_ip_addr = $entry->get_value('ipHostNumber');
            my $mac        = $entry->get_value('macAddress');
            #print "$ws_ip_addr\n";
            push(@zone_lines, $entry->get_value('cn') . "\t\tIN\tA\t" . $ws_ip_addr . "\n") if $mac;
        }
    }
    #}
    return @zone_lines;
} ## end sub gen_workstation_recs($$)


sub gen_printer_recs($)
{
    my ($ldap_result) = @_;
    my @zone_lines = ();
    my $network_address;

    # get network address from scLocation entry
    $network_address = get_network_addr($ldap_result);

    if (defined $network_address) {
        foreach my $entry ($ldap_result->entries) {
            if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => 'scPrinter')) {
                # calculate IP-Adress from cn and network address
                my $pr_cn = $entry->get_value('cn');
                my $pr_ip_addr = ldap_cn2addr($pr_cn, $network_address);
                push(@zone_lines, $pr_cn . "\t\tIN\tA\t" . $pr_ip_addr . "\n");
                # generate CNAME if scDnsName is set in the entry
                my $pr_cname = $entry->get_value('scDnsName');
                if (defined $pr_cname) {
                    push(@zone_lines, $pr_cname . "\tCNAME\t\t" . $pr_cn . "\n");
                }
            }
        }
    }
    return @zone_lines;
} ## end sub gen_printer_recs($)


sub get_network_addr($)
{
    my ($ldap_entry) = @_;

    my $network_address;
    $network_address = get_ldap_value($ldap_entry, 'scLocation', 'ipNetworkNumber');
    return $network_address;
} ## end sub get_network_addr($)


sub range2rev_network_addr($$)
{
    my ($self, $dhcp_range) = @_;

    my $network_address;
    my $ip_prefix;

    if (defined $dhcp_range) {
        $ip_prefix = analyze_dhcp_range($dhcp_range);
        if (defined $ip_prefix) {
            $network_address = $ip_prefix . ".0";
        }
    }

    return $network_address;
} ## end sub range2rev_network_addr($$)


###############?????????????????????#############
sub range2rev_addr($$)
{
    my ($dhcp_range, $network_addr) = @_;
    return $network_addr;
} ## end sub range2rev_addr($$)


sub write_zone_file($$$$)
{

    my ($self, $zonefilename, $zone_name, $zone_lines_ref) = @_;
    my $Zonefile_Path = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONEFILEPATH'};
    my $Zonefile_Conf = $self->{'Zonefile_Conf'};
    my $Zone_Conf     = $self->{'Zone_Conf'};
    # assemble path name
    my $base_zonename = lc(ldap_dn2dnsname($self->{'STATIC'}{'CONFIG'}{'ALL_LDAPBASE'}));
    my $path_name = $Zonefile_Path . "/" . zone_name2path($zone_name, $base_zonename);

    if (!check_dir($path_name)) {
        log_alert_printf('Can\'t create directory %s', $path_name);
        return 0;
    }

    my $output_filename = $path_name . "/" . $zonefilename;

    if (!open(OUTFILE, '>', $output_filename)) {
        log_alert_printf('Can\'t open %s for writing', $output_filename);
        return 0;
    }

    foreach my $line (@$zone_lines_ref) {
        print OUTFILE $line;
    }

    close(OUTFILE);

    # collect information for generation include config file
    $Zone_Conf->{$zonefilename} = zone_name2path($zone_name, $base_zonename) . "/" . $zonefilename;
    return 1;
} ## end sub write_zone_file($$$$)


sub gen_zone_conf($$$)
{
    my ($self, $zonefilename, $zone_name) = @_;
    my $Zonefile_Conf = $self->{'Zonefile_Conf'};
    my $Zone_Conf     = $self->{'Zone_Conf'};

    # assemble path name
    my $base_zonename = lc(ldap_dn2dnsname($self->{'STATIC'}{'CONFIG'}{'ALL_LDAPBASE'}));

    # collect information for generation include config file
    $Zone_Conf->{$zonefilename} = zone_name2path($zone_name, $base_zonename) . "/" . $zonefilename;
    return 1;
} ## end sub gen_zone_conf($$$)


sub backup_zone_file($$$)
{
    my ($self, $zonefilename, $zone_name) = @_;
    my $Zonefile_Path = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONEFILEPATH'};

    # assemble path name
    my $base_zonename = lc(ldap_dn2dnsname($self->{'STATIC'}{'CONFIG'}{'ALL_LDAPBASE'}));
    my $path_name = $Zonefile_Path . "/" . zone_name2path($zone_name, $base_zonename);

    my $output_filename = $path_name . "/" . $zonefilename;

    # check, if old zone file exists and is readable
    if (-r $output_filename) {
        my $backup_filename = $output_filename . ".bak";
        if (!copy($output_filename, $backup_filename)) {
            log_warn_printf('Can\'t create backup copy of %s', $output_filename);
            return 0;
        }
    }

    return 1;
} ## end sub backup_zone_file($$$)


sub zone_name2path($$)
{
    my ($zone_name, $remove_part) = @_;

    my $zone_path = $zone_name;
    my $new_zone_path;
    my @zone_path_parts;
    my @reverse_zone_path_parts;

    # replace . with \. for use in regex
    $remove_part =~ s/\./\\./g;
    # remove the unwanted part
    $zone_path =~ s/$remove_part//;
    # remove a trailing .
    $zone_path =~ s/\.$//;
    # replace . with /
    $zone_path =~ s#\.#/#g;

    # reverse order
    @zone_path_parts         = split(/\//, $zone_path);
    @reverse_zone_path_parts = reverse(@zone_path_parts);
    $new_zone_path           = join('/', @reverse_zone_path_parts);

    return $new_zone_path;
} ## end sub zone_name2path($$)


#######check what can happen
sub check_dir($)
{
    my ($dir_path) = @_;

    # todo: make this more robust, File::Path::mkpath dies if something
    # goes wrong.
    mkpath($dir_path);

    return 1;
} ## end sub check_dir($)


sub get_ldap_value($$$)
{
    my ($ldap_result, $oc_name, $attr_name) = @_;
    my $attr_content;
    my $objectclass;

    foreach my $entry ($ldap_result->entries) {
        # there should be only one such entry here
        # if one is found, it will be used.
        # sanity checks for LDAP Data are another project
        #my @objectclasses = $entry->get_value('objectClass');
        #($objectclass) = grep { $_ eq $oc_name } @objectclasses;
        #if ( defined $objectclass && $objectclass eq $oc_name )
        if (ldap_entry_has_value(ENTRY => $entry, ATTR_NAME => 'objectClass', VALUE => $oc_name)) {
            $attr_content = $entry->get_value($attr_name);
            last;
        }
    }
    return $attr_content;
} ## end sub get_ldap_value($$$)

sub gen_location_rev_zone($$$)
{
    my ($self, $ldap_result, $net_result) = @_;
    my $Rev_IP_Hash = $self->{'Rev_IP_Hash'};

    # do nothing, if the result is empty
    if ($ldap_result->count == 0) {
        return 1;
    }

    my %ip_hash;
    my $dhcp_range_network_addr;

    my @ldap_entries = $ldap_result->entries;

    # get the zonename by analyzing the dn of the first entry
    #print "get_zone_name\n";
    my $zone_name = get_zonename($ldap_entries[0]);

    # get network address from scLocation entry
    #print "get_network_adr\n";
    #print "netresult: ",$net_result,"\n";
    my $location_network_addr = $net_result->get_value('ipNetworkNumber');
    #print "$location_network_addr\n";

    # create branchserver hash first, before the other hashes are created; the
    # branchserver names should get the primary reverse lookup

    gen_ip_branchserver_hash($Rev_IP_Hash, $ldap_result, $zone_name);

    gen_ip_hash($Rev_IP_Hash, $ldap_result, $zone_name, 'scService',     'ipHostNumber', 'scDnsName');
    gen_ip_hash($Rev_IP_Hash, $ldap_result, $zone_name, 'scWorkstation', 'ipHostNumber', 'cn');

    if (defined $net_result->get_value('scDhcpRange')) {
        #print "defined 1\n";
        if (defined analyze_dhcp_range($net_result->get_value('scDhcpRange'))) {
            #print "defined 2\n";
            gen_ip_hash($Rev_IP_Hash, $ldap_result, $zone_name, 'scLocation', 'scDhcpRange', 'scDhcpRange',
                \&range2rev_addr, \&range2rev_network_addr);
            #print "range2rev_network_addr\n";
            $dhcp_range_network_addr = range2rev_network_addr($self, $net_result->get_value('scDhcpRange'));
            #print "range2rev_network_addr END\n";
        }
    }
} ## end sub gen_location_rev_zone($$$)


## @fn  gen_rev_zones($self)
# Generates reverse zones and writes the zone file if changed.
# @par Details:
# Uses ip_hash2zones() to obtain %rev_zones hash
# which is then written as a new zone file, if it differs from the previous zone file.
# @return void
sub gen_rev_zones($)
{
    my ($self) = @_;
    my $netmask = gen_cidr_from_mask($self->{'CACHE'}{'BRANCH'}{'NETMASK'});    #@MDA get netmask
    my $zonenetmask = ceil($netmask / 8) * 8;
    $zonenetmask = 24 if $zonenetmask > 24;

    my $Rev_IP_Hash   = $self->{'Rev_IP_Hash'};
    my $base_zonename = lc(ldap_dn2dnsname($self->{'STATIC'}{'CONFIG'}{'ALL_LDAPBASE'}));
    my $zone_name     = "rev." . $base_zonename;
    my %rev_zones     = ip_hash2zones($self);
    foreach my $network_addr (keys %rev_zones) {
        #print "gen_rev_zones(): $network_addr\n";
        my $rev_zone_name = get_rev_zone_name($network_addr . "/" . $zonenetmask);

        if (!defined $rev_zone_name) {
            log_error_print("invalid network address: $network_addr / $zonenetmask");
            next;
        }
        if (zone_changed($self, $rev_zones{$network_addr}, $zone_name, $rev_zone_name)) {
            # create backup copy of zone file
            backup_zone_file($self, $rev_zone_name, $zone_name);
            write_zone_file($self, $rev_zone_name, $zone_name, $rev_zones{$network_addr});
        }
        else {
            gen_zone_conf($self, $rev_zone_name, $zone_name);
        }
    }
} ## end sub gen_rev_zones($)

sub gen_cidr_from_mask($)
{
    my $netmask = shift;
    my @a       = split(/\./, $netmask);
    my $cidr    = 0;

    use integer;
    for (my $i = 0; $i < 4; $i++) {
        my $bits = $a[$i];

        if ($bits & 0x80) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x40) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x20) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x10) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x08) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x04) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x02) {
            $cidr = $cidr + 1;
        }
        if ($bits & 0x01) {
            $cidr = $cidr + 1;
        }
    }
    return $cidr;
} ## end sub gen_cidr_from_mask($)

sub gen_ip_branchserver_hash( $$$ )
{
    my ($ip_hash_ref, $ldap_result, $zone_name) = @_;
    my $ip_addr;
    my @ldap_entries    = $ldap_result->entries;
    my $network         = "";
    my $network_address = get_ldap_value($ldap_result, 'scLocation', 'ipNetworkNumber');
    my $netmask         = get_ldap_value($ldap_result, 'scLocation', 'ipNetmaskNumber');
    if ($network_address && $netmask) {
        $network = $network_address . "/" . $netmask;
    }
    for my $ldap_entry (@ldap_entries) {
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => 'scBranchServer')) {
            my $nc_entry = get_scNetworkCard(\@ldap_entries, $ldap_entry->dn(), $network);
            if ($nc_entry and ($ip_addr = $nc_entry->get_value('ipHostNumber'))) {
                my $dns_name = $ldap_entry->get_value('cn');
                $dns_name .= "." . $zone_name;
                $ip_hash_ref->{ inet_aton($ip_addr) } = [$dns_name];
            }
        }
    }
} ## end sub gen_ip_branchserver_hash( $$$ )


sub gen_ip_hash($$$$$$$$)
{
    my ($ip_hash_ref,  $ldap_result,       $zone_name,      $objectclass_name,
        $attr_name_ip, $attr_name_dnsname, $attr2ip_subref, $network_addr_subref
    ) = @_;

    my $network_addr;

    if (defined $network_addr_subref) {
        $network_addr = &$network_addr_subref($ldap_result);
    }

    my @ldap_entries = $ldap_result->entries;

    foreach my $ldap_entry (@ldap_entries) {
        #if ( $ldap_entry->get_value('objectClass') eq $objectclass_name )
        if (ldap_entry_has_value(ENTRY => $ldap_entry, ATTR_NAME => 'objectClass', VALUE => $objectclass_name)) {
            my $ip_addr;
            next
              if ($objectclass_name eq 'scWorkstation' && !$ldap_entry->get_value('macAddress'))
              ;    #skip invalidated workstations
            if (defined $attr2ip_subref) {
                $ip_addr = &$attr2ip_subref($ldap_entry->get_value($attr_name_ip), $network_addr);
            }
            else {
                $ip_addr = $ldap_entry->get_value($attr_name_ip);
            }
            if (!$ip_addr) {
                next;
            }

            if (defined($ldap_entry->get_value($attr_name_dnsname))) {
                my $dns_name = $ldap_entry->get_value($attr_name_dnsname);
                $dns_name =~ s/(\;.*)$//;    # multi-value field with semicolon
                my $canonical;
                if ($1) {
                    $canonical = ($1 =~ /^\;(c|canon|canonic|canonical)$/i);
                }
                my $dns_name_fq = $dns_name . "." . $zone_name;

                #print "DNS: $dns_name , ZONE_NAME: $zone_name\n";
                if ((!exists($ip_hash_ref->{ inet_aton($ip_addr) })) or
                    $canonical) {
                    $ip_hash_ref->{ inet_aton($ip_addr) } = [$dns_name_fq];
                }
            }
        }
    }

    return 0;
} ## end sub gen_ip_hash($$$$$$$$)


## @fn % ip_hash2zones($self)
# Generates reverse zones PTR records for $self->{'Rev_IP_Hash'}.
# @par Details:
# Use data from $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONETEMPLATE'} and $self->{'Rev_IP_Hash'} to
# create reverse zones, data are collected into a %rev_zones hash.
# This subroutine relies upon the fake calculation of a network using a 24 bit CIDR netmask.
# This allows for the creation of a zone file for each possible subnet.  However, the
# real network must be calculated in order to check whether to write the entry.  Otherwise,
# all .0 addresses will not be written into the reverse lookup space.
# @return %rev_zones hash filled with reverse zone PTR records.
sub ip_hash2zones($)
{
    my ($self)             = @_;
    my $Zone_Template_File = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONETEMPLATE'};
    my $ip_hash_ref        = $self->{'Rev_IP_Hash'};
    # at first this
    # this must be the actual netmask or the calculated network will be wrong.
    my $realnetmask = gen_cidr_from_mask($self->{'CACHE'}{'BRANCH'}{'NETMASK'});                           #@MDA get netmask
    my $zonenetmask = ceil($realnetmask / 8) * 8;
    $zonenetmask = 24 if $zonenetmask > 24;
    my ($realnetwork) = Net::IPv4Addr::ipv4_network($self->{'CACHE'}{'BRANCH'}{'NETNUM'} . "/" . $realnetmask);

    my %rev_zones;

    my $netmask_diff = $zonenetmask - $realnetmask;
    $netmask_diff = 0 if $netmask_diff < 0;
    my $num_networks = 1 << $netmask_diff;
    
    for (my $i = 0; $i < $num_networks; $i++) {
        my $ip_addr = (new Net::IP ("$realnetwork + " . ($i << (32 - $zonenetmask))))->last_ip();
        my ($zonenetwork) = Net::IPv4Addr::ipv4_network($ip_addr . "/" . $zonenetmask);
        my @new_rev_zone_lines;
        my $rev_zone_name;
        #print "ip_hash2zones(): $ip_addr\n";

        if (!exists($rev_zones{$zonenetwork})) {
            # create new zone lines array

            # get reverse-zone name
            $rev_zone_name = get_rev_zone_name($zonenetwork . "/" . $zonenetmask);
            # read zone-header file
            @new_rev_zone_lines = read_zone_template($self, $Zone_Template_File);
            # write informational comment into array
            push(@new_rev_zone_lines, "; Zone $rev_zone_name generated by ldap2dns.pl\n");
            # store array reference into hash
            $rev_zones{$zonenetwork} = \@new_rev_zone_lines;
        }
    }

    my $last_origin = "";
    foreach my $ip_int (sort(keys %{ $self->{'Rev_IP_Hash'} })) {
        my $ip_addr = inet_ntoa($ip_int);
        my ($zonenetwork) = Net::IPv4Addr::ipv4_network($ip_addr . "/" . $zonenetmask);
        my ($realnetwork) = Net::IPv4Addr::ipv4_network($ip_addr . "/" . $realnetmask);


        # for creating a zone which only has GENERATE entries from
        # a scDhcpRange, there may be a hash key id which is identical
        # to the network. The GENERATE record is created in the function
        # gen_dhcp_range_rec() but the zone header must already be created
        # by the block above.
        if ($realnetwork ne $ip_addr) {
            my $origin = get_rev_zone_name($ip_addr . "/255.255.255.0");
            $origin =~ s/[^.]*\.//;

            if ($origin ne $last_origin) {
                push(@{ $rev_zones{$zonenetwork} }, "\n");
                push(@{ $rev_zones{$zonenetwork} }, '$ORIGIN ' . "$origin.\n");
                $last_origin = $origin;
            }


            # add a PTR record
            # last octet of $ip_addr, this works because the regex matches
            # up to the last dot found
            (my $last_octet = $ip_addr) =~ s/.*\.//;
            push(@{ $rev_zones{$zonenetwork} }, "$last_octet\t\t\tPTR\t" . $ip_hash_ref->{$ip_int}[0] . ".\n");
        }
    }
    return %rev_zones;
} ## end sub ip_hash2zones($)

## @fn $ ip_reverse($network)
# Replaces reverse_ip/ip_reverse in Net::IP.
# @par Details:
# ip_reverse of Net::IP is broken in case of ip containing leading zeros, e.g. 172.22.0.0.
# Here, number of quads remaining is correctly given with respect to mask
# @return in-addr.arpa. reversed ip address name.
# replace reverse_ip/ip_reverse in Net::IP
#TODO: try to get by without Net::IP
sub ip_reverse($)
{
    my ($network) = (@_);
    my $ip = new Net::IP($network);

    return if (not defined $ip or not $ip->is_prefix());

    my $mask_length    = $ip->prefixlen();
    my $prefix         = $ip->ip();
    my @quads          = split /\./, $prefix;
    my $no_quads       = floor($mask_length / 8);
    my @reverse_groups = reverse @quads[0 .. ($no_quads - 1)];
    return join '.', @reverse_groups, 'in-addr', 'arpa.';
} ## end sub ip_reverse($)

## @fn $ get_rev_zone_name($network)
# calls ip_reverse to obtain Replaces reverse_ip/ip_reverse in Net::IP.
# @par Details:
# a wrapper to call ip_reverse(), only removes trailing '.'
# @return in-addr.arpa reversed ip address name.
sub get_rev_zone_name($)
{
    my ($network) = @_;
    my $rev_zone_name;
    my $ip;

    $rev_zone_name = ip_reverse($network);
    if (defined $rev_zone_name) {
        # remove trailing dot yielded by reverse_ip()
        chop($rev_zone_name);
    }
    return $rev_zone_name;
} ## end sub get_rev_zone_name($)


sub gen_config_file($)
{

    my ($self)        = @_;
    my $Zone_Conf     = $self->{'Zone_Conf'};
    my $Zoneconf_Name = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONECONFNAME'};
    my $Zonefile_Path = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONEFILEPATH'};
    if (!open(FH, '>', $Zoneconf_Name)) {
        log_alert_printf('Can\'t open %s for writing', $Zoneconf_Name);
        return 0;
    }

    print FH "# include config file generated by ldap2dns.pl\n\n";
    foreach my $zone_name (sort keys %{$Zone_Conf}) {
        print FH "zone \"$zone_name\" IN {\n";
        print FH "\ttype master;\n";
        print FH "\tfile \"" . $Zonefile_Path . "/" . $Zone_Conf->{$zone_name} . "\";\n";
        #print FH "\tcheck-names fail;\n";
        print FH "\tallow-update { none; };\n";
        print FH "};\n\n";
    }
    close(FH);
    return 1;
} ## end sub gen_config_file($)


sub analyze_dhcp_range($)
{
    my ($ip_range) = @_;

    my $range_prefix;
    my @range;

    if (defined $ip_range) {

        # remove all whitespace
        $ip_range =~ s/\s//g;
        my @ip_range = split(/,/, $ip_range);
        # check if first 3 octets are identical
        my @ip1 = split(/\./, $ip_range[0]);
        my @ip2 = split(/\./, $ip_range[1]);
        my $ip_prefix = join(".", @ip1[0 .. 2]);
        if ($ip_prefix eq join(".", @ip2[0 .. 2])) {
            $range_prefix = $ip_prefix;
            @range = ($ip1[3], $ip2[3]);
        }

    }
    else {
        log_warn_print('analyze_dhcp_range() called with undef');
        show_callstack();
    }

    if (wantarray) {
        return @range;
    }
    else {
        return $range_prefix;
    }
} ## end sub analyze_dhcp_range($)


sub zone_changed($$$$)
{
    my ($self, $zone_lines_ref, $zone_name, $rev_zone_name) = @_;
    my $Zonefile_Path = $self->{'STATIC'}{'CONFIG'}{'POS_LDAP2DNS_ZONEFILEPATH'};
    my $Changed_Zones = $self->{'Changed_Zones'};


    my $base_zonename = lc(ldap_dn2dnsname($self->{'STATIC'}{'CONFIG'}{'ALL_LDAPBASE'}));
    # read old zone file into array
    my $old_zone_file;
    if (defined $rev_zone_name) {
        $old_zone_file = $Zonefile_Path . "/" . zone_name2path($zone_name, $base_zonename) . "/" . $rev_zone_name;
    }
    else {
        $old_zone_file = $Zonefile_Path . "/" . zone_name2path($zone_name, $base_zonename) . "/" . $zone_name;
    }


    if (!-e $old_zone_file) {
        log_info_printf('zone not found, %s will be created', $old_zone_file);
        my $serial = incr_serial('0000000000');
        replace_serial($zone_lines_ref, $serial);
        log_info_print("Generated new serial $serial");
        return 1;
    }
    if (!open(OLD_ZONE, '<', $old_zone_file)) {
        log_warn_printf('Can\'t open %s for zone compare, will try to create new zone file', $old_zone_file);
        my $serial = incr_serial('0000000000');
        replace_serial($zone_lines_ref, $serial);
        log_info_print("Generated new serial $serial");
        return 1;
    }
    my @old_zone_lines = <OLD_ZONE>;
    close(OLD_ZONE);
    my $serial = get_serial(\@old_zone_lines);
    replace_serial($zone_lines_ref, $serial);

    my $zone_changed = !string_array_compare(\@old_zone_lines, $zone_lines_ref);
    if ($zone_changed) {
        log_info_printf('zone changed, updating %s', $old_zone_file);
        $serial = incr_serial($serial);
        replace_serial($zone_lines_ref, $serial);
        log_info_print("Incrementing serial to $serial");
        if (defined $rev_zone_name) {
            push(@{$Changed_Zones}, $rev_zone_name);
        }
        else {
            push(@{$Changed_Zones}, $zone_name);
        }
    }
    else {
        log_info_printf('zone unchanged, keeping %s', $old_zone_file);
    }

    return $zone_changed;
} ## end sub zone_changed($$$$)


sub string_array_compare($$)
{
    my ($arr1_ref, $arr2_ref) = @_;
    if (@$arr1_ref != @$arr2_ref) {
        return 0;
    }

    #sorting the arrays
    my @arr1 = sort @$arr1_ref;
    my @arr2 = sort @$arr2_ref;
    # arrays have same length, compare elements
    my $index = 0;
    foreach my $arr1_elem (@arr1) {
        if ($arr1_elem ne $arr2[$index]) {
            return 0;
        }
        $index++;
    }
    return 1;
} ## end sub string_array_compare($$)


# get the relevant zone name from the location LDAP entry
# parameters: LDAP entry for scLocation
# if an associatedDomain is present, use it
# otherwise generate the domain name from the LDAP DN path
sub get_zonename($)
{
    my $ldap_entry = shift;
    my $zonename   = "";

    if (!($zonename = $ldap_entry->get_value('associatedDomain'))) {
        my $dn = $ldap_entry->dn();

        my @zonename_parts = ();

        #print "get_zonename() dn:$dn\n";
        #my $dn_base = shiftbase($dn);
        my $dn_base = $dn;

        #print "get_zonename() dn_base:$dn_base\n";
        my @dn_parts = split(/,/, $dn_base);
        foreach my $dn_part (@dn_parts) {
            my @tmp_dn_part = split(/=/, $dn_part);
            push(@zonename_parts, $tmp_dn_part[1]);
        }
        # build zonename with all characters lowercase
        $zonename = lc(join(".", @zonename_parts));
    }
    return $zonename;
} ## end sub get_zonename($)

sub get_location_baselist($$)
{

    my ($self) = @_;
    my $modification_time = $self->{'Modification_Time'};

    # generate list of bases below a root base which have an organizationalUnit
    # entry but ignore ou=global
    my %baselist;

    # there is no scope "two", so we have to search twice with scope "one"
    my $result =
      $self->{'CACHE'}{'LDAP'}
      ->search(base => $self->{'STATIC'}{'BRANCHBASE'}, scope => 'sub', filter => '(objectClass=scLocation)');

    $result->code && die $result->error;

    foreach my $entry ($result->entries) {
        $baselist{ $entry->dn }             = {};
        $baselist{ $entry->dn }{'CONFIG'}   = $entry;
        $baselist{ $entry->dn }{'MODIFIED'} = 0;
    }


    ### perhaps add Filerobjects
    my $mod_result = $self->{'CACHE'}{'LDAP'}->search(
        base   => $self->{'STATIC'}{'BRANCHBASE'},
        scope  => 'sub',
        filter => '(modifyTimestamp>=' . $modification_time . 'Z)',
        attrs  => ['modifyTimestamp']
    );

    $mod_result->code && die $mod_result->error;


    foreach my $entry ($mod_result->entries) {
        my @a = split(/,/, $entry->dn);
        shift @a;
        while (@a) {
            if (defined($baselist{ join(",", @a) })) {
                $baselist{ join(",", @a) }{'MODIFIED'} = 1;
                last;
            }
            shift @a;
        }
    }

    return \%baselist;
} ## end sub get_location_baselist($$)

sub ldap2dns
{
    my ($self) = @_;

    # generate zones for all locations
    $self->gen_location_zones();

    # generate reverse zones
    $self->gen_rev_zones();

    # generate config include file
    $self->gen_config_file();

    if (@{ $self->{'Changed_Zones'} }) {
        system("/etc/init.d/named reload");
    }

    return 0;
} ## end sub ldap2dns
