## @file
# Implementation of POS::posImageMonitor

## @class
# Used to check for the image notification files and store the information in LDAP
package POS::posImageMonitor;

use strict;

use POS::Syslog;
use POS::LdapWorkstation;
use POS::CheckFile;
use POS::Config;
use POS::posLeaseMonitor;
use POS::posDataCache;
use MIME::Base64 ();
use File::Basename;
use POS::LdapUtil;
use Digest::MD5;


# version  of the package

our $VERSION = 0.1;

## Need to decide whether to use the ERRORCODE consistently or
## return code.  Haven't seen enough of the code yet to determine
## which is more common.
sub new
{
    my $package = shift;
    my $self    = {
        #init data
        'STATIC' => undef,    ## config data
        'CACHE'  => undef,    ## cached ldap data

        #used data
        'BOOTVERSION' => undef,

        'DELAY_TIME'    => 45,      # seconds of delay in case of errors and between loops
        'LEASEFILEOBJ'  => undef,
        'UPDATEFILEOBJ' => undef,
    };
    $self->{'ERROR'}{'CODE'}    = 0;
    $self->{'ERROR'}{'MESSAGE'} = '';

    %$self = (%$self, @_);

    bless($self);

    return unless $self->_init_process();
    return $self;
} ## end sub new


# Preconditions: Object must exist
# Postconditions: $self->{'STATIC'}{'CONFIG'} is valid
# Returns: 1 - configuration is loaded
#          0 - error loading configuration
# Notes:
## @method void load_config()
# Loads configuration from @em branchserver.conf or @em adminserver.conf file.
# @par Details:
# Reads data from the @em branchserver.conf or @em adminserver.conf file, via the POS::Config module, into
# the $self->{'STATIC'}{'CONFIG'} hash. Afterwards, the data are checked for valid values of important entries.
# @return 1 on success, 0 on failure (read problems, missing or invalid crucial entries)
sub _init_process
{
    my $self = shift;

    #$self->{'STATIC'}{'CONFIG'} = POS::Config->new($self->{'CONFIGFILEPATH'});
    if (!$self->{'STATIC'}{'CONFIG'}) {
        log_alert_print('Can\'t read configuration, exit now');
        # Unable to load configuration
        return 0;
    }

    #if ($self->get_netmask() == 0) {
    # error getting the POS::IP
    #    return 0;
    #}

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'})) {
        log_alert_print("TFTP_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }


    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'})) {
        log_alert_print("TFTP_CR_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_UPLOAD_BASE'})) {
        log_alert_print("TFTP_UPLOAD_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'TFTP_CR_ROLLBACK_BASE'})) {
        log_alert_print("TFTP_CR_ROLLBACK_BASE not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'POS_BOOTVERSION_PREFIX'})) {
        log_alert_print("POS_BOOTVERSION_PREFIX not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'POS_ROLLBACKLIST_PREFIX'})) {
        log_alert_print("POS_ROLLBACKLIST_PREFIX not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    if (!defined($self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'})) {
        log_alert_print("POS_CONFIG_PREFIX not defined, cannot proceed. Corrupted branchserver.conf file?...");
        return 0;
    }

    return 1;
} ## end sub _init_process


###############################################################################
###############################################################################
# rollback functionality related code
###############################################################################
## @method void refresh_rollback_list()
# Refreshes rollback file if needed
# @par Details:
# Reads list of all rolback files via get_rollbacfg_list(),
# reads LDAP (scNotifiedImage) using get_nots_from_ldap(),
# reads rollback file using get_nots_from_file(),
# and compares rollback files data with LDAP data, if they differ,
# the corresponding rollback file is updated write_rollback_file().
# Multiple MAC addresses per workstation (and ws<->rollback relation)
#  are determined by actual LDAP ws object state (not one when rollback was done)
# @return 1
sub refresh_rollback_list
{
    my $self = shift;
    #if (POS::LdapUtil::check_boolean_attribute($self->{'LDAP'},$self->{'ROOTBASE'},'scAllowRoles') == 0 ) {
    my $rollbackdir  = $self->{'STATIC'}{'ROLLBACK_DIR'};
    my $cfgbasepname = $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_ROLLBACK_BASE'};
    my $rlbasepathname =
      $self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'} .
      "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'} . '/' . $self->{'STATIC'}{'CONFIG'}{'POS_ROLLBACKLIST_PREFIX'};

    $self->{'ROLLBACKCFGS'} = $self->get_rollbacfg_list($rollbackdir);
    $self->{'ROLLBACKNOTS'} = $self->get_rollbacfg_nots();

    foreach my $wsid (keys %{ $self->{'CACHE'}{'WS'}{'ACT'} }) {
        #read data from LDAP, accept if roled back files exists
        my $ws_ref = $self->{'CACHE'}{'WS'}{'ACT'}{$wsid};
        my @macs   = $ws_ref->get_value('macAddress');
        $self->{'ROLLBACK'}{$wsid}{'WS'}   = $ws_ref;
        $self->{'ROLLBACK'}{$wsid}{'MACS'} = \@macs;
        $self->{'ROLLBACK'}{$wsid}{'LDAP'} = $self->get_nots_from_ldap($self->{'ROLLBACK'}{$wsid});
        #read data from rollback files and links
        $self->{'ROLLBACK'}{$wsid}{'FILE'} = $self->get_nots_from_file($rlbasepathname, $self->{'ROLLBACK'}{$wsid});
        if (POS::posLeaseMonitor::compare_hashes(
                '',
                $self->{'ROLLBACK'}{$wsid}{'FILE'},
                $self->{'ROLLBACK'}{$wsid}{'LDAP'},
                'MACS_CMPHASH', 'HASH', 'IMAGE', 'INDEX'
            )
          ) {    #1=different
            log_info_print("Updating rolback files for \'$wsid\'");
            $self->write_rollback_file($rlbasepathname, $self->{'ROLLBACK'}{$wsid});
        }

    }


    return 1;
} ## end sub refresh_rollback_list


## @method void cleanup_rollback_files()
# Cleans old rollback files if needed
# @par Details:
# Reads list of all rollback config.MAC.hash
# reads all referenced associated config files
# reads all referenced pxe.hash files
# reads (from those pxe files) all used initrd and kernel files
# deletes all what is present and not read before
# if initrd or kernel is deleted, associated file
#  is also checked in tftpboot/boot dir and deleted if present
# @return 1
sub cleanup_rollback_files
{
    my $self         = shift;
    my $rollcfg_dir  = $self->{'STATIC'}{'ROLLBACK_DIR'};
    my $rlbkbootpath = $self->{'STATIC'}{'ROLLBACK_BOOTDIR'};
    my $filepref     = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    my %safeconfigs;    #vals are just flag
    my %safeboots;      #vals are /tftpboot/boot pathnames for initrd and linurc, flag only for pxe


    #   delete all rollback/config.*.hash that are not referenced in any scNotifiedImage in ldap
    #  but only if not too young (so it may be still booting/expected to boot)

    my $waittime   = 10000;
    my @all_hashes = keys %{ $self->{'ROLLBACKCFGS'} };
    my %keep_hashes;
    foreach my $hash (@all_hashes) {
        $keep_hashes{$hash} += 1 if defined $self->{'ROLLBACKNOTS'}{$hash};
        $keep_hashes{$hash} += 2 if time() - $self->{'ROLLBACKCFGS'}{$hash}{'MTIME'} < $waittime;
    }
    my @cmds;
    foreach my $hash (@all_hashes) {
        if (!defined $keep_hashes{$hash}) {
            @cmds = (@cmds, ['unlink', $self->{'ROLLBACKCFGS'}{$hash}{'PATHNAME'}]);
            delete $self->{'ROLLBACKCFGS'}{$hash};
        }
    }
    $self->{'STATIC'}->runcmds(\@cmds);

    #now, remove all other rollback files that are no longer referenced in any remaining rollback/config.hash
    foreach my $hash (keys %{ $self->{'ROLLBACKCFGS'} }) {    #need to use updated ones
        my $cfg_hr = $self->rollback_parse_cfg($self->{'ROLLBACKCFGS'}{$hash}{'PATHNAME'}, 'MAC');
        $self->rollback_parse_pxe($cfg_hr, 'MAC');            #no need for real mac
                                                              #store present configs
        foreach my $cfghash (keys(%{ $cfg_hr->{'CONFIGS'} })) {
            $safeconfigs{ $cfg_hr->{'CONFIGS'}{$cfghash}{'ROLLBACKPATH'} } = 1;
        }
        my $pxe_hr = $self->{'CACHE'}->load_pxefile($cfg_hr->{'PXE'});
        if (!defined $pxe_hr) {
            log_warn_printf(
                "Rollback pxe file %s not found, rollback partly invalid, skipping kernel and initrd check for this",
                $cfg_hr->{'PXE'}{'ROLLBACKPATH'});
            next;
        }
        #store pxe
        $safeboots{ $cfg_hr->{'PXE'}{'ROLLBACKPATH'} } = 1;
        $safeboots{  $self->{'CACHE'}->name_to_grub_name($cfg_hr->{'PXE'}{'ROLLBACKPATH'}) } = 1;
        #store initrd
        $safeboots{ $rlbkbootpath . '/' . $pxe_hr->{'INITRD'} } = 1;
        $safeboots{ $rlbkbootpath . '/' . $pxe_hr->{'KERNEL'} } = 1;
    }

    #####
    #parse rollback directories and delete all unreferenced files
    log_info_printf("Checking unused rollback files for removal...");
    $self->sanitize_dir($self->{'STATIC'}{'ROLLBACK_CONFIGSDIR'}, \%safeconfigs);
    $self->sanitize_dir($rlbkbootpath, \%safeboots, $self->{'STATIC'}{'TFTP_BOOT_DIR'});

    return 1;

} ## end sub cleanup_rollback_files

#delete all but safe files in $dir,
#  if $extdir is supplied, also tryies to delete the file there
# if that file was present and deleted in $dir (useful for boot images)
sub sanitize_dir
{
    my ($self, $dir, $safefiles_hr, $extdir) = @_;
    my @cmds;
    opendir(DIRc, $dir);
    my @files = grep { -f "$dir/$_" } readdir(DIRc);
    foreach my $file (@files) {
        my $key = $dir . '/' . $file;
        unless (defined $safefiles_hr->{$key}) {
            @cmds = (@cmds, ['unlink', $key]);
            if (defined $extdir) {    #not a flag, full pathname, check that path also
                my $epname = $extdir . '/' . $file;
                @cmds = (@cmds, ['unlink', $epname]) if -e $epname;
            }
        }
    }

    closedir(DIRc);
    $self->{'STATIC'}->runcmds(\@cmds);

} ## end sub sanitize_dir

## @method $ get_nots_from_ldap( $ ws_ref, $ basepname)
# Reads rollback data from LDAP
# @par Details:
#  Reads rollback data from LDAP (scNotification parameter of ws_ref),
#  checks if their associated configs really exists, ignores notification from non-role based entries (without hash).
# @param $ws_ref a workstation dn
# @param $basepname directory containing the rolled-back config files + filename prefix:
# @return \\%%rollbackhashall->{$rollbackhash{'HASH'}} hash reference of all rollback config files witha given MAC, containing:
# @arg @c 'IMAGE' image file name
# @arg @c 'BVTIME' time of notification reception
# @arg @c 'HASH' hash id of the rolledback config file
# @arg @c 'MAC' MAC address of the terminal
# @arg @c 'PNAME' TFTP_CR_ROLLBACK_BASE/POS_CONFIG_PREFIX
# @arg @c 'PATHNAME' full pathname of the rolled back config file
# @arg @c 'INDEX' index of ordering by the 'BVTIME'
sub get_nots_from_ldap
{
    my ($self, $rollhash_ref) = @_;
    my %rollbackhashall;
    my @macs   = @{ $rollhash_ref->{'MACS'} };
    my $ws_ref = $rollhash_ref->{'WS'};
    if ($ws_ref->get_value('scNotifiedImage')) {
        foreach my $ims ($ws_ref->get_value('scNotifiedImage')) {    #can be multival
            my %rollbackhash;
            #log_debug_printf("scNotifiedImage=$ims");
            my @imparts = split(';', $ims);
            foreach my $part (@imparts) {                            #trim ; separated data
                $part =~ s/^\s*//;
                $part =~ s/\s*$//;
                #log_debug_printf("scNotifiedImageT=\'$part\'");
            }
            my $pos_image = POS::Image->new();
            my @tmp = split('/', $imparts[0]);
            $rollbackhash{'IMAGE'}  = pop @tmp;
            $rollbackhash{'BVTIME'} = $imparts[1];
            my @imgnameparts = split('-', $rollbackhash{'IMAGE'});
            my $img_version  = pop @imgnameparts;
            my $image_name   = join('-', @imgnameparts);

            my $hash = $imparts[2];
            if (!defined $hash) {
                log_warn_printf("No hash in scNotifiedImage entry \'$ims\' of %s, skipping.", $ws_ref->dn());
                next;
            }
            $rollbackhash{'HASH'} = $hash;
       # find rolled back config pathname, check if exists (need just one, for any mac of ws, since links are not saved)
            my $rollbk_ref = $self->{'ROLLBACKCFGS'}{ $rollbackhash{'HASH'} };
            if (!defined $rollbk_ref) {
                log_info_printf("no rollback config found for record \'$ims\' of \'%s\' scWorkstation, skipping.",
                    $ws_ref->dn());
                next;
            }
            elsif (!$pos_image->lookup($self->{'STATIC'}{'TFTP_IMAGE_DIR'}, $image_name, $img_version)) {
                log_info_printf("image file not found for record \'$ims\' of \'%s\' scWorkstation, skipping.",
                    $ws_ref->dn());
                next;
            }
            else {
                $rollbackhash{'PNAME'} =
                  $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_ROLLBACK_BASE'} . '/' . $rollbk_ref->{'NAME'};  #relative to KIWI
                $rollbackhash{'PATHNAME'} = $self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'} . '/' . $rollbackhash{'PNAME'};
                $rollbackhash{'MACS_CMPHASH'} = join(',', sort @macs);
            }

            $rollbackhashall{ $rollbackhash{'HASH'} } = \%rollbackhash;
        }
        #compute index (reflecting sort by mtime, highest mtime=lowest index)
        my @ldaparray       = values %rollbackhashall;
        my @ldaparraysorted = sort {
            $b->{'BVTIME'} <=> $a->{'BVTIME'}                                                          #num. comparison
        } @ldaparray;
        my $index = 0;
        foreach my $rollback (@ldaparraysorted) {
            $rollbackhashall{ $rollback->{'HASH'} }{'INDEX'} = $index;
            $index++;
        }

    }
    return \%rollbackhashall;
} ## end sub get_nots_from_ldap

## @method $ get_rollbacfg_list( $ basedir)
# Refreshes rollback file if needed.
# @par Details:
# Reads list of all rollback files located in $basedir,
# gathers name, MAC, hash id, creation time.
# @param $basedir directory containing the rollback files
# @return \\%%cfgmachash->{$mac}{$hash} hash reference containing:
# @arg @c 'PATHNAME' image file name
# @arg @c 'MAC' MAC address of the terminal
# @arg @c 'HASH' hash id of the rolledback config file
# @arg @c 'MTIME' time of the rolled back config file creation
sub get_rollbacfg_list
{
    my ($self, $basedir) = @_;
    my %cfgmachash;
    my $prefix = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    # print " $dir $prefix \n";
    my @lines = <$basedir/$prefix*>;    # perl-style file globbing operator
                                        #      print @lines,"\n";
    foreach my $pathname (@lines) {
        chomp $pathname;
        my $filtmp  = $self->{'STATIC'}{'MACCFG'}{'MACFILTER'};
        my $mp      = $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP'};
        my $cfgfile = basename($pathname);
        if ($cfgfile =~ /^$prefix\./ && $cfgfile =~ /\.([\da-zA-Z]*)$/) {    #config.???.md5hash
#            if ($cfgfile =~ /^$prefix\.([\d$filtmp]*)\.([\da-zA-Z]*)$/) {    #config.MAC.md5hash
            #if ($filename =~ /^$prefix\.([\d$filtmp]*)($|\.([\da-zA-Z]*$))/)
            #mac is no longer useful/needed for multi-MAC cases in FP1
            #my $outmac = $1;
            my $hash = $1;
            if ($hash) {

                $cfgmachash{$hash}{'NAME'}     = $cfgfile;
                $cfgmachash{$hash}{'PATHNAME'} = $pathname;
                $cfgmachash{$hash}{'HASH'}     = $hash;
                $cfgmachash{$hash}{'MTIME'}    = (stat($pathname))[9];
            }
        }
    }
    return \%cfgmachash;
} ## end sub get_rollbacfg_list


#returns hash keyed by scNotification hashes, values are lists of workstation objects referencing them
sub get_rollbacfg_nots
{
    my ($self) = @_;
    my %notshash;
    foreach my $ws_ref (values %{ $self->{'CACHE'}{'WS'}{'ALL'} }) {
        my @notifs = $ws_ref->get_value('scNotifiedImage');
        my $wscn   = $ws_ref->get_value('cn');
        foreach my $notif (@notifs) {
            my @parts = split(';', $notif);
            my $hash = $parts[2];
            $notshash{$hash}{$wscn} = $ws_ref if defined $hash;    #by wscn to remove duplicates within $ws_ref
        }
    }

    return \%notshash;
} ## end sub get_rollbacfg_nots


#read rollback.MACs list files and get data on configs from there
#data is sorted by time in the file
#todo: get macs directly as a parameter??
#TODO: info on return error method via special hash
## @method $ get_nots_from_file( $pathname, $rollhash_ref)
# Reads rollback data from @em rollback.MAC file
# @par Details:
#  Reads rollback data from @em  rollback.MACs files in $basepathname,
#  checks if their associated config names make sense and really exists.
#  This is done only if @em  rollback.MAC file changed after previous read (and the
#  data are present in $rollhash_ref )
# @param $pathname pathname of the @em  rollback.MAC file
# @param $rollhash_ref hash reference on previously read, current data
# @return \\%%frollshash hash reference for the rollback information read from the @em  rollback.MAC file, containing:
# @arg @c 'IMAGE' image file name
# @arg @c 'BVTIME' time of notification reception
# @arg @c 'HASH' hash id of the rolledback config file
# @arg @c 'MAC' MAC address of the terminal
# @arg @c 'DESCRIPTION' description of the rollback (human readable summary)
# @arg @c 'PNAME' TFTP_CR_ROLLBACK_BASE/POS_CONFIG_PREFIX
# @arg @c 'PATHNAME' full pathname of the rolled back config file
# @arg @c 'INDEX' index of ordering by the 'BVTIME'
sub get_nots_from_file
{
    my ($self, $basepathname, $rollhash_ref) = @_;
    my %frollshash;
    my $prefix = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    $rollhash_ref->{'FILETIME'} = 0 if !defined $rollhash_ref->{'FILETIME'};    #init
    my $filemac;
    my @macs;
    my $pathname;
    #check what rollback.MAC files/links exist for each MAC
    foreach my $mac (@{ $rollhash_ref->{'MACS'} }) {
        my $tmpathname = $basepathname . '.' . $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($mac);
        if (-e $tmpathname) {
            @macs = (@macs, $mac);
            if (-e $tmpathname && !-l $tmpathname) {                            #find which one is file
                $pathname = $tmpathname;
                $filemac  = $mac;
            }
        }
    }
    return undef if !defined $pathname;
    my $rollfiletime = (stat($pathname))[9];
    if (defined $rollfiletime) {
        if ($rollfiletime == $rollhash_ref->{'FILETIME'} && defined $rollhash_ref->{'FILE'}) {
            # if file not changed, and FILE hash already stored, just return data from memory
            log_debug_printf("reading data for \'%s\' from stored hash", $pathname);
            %frollshash = %{ $rollhash_ref->{'FILE'} };
        }
        else {
            log_debug_printf("reading file \'%s\'", $pathname);
            if (open(FHH, '<', $pathname)) {
                my @lines = <FHH>;
                my $index = 0;       #line number corrected for repeated entries
                foreach my $line (@lines) {
                    $line =~ s/\n//;      #remove new line
                    $line =~ s/\s*$//;    #remove leading whitespace
                    $line =~ s/^\s*//;    #remove trailing whitespace
                    my @parts = split('\|', $line);
                    next unless @parts;    #ignore empty lines
                    my $pname  = shift @parts;
                    my $bvtime = shift @parts;
                    my $image  = shift @parts;    #ignore description text
                    my $desc   = shift @parts;

                    my $filtmp = $self->{'STATIC'}{'MACCFG'}{'MACFILTER'};
                    my $mp     = $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP'};
                    my $name   = basename($pname);
                    #TODO: also assoc. config., pxe check
                    if ($name =~ /^$prefix\./ && $name =~ /\.([\da-zA-Z]*)$/) {
                        my $hash = $1;
                        $frollshash{$hash}{'BVTIME'}      = $bvtime;
                        $frollshash{$hash}{'IMAGE'}       = $image;
                        $frollshash{$hash}{'DESCRIPTION'} = $desc;
                        $frollshash{$hash}{'HASH'}        = $hash;
                        $frollshash{$hash}{'PNAME'}       = $pname;    #reduced pathname
                        $frollshash{$hash}{'PATHNAME'}     = $self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'} . "/" . $pname;
                        $frollshash{$hash}{'INDEX'}        = $index;
                        $frollshash{$hash}{'MACS_CMPHASH'} = join(',', sort @macs);
                    }
                    else {    # rollback not found, could also be due to MAC change
                        $frollshash{'!!!!!!!!!!!!!!'}{'MAC'} =
                          '!!ERROR!!';    #mark this bad line in HASH also so comparison fails
                        log_warn_printf("invalid path \'$pname\' in the $pathname file, skipping.");
                        return (\%frollshash);    # as if file was empty, it is invalid and must be created again
                    }
                    $index++;
                }    #foreach line
                close(FHH);
                $rollhash_ref->{'FILETIME'} = $rollfiletime;
            }    #open
            else {
                log_warn_printf('Can\'t open file \'%s\' ', $pathname);
            }
        }
    }
    return (\%frollshash);
}    #end of get_nots_from_file

## @method $ write_rollback_file( $pathname, $rollhash_ref)
# Writes rollback data into rollback.MAC file
# @par Details:
#  Writes rollback data from $rollhash_ref into rollback.MACs file $pathname,
#  ordered by time ('BVTIME').
# @param $basepathname pathname of the directory of rollback.MACs files
# @param $rollhash_ref hash reference on previously read, current data
# @return 0 if file written, 1 on problems
sub write_rollback_file
{
    my ($self, $basepathname, $rollhash_ref) = @_;
    my @macs     = @{ $rollhash_ref->{'MACS'} };                                               #at least one must exists
    my $pathname = $basepathname . '.' . $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($macs[0]);

    my ($fname, $path) = fileparse($pathname);
    my $tmppathname = $path . '/' . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $fname;
    my @rolllist;
    if (open(FHW, '>', $tmppathname)) {
        log_info_printf("writting file \'%s\' ", $pathname);
        foreach my $hash (keys %{ $rollhash_ref->{'LDAP'} }) { #create intermediate hash with keys being the depth level
            $rolllist[$rollhash_ref->{'LDAP'}{$hash}{'INDEX'}] = $rollhash_ref->{'LDAP'}{$hash};
        }
        foreach my $rollcfg (@rolllist) {
            my $ltime = localtime($rollcfg->{'BVTIME'});
            my $desc  = "Image \'$rollcfg->{'IMAGE'}\' of $ltime";
            print FHW "$rollcfg->{'PNAME'}|$rollcfg->{'BVTIME'}|$rollcfg->{'IMAGE'}|$desc\n";
        }
        close(FH);
        system('mv', $tmppathname, $pathname);
        #make links, (not for the first one...)
        for (my $i = 1; $i < @macs; $i++) {
            my $lnkconf_file = $basepathname . '.' . $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($macs[$i]);
            system("ln -sf \"$fname\" \"$lnkconf_file\"");
        }
        return 0;
    }
    else {
        log_warn_printf('Can\'t open file \'%s\' ', $pathname);
        return 1;
    }


} ## end sub write_rollback_file


## @method $ trim_space( $entry)
# Trims whitespace at the beginning and at the end.
# @par Details:
# Procedure to trim whitespace ('\s+') from the beginning and end of a string.
# @paramaters $entry input string
# @return $entry trimmed string
sub trim_space
{
    my ($self, $entry) = @_;
    $entry =~ s/^\s+//;
    $entry =~ s/\s+$//;
    return $entry;
} ## end sub trim_space


###############################################################################
###############################################################################


## @method void load_bootversion()
# Gathers uploaded bootversion files.
# @par Details:
# find_bootversion() is used to gather
#  uploaded bootversion files and information into the $self->{'BOOTVERSION'} hash.
# @return 1 on success, 0 on failure
sub load_bootversion
{
    my $self = shift;

    $self->{'BOOTVERSION'} = $self->find_bootversion();
    (%{ $self->{'BOOTVERSION'} }) ? return 1 : return 0;    #if empty hash return 0

} ## end sub load_bootversion

## originally in the posImageInstallNotify.pm
## @method $ find_bootversion()
# Checks for an uploaded bootversion file.
# @par Details:
# Checks all filenames in '/srv/tftpboot/upload'. If bootversion.<MAC> found,
# it is stored, together with image pathname and version string obtained from it.
# @return \\%%bootversionhash hash reference containing data found
sub find_bootversion()
{
    my ($self) = @_;
    my %bootversionhash;
    my $dir      = $self->{'STATIC'}{'UPLOAD_DIR'};
    my $dirbk    = $self->{'STATIC'}{'UPLOAD_BCK_DIR'};
    my $prefix   = $self->{'STATIC'}{'CONFIG'}{'POS_BOOTVERSION_PREFIX'};    #i.e. 'bootversion'
    my $archname = $prefix;
    my $waittime = 2;                  #read only bootverisons at least 2s old to prevent partial uploads
                                       # print " $dir $prefix \n";
    my @lines    = <$dir/$prefix*>;    # perl-style file globbing operator
    FILE:
    foreach my $file (@lines) {
        chomp $file;
        my $filtmp = $self->{'STATIC'}{'MACCFG'}{'MACFILTER'};
        my $mp     = $self->{'STATIC'}{'MACCFG'}{'POS_MACSEP'};
        my $a      = basename($file);
        $a =~ /^$prefix\.([\d$filtmp]*)$/;    # DO NOT ignorecase search for hex MAC address
        my $outmac = $1;
        if ($outmac) {
            my $mac =
              $self->{'STATIC'}{'MACCFG'}->outmac_to_mac($outmac);    # for internal purpose, always standard format
            if (open(FH, '<', $file)) {
                my $mtime = (stat(FH))[9];
                my $tdiff = time() - $mtime;
                if ($tdiff > $waittime) {
                    my %datahash;
                    log_info_printf("Processing \'$file\'...");
                    my @flines = <FH>;
                    close FH;
                    #backup bootversion file
                    my @cmds = $self->{'STATIC'}->backup($file, $dirbk, $archname);
                    $self->{'STATIC'}->runcmds(\@cmds);
                    if (@flines < 2) {
                        log_warn_printf(
                            "Invalid data in \'$file\', no image pathname or version on first two lines, ignoring...");
                        next FILE;
                    }
                    my $flinePath = shift(@flines);
                    chomp $flinePath;
                    if (!length $flinePath) {
                        log_warn_printf("Invalid data in \'$file\', image pathname empty, ignoring this file...");
                        next FILE;
                    }
                    my $flineVersion = shift(@flines);
                    if (!length $flineVersion) {
                        log_warn_printf("Invalid data in \'$file\', image version empty, ignoring this file...");
                        next FILE;
                    }
                    chomp $flineVersion;
                    if ($flineVersion !~ /^[\d\.]+$/) {
                        log_warn_printf("Invalid image version \'$flineVersion\', ignoring this file...");
                        next FILE;
                    }

                    $flinePath .= "-" . $flineVersion;
                    #push(@{ $self->{'BOOTVERSION'}{$mac} }, $flinePath);
                    #push(@{ $self->{'BOOTVERSION'}{$mac} }, $mtime);
                    #read the rest of the file, having same format as the hwtype.MAC file, i.e. KEY=VALUE
                    foreach my $line (@flines) {
                        chomp $line;
                        my ($arg, $val) = split(/=/, $line, 2);
                        if ($arg && $val) {
                            $self->trim_space($val);
                            #$val =~ s/^\s+//;
                            #$val =~ s/\s+$//;
                            if (!defined $self->{'STATIC'}{'WHITELIST'}{'BOOTVERISON'}{$arg}) {
                                log_warn_printf(
                                    "key \'$arg\' in file \'$file\' is not whitelisted, ignoring it...");
                                next;
                            }
                            my $regexp = $self->{'STATIC'}{'WHITELIST'}{'BOOTVERISON'}{$arg};
                            if ($val !~ /$regexp/) {
                                log_warn_printf(
"value \'$val\' of key \'$arg\' in file \'$file\' does not conform to regexp /$regexp/, ignoring this file..."
                                );
                                next FILE;
                            }
                            $datahash{$arg} = $val;
                        }
                    }
                    $bootversionhash{$mac}{'PATHNAME'} = $flinePath;      #POS 11 SP1
                    $bootversionhash{$mac}{'VERSION'}  = $flineVersion;
                    $bootversionhash{$mac}{'TIME'}     = $mtime;
                    $bootversionhash{$mac}{'MAC'}      = $mac;
                    $bootversionhash{$mac}{'DATA'}     = \%datahash;
                }    #fi of time test
                else {
                    log_info_printf('File \'%s\' is too young (%ss), maybe not yet complete, skipping this time...',
                        $file, $tdiff);
                }
            }    #fi of file test
            else {
                log_warn_printf('Can\'t open %s', $file);
            }
        }
    }
    return (\%bootversionhash);
} ## end sub find_bootversion()


###############################################################################
###############################################################################


## @method void process_image_reports( $ ws_ref, $ basepname)
# Reads uploaded bootversion information and updates LDAP
# @par Details:
# Reads bootversion information from $self->{'BOOTVERSION'}, updates
# the scNotifiedImage attribute in correspoding scWorkstation object by create_workstations_from_image_reports()
# and if requested, provides rollback functionality via copy_referred_cfg().
# @param $ws_ref a workstation dn
# @param $basepname directory containing the rolled-back config files + filename prefix:
# @return 1
sub process_image_reports
{
    my $self = shift;

    #TODO: simplify sub, put into he cycle bellow
    $self->create_workstations_from_image_reports();

    foreach my $mac (keys(%{ $self->{'BOOTVERSION'} })) {
        my $bootvhash_ref = $self->{'BOOTVERSION'}{$mac};
        $self->rollback_referred_cfg($bootvhash_ref);
    }

} ## end sub process_image_reports

#copy stored rollback config.(MAC).hash back onto current config.MAC (and store current one too..)
# recreate all links to config.MAC, pxe and configuration files, according to the current workstation entries
## @method void copy_referred_cfg( $bv_ref)
# Provides rollback functionality
# @par Details:
# If rollback is requested in the boot notification (POS_ROLLBACK=1), relevant @em config.MAC.hash is
# copied  from the rollback directory back into the KIWI directory as the @em config.MAC file.
# Checks if their associated configs really exists, ignores notification from non-role based
# entries (without hash).
# @param $bv_ref a hash reference for bootversion notification data ($self->{'BOOTVERSION'}{$mac})
# @return 1
sub rollback_referred_cfg
{
    my ($self, $bv_ref) = @_;

    my $mainmac = $bv_ref->{'MAC'};    #use to find associated workstation and to name rolled back files
    my $ws_ref = $self->{'CACHE'}{'WS'}{'BY-MAC'}{$mainmac};
    if (!defined $ws_ref) {
        log_warn_printf(
'LDAP-Error: Cannot find scWorkstation with MAC:  %s, invalid/deleted registration? Skipping this rollback...',
            $mainmac
        );
        return undef;
    }
    my @macs = $ws_ref->get_value('macAddress');
    my $hash = $bv_ref->{'DATA'}{'POS_CFG_HASH'};
    if (defined $hash && $bv_ref->{'DATA'}{'POS_ROLLBACK'}) {    #rollback image was used
        my @linkmacs;                                            #(these other macs will be used to create links)
        foreach my $mac (@macs) {
            @linkmacs = (@linkmacs, $mac) unless $mac eq $mainmac;
        }

        my $rollcfgpathname = $self->{'ROLLBACKCFGS'}{$hash}{'PATHNAME'};
        if (-e $rollcfgpathname) {
            #load rollback config.MAC, get config files data, get pxe file data
            my $cfg_hr = $self->rollback_parse_cfg($rollcfgpathname, $mainmac, \@linkmacs);
            $self->rollback_parse_pxe($cfg_hr, $mainmac);

            #rollback stored config.MAC
            log_info_printf("rollback of \'$cfg_hr->{'PATHS'}{'ROLLBACK'}\'");
            $self->rollback_config($cfg_hr, $mainmac);

            #rollback stored associated config files
            log_info_printf("rollback of configuration files");
            $self->rollback_assoc_cfgs($cfg_hr, $mainmac);

            #rollback pxe file (always as a specific pxe)
            log_info_printf("rollback of pxe file");
            $self->rollback_pxe($cfg_hr, $mainmac);
        }
        else {
            #maybe create error config if rollback not found??
            log_error_printf("Cannot find requested \'$rollcfgpathname\' config.MAC will not be rolled back.");
            return undef;
        }
    }
} ## end sub rollback_referred_cfg


#restore config.MAC
#also change MAC dir of assoc. configs
sub rollback_config
{
    my ($self, $cfghash_ref, $mainmac) = @_;
    #backup previous if exists (as file)
    my @all_cfgmacs = ($cfghash_ref->{'PATHS'}{'KIWI'}, @{ $cfghash_ref->{'PATHS'}{'KIWI_LINKS'} });
    foreach my $cfg_pathname (@all_cfgmacs) {
        #TODO make a rollback of current config.MAC file?
        if (-e $cfg_pathname && !-l $cfg_pathname) {
            my $dirbk    = $self->{'STATIC'}{'KIWI_BCK_DIR'};
            my $archname = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
            log_info_printf("backup of \'$cfg_pathname\' into \'$dirbk\'");
            my @cmdbks = $self->{'STATIC'}->backup($cfg_pathname, $dirbk, $archname, 1);
            $self->{'STATIC'}->runcmds(\@cmdbks);    #backup file
        }
    }
    #restore rolled back one
    my @cmds;
    open(FI, "$cfghash_ref->{'PATHS'}{'ROLLBACK'}");
    my @rconfig_a = <FI>;
    close FI;
    my $modline = $cfghash_ref->{'MOD'}{'CONF'};
    open(FO, '>', $cfghash_ref->{'PATHS'}{'ROLLBACK_TMP'});
    foreach my $line (@rconfig_a) {
        $line =~ s/(^CONF=).*$/$1$modline/g if defined $modline;
        print FO $line;
    }
    close FO;
    @cmds = (@cmds, ['move', $cfghash_ref->{'PATHS'}{'ROLLBACK_TMP'}, $cfghash_ref->{'PATHS'}{'KIWI'}]);
    #make links
    foreach my $linkpathname (@{ $cfghash_ref->{'PATHS'}{'KIWI_LINKS'} }) {
        @cmds = (@cmds, ['symlink', $cfghash_ref->{'PATHS'}{'NAME'}, $linkpathname]);
    }
    return $self->{'STATIC'}->runcmds(\@cmds);

} ## end sub rollback_config

#todo: unify with posLeaseMonitor's code for assoc. cfg. creation
sub rollback_assoc_cfgs
{
    my ($self, $cfghash_ref, $mainmac) = @_;
    my @cmds;
    #ensure directory really exists
    my $cfgpath = $cfghash_ref->{'PATHS'}{'KIWIDIR'};
    @cmds = (@cmds, ['unlink', $cfgpath]) if (-l $cfgpath || !-d $cfgpath);    #remove links, files
    @cmds = (@cmds, ['external', 'install', '-d', $cfgpath]);
    foreach my $cf_hr (values %{ $cfghash_ref->{'CONFIGS'} }) {
        @cmds = (@cmds, ['copy', $cf_hr->{'ROLLBACKPATH'}, $cf_hr->{'KIWIPATH'}]);
    }
    #make links
    foreach my $linkdir_pathname (@{ $cfghash_ref->{'PATHS'}{'KIWI_DIRLINKS'} }) {
        #TODO: maybe keep old configs, just in case?
        #remove dir so link can be created
        @cmds = (@cmds, $self->{'STATIC'}->rmdir_safe($linkdir_pathname, ['/KIWI/[^*/]+$']))
          if (-d $linkdir_pathname && !-l $linkdir_pathname);
        @cmds = (@cmds, ['symlink', $cfghash_ref->{'PATHS'}{'CFGDIRNAME'}, $linkdir_pathname]);
    }
    return $self->{'STATIC'}->runcmds(\@cmds);


} ## end sub rollback_assoc_cfgs

sub rollback_pxe
{
    my ($self, $cfghash_ref, $mainmac) = @_;
    my @cmds;
    my $pxepn = $cfghash_ref->{'PXE'}{'KIWIPATH'};
    my $pxepn_grub = $self->{'CACHE'}->name_to_grub_name($pxepn);
    @cmds = (@cmds, ['unlink', $pxepn]) if (-e $pxepn);    #remove needed for links
    @cmds = (@cmds, ['unlink', $pxepn_grub]) if (-e $pxepn_grub);    #remove needed for links
    @cmds = (@cmds, ['copy', $cfghash_ref->{'PXE'}{'ROLLBACKPATH'}, $pxepn]);
    @cmds = (@cmds, ['copy', $self->{'CACHE'}->name_to_grub_name($cfghash_ref->{'PXE'}{'ROLLBACKPATH'}), $pxepn_grub]);
    #make links
    foreach my $linkpxe_pathname (@{ $cfghash_ref->{'PXE'}{'KIWIPATH_LINKS'} }) {
        @cmds = (@cmds, ['symlink', $cfghash_ref->{'PXE'}{'KIWINAME'}, $linkpxe_pathname]);    #force
        @cmds = (@cmds, ['symlink', $self->{'CACHE'}->name_to_grub_name($cfghash_ref->{'PXE'}{'KIWINAME'}), $self->{'CACHE'}->name_to_grub_name($linkpxe_pathname)]);
    }

    #rollback initrd and kernel file
    $self->{'CACHE'}->load_pxefile($cfghash_ref->{'PXE'});
    my $rollinitrdpname = $cfghash_ref->{'PXE'}{'PATH'} . '/' . $cfghash_ref->{'PXE'}{'FILE'}{'INITRD'};
    my $rollkernelpname = $cfghash_ref->{'PXE'}{'PATH'} . '/' . $cfghash_ref->{'PXE'}{'FILE'}{'KERNEL'};
    @cmds = (@cmds, ['copy', $rollinitrdpname, $self->{'CACHE'}{'DEFAULT_PXE'}{'BOOTPATH'}]);
    @cmds = (@cmds, ['copy', $rollkernelpname, $self->{'CACHE'}{'DEFAULT_PXE'}{'BOOTPATH'}]);

    return $self->{'STATIC'}->runcmds(\@cmds);

} ## end sub rollback_pxe

##parses rolled back config.MAC file
##also return pxenames for pxe links, if $linkmac_ref supplied
sub rollback_parse_cfg
{
    my ($self, $pathname, $mainmac, $linkmac_ref) = @_;
    my %cfghash;
    my %error_hash;
    #gather pathnames
    my $rollbackdir = $self->{'STATIC'}{'ROLLBACK_DIR'};
    my $cpref       = $self->{'STATIC'}{'CONFIG'}{'POS_CONFIG_PREFIX'};
    my $rollbacktmpcfgbase =
      $self->{'STATIC'}{'ROLLBACK_DIR'} . '/' . $self->{'STATIC'}{'CONFIG'}{'POS_TMP_PREFIX'} . $cpref;
    #/srv/tftpboot/KIWI/rollback/TMP_config

    my $cfgdir =
      $self->{'STATIC'}{'CONFIG'}{'TFTP_BASE'} .
      "/" . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'};    #/srv/tftpboot / KIWI
    my $cfgbase    = $cfgdir . '/' . $cpref;                                 #/srv/tftpboot/KIWI/config
    my $outmainmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($mainmac);
    my $currentcfg = $cfgbase . '.' . $outmainmac;
    $cfghash{'MOD'}{'CONF'} = undef;

    my ($data_hr) = $self->parse_file($pathname);
    return undef if !defined $data_hr;
    $cfghash{'LINKMACS'}              = $linkmac_ref;
    $cfghash{'PATHS'}{'ROLLBACK'}     = $pathname;
    $cfghash{'PATHS'}{'NAME'}         = $cpref . '.' . $outmainmac;
    $cfghash{'PATHS'}{'KIWI'}         = $currentcfg;
    $cfghash{'PATHS'}{'CFGDIRNAME'}   = $outmainmac;
    $cfghash{'PATHS'}{'KIWIDIR'}      = $cfgdir . "/" . $outmainmac;
    $cfghash{'PATHS'}{'ROLLBACK_TMP'} = $rollbacktmpcfgbase . '.' . $outmainmac;
    #gather linknames
    my @cfg_links;       #for config.MAC
    my @cfg_dirlinks;    #for assoc configs
    foreach my $lnkmac (@{$linkmac_ref}) {
        my $outlnkmac = $self->{'STATIC'}{'MACCFG'}->mac_to_outmac($lnkmac);
        @cfg_links    = (@cfg_links,    $cfgbase . '.' . $outlnkmac);
        @cfg_dirlinks = (@cfg_dirlinks, $cfgdir . "/" . $outlnkmac);
    }
    $cfghash{'PATHS'}{'KIWI_LINKS'}    = \@cfg_links;
    $cfghash{'PATHS'}{'KIWI_DIRLINKS'} = \@cfg_dirlinks;

    #parse CONFIGS
    $cfghash{'DATA'} = $data_hr;
    my $modcfgline;
    my $cfgline = $cfghash{'DATA'}{'CONF'};
    $cfgline = '' if !defined $cfgline;
    my @cfgs_a = split(',', $cfgline);
    foreach my $cfg (@cfgs_a) {
        my @cfg_a = split(';', $cfg);
        my $cfg_hash = $cfg_a[4];
        if (!defined $cfg_hash) {
            log_warn_printf("CONF line invalid in \'%s\' (config without hash: %s), this cfg. cannot be rolled back...",
                $pathname, $cfg);
        }
        else {
            my $cfg_pname = $cfg_a[0];    #assuming /KIWI/<MAC>/cfgpath structure in CONF
            $cfg_pname =~ s@/.*/.*/@@ig;
            $cfg_a[0] = '/' . $self->{'STATIC'}{'CONFIG'}{'TFTP_CR_BASE'} . '/' . $outmainmac . '/' . $cfg_pname;
            my $cfg_kiwipname = $cfg_a[0];
            $cfghash{'CONFIGS'}{$cfg_hash}{'ROLLBACKPATH'} =
              $self->{'STATIC'}{'ROLLBACK_CONFIGSDIR'} . '/' . $cfg_pname . '.' . $cfg_hash;
            $cfghash{'CONFIGS'}{$cfg_hash}{'KIWIPATH'} = $self->{'STATIC'}{'TFTP_DIR'} . $cfg_kiwipname;
            #add more data if needed
            $modcfgline .= ',' if $modcfgline;
            $modcfgline .= join(';', @cfg_a);
        }
    }
    $cfghash{'MOD'}{'CONF'} = $modcfgline;
    return (\%cfghash);
} ## end sub rollback_parse_cfg


##uses conf hash to create CONF= line with current MAC
##also return pxenames for pxe links, if $linkmac_ref supplied
sub rollback_parse_pxe
{
    my ($self, $cfghash_ref, $mac) = @_;
    #parse pxe
    my $pxehash = $cfghash_ref->{'DATA'}{'POS_KERNEL_PARAMS_HASH_VERIFY'};
    if (!defined $pxehash) {
        log_warn_printf("pxe hash (POS_KERNEL_PARAMS_HASH_VERIFY) not defined in \'%s\', pxe file rollback may fail...",
            $cfghash_ref->{'PATHNAME'});
        $pxehash = '';
    }
    my $pxename = $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($mac);
    $cfghash_ref->{'PXE'}{'NAME'}         = $self->{'STATIC'}{'ROLLBACK_PXEPREFIX'} . '.' . $pxehash;
    $cfghash_ref->{'PXE'}{'PATH'}         = $self->{'STATIC'}{'ROLLBACK_BOOTDIR'};
    $cfghash_ref->{'PXE'}{'ROLLBACKPATH'} = $cfghash_ref->{'PXE'}{'PATH'} . '/' . $cfghash_ref->{'PXE'}{'NAME'};
    $cfghash_ref->{'PXE'}{'KIWINAME'}     = $pxename;
    $cfghash_ref->{'PXE'}{'KIWIPATH'}     = $self->{'STATIC'}{'TFTP_PXE_DIR'} . '/' . $pxename;
    my @pxe_links;

    foreach my $lnkmac (@{ $cfghash_ref->{'LINKMACS'} }) {
        my $lnkpxename = $self->{'STATIC'}{'MACCFG'}->mac_to_pxename($lnkmac);
        @pxe_links = (@pxe_links, $self->{'STATIC'}{'TFTP_PXE_DIR'} . '/' . $lnkpxename);
    }
    $cfghash_ref->{'PXE'}{'KIWIPATH_LINKS'} = \@pxe_links;
    return ($cfghash_ref);

} ## end sub rollback_parse_pxe


## general file parser for 'KEY=VALUE' type of config file
sub parse_file
{
    my ($self, $pathname) = @_;
    my %datahash;
    if (open(FH, '<', $pathname)) {
        binmode(FH);    #seems not really necessary
        my @lines = <FH>;
        close FH;
        #      my $md5hash = md5_hex(@lines);
        foreach my $line (@lines) {
            chomp $line;
            my ($arg, $val) = split(/=/, $line, 2);
            if ($arg) {    # allow empty values ('arg=')
                $val =~ s/^\s+//;
                $val =~ s/\s+$//;
                $datahash{$arg} = $val;
            }
        }
    }
    else {
        log_error_printf("Cannot open hwtype file \'$pathname\', skipping...");
        return (undef);
    }
    return (\%datahash);

} ## end sub parse_file


# Preconditions: $self->{'STATIC'}{'CONFIG'} $self->{'CR2CONFIG'}{'BOOTVERSION'}
# Postconditions:
# Returns: 1 - has leases in hash
#          0 - unable to get lease data
# Notes:
## @method void create_workstations_from_image_reports()
# Writes information from uploaded bootversion files into LDAP.
# @par Details:
# If boot notification file @em bootversion.MAC contains POS_CFG_HASH hash code entry, attribute
# scNotifiedImage of the scWorkstation object is updated (unless the hash is same in the last entry,
# i.e. no change in the configuration has occured since the last boot notification).
# Number of previously used image files stored is given by scPosleasesMaxNotify
# value in the @em branchserver.conf file, all excessive entries are deleted.
# Legacy entries (without hash code) are ignored.
# @return 1 if has leases in hash, 0 if unable to get lease data
#TODO: rename to something sensible
sub create_workstations_from_image_reports
{
    my $self = shift;

    ## stepping through all found mac-adresses
    my %leases_hash = %{ $self->{'BOOTVERSION'} };    #only here used
    foreach my $mac (keys %leases_hash) {
        ## Now we are going to lookup this workstation in the directory
        my $wsentry = $self->{'CACHE'}{'WS'}{'BY-MAC'}{$mac};
        if (defined $wsentry) {
            my $mac_dn  = $wsentry->dn();
            my $ipaddr  = $self->{'BOOTVERSION'}{$mac}{'DATA'}{'IPADDR'};
            my $biosver = $self->{'BOOTVERSION'}{$mac}{'DATA'}{'HWBIOS'};
            #-> always modifies WS's scPosRegisterBiosVersion and therefore also obj's timestamp
            $biosver = $wsentry->get_value('scPosRegisterBiosVersion') if !defined $biosver;
            my $bootvtime =
              $self->{'BOOTVERSION'}{$mac}{'TIME'};    # set timestamp of last boot - needed for feature 313798
            my $hash = $self->{'BOOTVERSION'}{$mac}{'DATA'}{'POS_CFG_HASH'};
            $hash = '' if !defined $hash;
            if ($hash ne '' || !defined $ipaddr)
            { # only write image update information if HASH present (means we have new image and not just IP update or other)
                    #             `.-> hack to write image version also for old (just first two lines) bootversion files
                my $imageString = $self->{'BOOTVERSION'}{$mac}{'PATHNAME'};
                $imageString .= ";" . $self->{'BOOTVERSION'}{$mac}{'TIME'};    #add mtime
                $imageString .= ";" . $hash;    #add hash of the config file used to create reported image

                $self->update_rollback_time($self->{'ROLLBACKCFGS'}, $bootvtime, $hash) if $hash;

                #read scNotified attribute
                #my @imlist=$imentry->get_value('scNotifiedImage');
                my @imlistsorted = sort {
                    my @tmp = split(';', $a);
                    my $la = $tmp[1];    #by mtime, newst first
                    @tmp = split(';', $b);
                    my $lb = $tmp[1];
                    $lb <=> $la;
                } $wsentry->get_value('scNotifiedImage');
                #check if current bootversion is same as last entry. If so, ignore it (only for non empty hash).
                my $lasthash;
                if (@imlistsorted) {    #care only if non empty
                    my @tmp = split(';', $imlistsorted[0]);
                    $lasthash = $tmp[2];
                }
                $lasthash = '' if !defined $lasthash;
                if ($hash eq '' || $hash ne $lasthash) {    #first part needed for legacy funct.
                    my $imnumber    = @imlistsorted;
                    my $maximnumber = $self->{'CACHE'}{'POSLEASES_MAXNOTIFY'};
                    $maximnumber = 1 unless defined $maximnumber;    #if not valid integer, things may happen
                                                                     #trim older messages
                    while (@imlistsorted >= $maximnumber && @imlistsorted > 0)
                    {    #be sure not to cycle in any case after all membres out
                        my $delentry = pop @imlistsorted;
#                        #delete old rollback/config.MAC.hash files - dereferenced to cleanup_rollback_files
#                        my @cmds = $self->remove_cfgs_older_than($delentry);
#                        $self->{'STATIC'}->runcmds(\@cmds);
                        log_debug_printf('deleting scNotifiedImage %s from %s', $delentry, $mac_dn);
                        my $ret =
                          $self->{'CACHE'}{'LDAPCACHE'}->modify($mac_dn, delete => { scNotifiedImage => $delentry });
                        if ($ret->code) {
                            log_error_printf('LDAP-Error: Can\'t delete scNotifiedImage %s of %s: Error=%s ',
                                $delentry, $mac_dn, $ret->error);
                        }
                    }
                    log_debug_printf('adding %s into scNotifiedImage of %s', $imageString, $mac_dn);
                    my $ret =
                      $self->{'CACHE'}{'LDAPCACHE'}->modify($mac_dn, add => { scNotifiedImage => $imageString });
                    #$ret = $self->{'LDAP'}->modify($mac_dn, replace => { scNotifiedImage => $imageString });
                    if ($ret->code) {
                        log_error_printf('LDAP-Error: Can\'t add %s to %s: Error=%s ',
                            $imageString, $mac_dn, $ret->error);
                    }
                }
            }
            #update ip address always (POS11 SP1 funct.)
            if (defined $ipaddr && $self->{'STATIC'}{'EXTDHCP'}) {
                log_debug_printf('replacing IP %s into %s', $ipaddr, $mac_dn);
                my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($mac_dn, replace => [ipHostNumber => $ipaddr]);
                if ($ret->code) {
                    log_error_printf('LDAP-Error: Can\'t modify ip %s: Error=%s ', $ipaddr, $ret->error);
                }
            }
            #update bios version
            if (defined $biosver) {
                log_debug_printf('updating HWBIOS %s into %s', $biosver, $mac_dn);
                my $ret =
                  $self->{'CACHE'}{'LDAPCACHE'}->modify($mac_dn, replace => [scPosRegisterBiosVersion => $biosver]);
                if ($ret->code) {
                    log_error_printf('LDAP-Error: Can\'t modify biosversion %s: Error=%s ', $biosver, $ret->error);
                }
            }
            #update last boot time
            log_debug_printf('setting boot time to %s', $bootvtime);
            my $ret = $self->{'CACHE'}{'LDAPCACHE'}->modify($mac_dn, replace => [scLastBootTime => $bootvtime]);
            if ($ret->code) {
                log_error_printf('LDAP-Error: Can\'t modify scLastBootTime %s: Error=%s ', $bootvtime, $ret->error);
            }

        }
        else {
            ## workstation not in LDAP
            log_warn_printf('Nonexistant scWorkstation for MAC: %s', $mac);
        }
    } ## end foreach loop

    return 1;
} ## end sub create_workstations_from_image_reports

#TODO: use data that will be gathered in refresh_rollback_list
sub update_rollback_time
{
    my ($self, $rollcfgs_ref, $bootvtime, $hash) = @_;
    #must update mtime of related rollback config file , otherwise may get deleted even if still in use (bnc690858)
    my $bootstrtime = localtime($bootvtime);
    #find the rollback of this workstation, do for all MACs, one must be present as filename
    if (defined $rollcfgs_ref->{$hash}) {
        my $rollcfg_ref = $rollcfgs_ref->{$hash};
        my $rollstrtime = localtime($rollcfg_ref->{'MTIME'});
        log_info_printf(
            'updating time of \'%s\' from \'%s\' to \'%s\'...',
            $rollcfg_ref->{'PATHNAME'},
            $rollstrtime, $bootstrtime
        );
        utime $bootvtime, $bootvtime, $rollcfg_ref->{'PATHNAME'};
        $rollcfg_ref->{'MTIME'} = $bootvtime;
    }
} ## end sub update_rollback_time


"1;"
