#!/usr/bin/perl -w
use strict;
use warnings;

use POS::Config;
use POS::Syslog;
use POS::Pid;
use POS::Admin::ldapConfig;
use POS::Admin::LdapStruct;    #i.e. for global variables

log_parse_argv();

$SIG{'INT'} = 'break';
my $pid = new POS::Pid();
if ($pid->{'ERROR'}) {
    log_alert_printf('already running: %s', $pid->{'ERROR_MSG'});
    exit(1);
}

our $str_remove         = 'remove';
our $str_modify         = 'modify';
our $str_add            = 'add';
our $str_query          = 'query';
our $str_help           = 'help';
our $str_generate       = 'generate';
our $str_check          = 'validate';
our $str_checkXML       = 'validateXML';
our $str_checkLDIF      = 'validateLDIF';
our $str_import         = 'import';
our $str_export         = 'export';
our $str_refresh        = 'refresh';
our $sub_str_objectlist = 'optionlist';

my %LDAP_GENERAL_OPTIONS = (value => ['user', 'password'], description => ['username', 'password'],);

# WARNING!
# Don't forget to modify ldapConfig/POS::Admin when changing parameter names
my %LDAP_ACTIONS = (
    $str_add => {
        value       => { must => ['base'], may => ['nosm'] },
        description => {
            must => ['the base in which to search for objects'],
            may  => ['disable related action on SUSE Manager server']
        }
    },
    $str_remove => {
        value       => { must => ['DN'], may => ['recursive', 'nosm'], },
        description => {
            must => ['distinguished name of the element to remove'],
            may => ['option to delete an object with all sub-objects', 'disable related action on SUSE Manager server'],
        }
    },
    $str_query => {
        value       => { must => [], may => ['base', 'DN', 'full', 'list', 'showDN'] },
        description => {
            must => [],
            may  => [
                'the base in which to search for objects',
                'regexp of distinguished name of the element(s) to query',
                'switch to show attributes',
                'switch to show results without tree formatting',
                'displays all DN of found objects also in the tree view'
            ],
        }
    },
    $str_modify => {
        value       => { must => ['DN'], may => ['multival', 'newBase', 'nosm'], },
        description => {
            must => ['distinguished name of the element to modify'],
            may  => [
                'option to modify entries of a multivalued object',
                'new base where to move specified entry',
                'disable related action on SUSE Manager server'
            ],
        }
    },
    $str_help => {    #str_help needed for all actions list
        value => {
            must => [],
            may  => [
                $str_add,            $str_remove,                $str_query,    $str_modify,
                $sub_str_objectlist, $str_generate . $str_check, $str_checkXML, $str_checkLDIF,
                $str_export,         $str_import
            ],
        },
        description => {
            must => [],
            may  => [
                "help on $str_add",
                "help on $str_remove",
                "help on $str_query",
                "help on $str_modify",
                "help on $sub_str_objectlist",
                "help on $str_generate",
                "help on $str_check",
                "help on $str_checkXML",
                "help on $str_checkLDIF",
                "help on $str_export",
                "help on $str_import"
            ],
        }
    },
    $str_generate => {
        value => { must => ['base'], may => [], },
        description =>
          { must => ["the branch base for which Offline Installation File is about to be generated"], may => [], }
    },
    $str_check => {
        value       => { must => [], may => ['base'] },
        description => { must => [], may => ['base specifies DN of subtree to check'] }
    },
    $str_checkXML => {
        value       => { must => ['file'],                         may => ['schema'] },
        description => { must => ['path to XML file to validate'], may => ['path to XML validation RelaxNG schema'] }
    },
    $str_checkLDIF => {
        value       => { must => ['file'],                          may => [] },
        description => { must => ['path to LDIF file to validate'], may => [] }
    },
    $str_import => {
        value       => { must => [], may => ['file',               'type'] },
        description => { must => [], may => ['filename to import', 'type of importing file [XML|LDIF]'] }
    },
    $str_export => {
        value       => { must => [], may => ['file', 'type', 'base'] },
        description => {
            must => [],
            may => ['filename to export', 'type of exporting file [XML|LDIF]', 'base specifies DN of subtree to export']
        }
    },
    $str_refresh => { value => { must => [], may => [] }, description => { must => [], may => [] } }
);

my @LDAP_FILE_ATTRIBUTES = ('scConfigFileData');
my @LDAP_LOCAL_FILE      = ('scConfigFileLocalPath');

#generate @LDAP_ADD_ACTIONS
my @LDAP_ADD_ACTIONS = sort(keys(%LDAP_STRUCT));

#add multivalue spec into %LDAP_STRUCT (for use by posAdmin.pl)
# from data in %LDAP_ATTR_STRUCT
foreach my $option (sort(keys(%LDAP_STRUCT))) {
    my $must_ref = $LDAP_STRUCT{$option}{'attributes'}{'must'};
    my $may_ref  = $LDAP_STRUCT{$option}{'attributes'}{'may'};
    my @all;
    @all = (@all, @$must_ref) if defined $must_ref;
    @all = (@all, @$may_ref)  if defined $may_ref;
    my @multivals;
    foreach my $attr (@all) {
        if (defined $LDAP_ATTR_STRUCT{$attr}) {
            if ($LDAP_ATTR_STRUCT{$attr}{'SYNTAX'} !~ /SINGLE-VALUE\s*$/) {
                @multivals = (@multivals, $attr);
            }
        }
    }
    $LDAP_STRUCT{$option}{'attributes'}{'multival'} = \@multivals if @multivals;
}

# Admin server always use "/etc/SLEPOS/adminserver.conf" config file - bnc#511347
# check first if adminserver.conf is present, if not, use branchserver.conf - bnc#595732
my $config;
# let POS::Config to determine correct paths
eval { $config = POS::Config->new('admin') };
if ($@) {
    # admin configuration not found, try branch
    eval { $config = POS::Config->new('branchadmin') };
    if ($@) {
        # no config file found!
        log_crit_print("No suitable configuration file found!");
        exit(1);
    }
    log_notice_print("Using branchserver configuration file");
}

my %options = ();

my @all_opts = ();
initGlob();

log_info_print "Parsing options:";
my $base_action = getBaseAction(\%options, \@ARGV);    #have non-help baseaction here, otherwise we have already ended


filter_params(\%options);    #hack to supplement dummy dhcp options in scLocation when external dhcp server is specified
                             # in general, modify parameters for specific cases here

check_options($base_action, \%options);
#if we are here, no error so far
log_info_print " action = '$base_action':";
foreach my $op (keys(%options)) {
    if ($op ne $base_action) {
        my $option = $options{$op};
        $option = "" if (!defined $option);
        if ($op eq "password") {
            $option = "********";
        }
        log_info_printf("    %-26s   %-16s", $op . ":", $option);
    }
}

my $session = new POS::Admin::ldapConfig(OPTIONS => \%options, CONFIG => $config, BASEACTION => $base_action,);
if ($session->{'ERROR_CODE'}) {    #this error class should not happen at this early stage, just for future use
    log_error_print "Error in session initialization: (", $session->{'ERROR_CODE'}, "): ", $session->{'ERROR_MSG'};
    log_error_print "Operation failed.";
    exit 1;
}
if ($session->{'LDAP_ERROR_CODE'}) {
    log_error_print "Error in session initialization: (", $session->{'LDAP_ERROR_CODE'}, "): ",
      $session->{'LDAP_ERROR_MSG'};
    log_error_print "Operation failed.";
    exit 1;
}

my $ret;
my $opt_multival = 0;
if (exists($session->{'OPTIONS'}{'multival'})) {
    log_debug_print "multival option";
    delete($session->{'OPTIONS'}{'multival'});
    $opt_multival = 1;
}
log_debug_printf('Base Action: %s', $base_action);
if ($base_action eq $str_add) {
    my $found_ldap_action = 0;
    for my $i (@LDAP_ADD_ACTIONS) {
        if (exists($session->{'OPTIONS'}{$i})) {
            log_debug_print "Found LDAP Action: " . $i;
            $found_ldap_action          = 1;
            $session->{'SPECIFICATION'} = $i;
            $ret                        = $session->add($i, $session->{'OPTIONS'});
            last;
        }
    }
    if (!$found_ldap_action) {
        #should not happen unless line parsing and checking in posAdmin.pl disabled
        log_debug_print "no LDAP Action found";
        $session->{'ERROR_CODE'} = 1;
        $session->{'ERROR_MSG'}  = "no LDAP Action found\n";
    }
}
elsif ($base_action eq $str_query) {
    for my $i (@LDAP_ADD_ACTIONS) {
        if (exists($session->{'OPTIONS'}{$i})) {
            $session->{'SPECIFICATION'} = $i;
            $ret = $session->query($i, $session->{'OPTIONS'});
            last;
        }
    }
    #assume all objects, if none specified
    $ret = $session->query(undef, $session->{'OPTIONS'}) unless defined $session->{'SPECIFICATION'};
}
elsif ($base_action eq $str_generate) {
    $ret = $session->queryGenerate($session->{'OPTIONS'});
}
elsif ($base_action eq $str_modify) {
    for my $i (@LDAP_ADD_ACTIONS) {
        if (exists($session->{'OPTIONS'}{$i})) {
            $session->{'SPECIFICATION'} = $i;
            $ret = $session->modify($i, $session->{'OPTIONS'}) if (!$opt_multival);
            $ret = $session->modify_multival($i, $session->{'OPTIONS'}) if ($opt_multival);
            last;
        }
    }
}
elsif ($base_action eq $str_remove) {
    $ret = $session->remove($session->{'OPTIONS'});    # if (! $opt_multival);
}
elsif ($base_action eq $str_check) {
    # check $base option, if present search LDAP as base and pass result to validator, $base as SCOPE
    $ret = $session->validateLDAP();
}
elsif ($base_action eq $str_checkXML) {
    my $xml    = $session->{'OPTIONS'}{'file'};
    my $schema = $session->{'OPTIONS'}{'schema'};
    $ret = $session->validateXML($xml, $schema);
}
elsif ($base_action eq $str_checkLDIF) {
    $session->{'DRYRUN'} = 1;
    $ret = $session->importLDIF();
}
elsif ($base_action eq $str_import) {
    # determine type of import and pass to respective handlers
    my $type = $session->{'OPTIONS'}{'type'};
    my $file = $session->{'OPTIONS'}{'file'};
    if (not $type) {
        if (not $file) {
            log_error_print('Either --type or --file must be used with --import action');
        }
        ($type) = ($file =~ /\.(xml|ldif)$/);
    }
    # check once again if we managed to get type from extension
    if (not $type) {
        log_error_print('Unable to guess import type. Use --type option.');
    }
    elsif ($type =~ /^ldif$/i) {
        $ret = $session->importLDIF();
    }
    elsif ($type =~ /^xml$/i) {
        $ret = $session->importXML();
    }
    else {
        log_error_printf('Unknown import type %s', $type);
    }
}
elsif ($base_action eq $str_export) {
    my $type = $session->{'OPTIONS'}{'type'};
    my $file = $session->{'OPTIONS'}{'file'};
    if (not $type) {
        if (not $file) {
            log_error_print('Either --type or --file must be used with --export action');
        }
        ($type) = ($file =~ /\.(xml|ldif)$/);
        # check once again if we managed to get type from extension
        if (not $type) {
            log_error_print('Unable to guess export type. Use --type option.');
        }
    }
    # determine type of export and pass to respective handlers
    if ($type =~ /^ldif$/i) {
        $ret = $session->exportLDIF();
    }
    elsif ($type =~ /^xml$/i) {
        $ret = $session->exportXML();
    }
    else {
        log_error_printf('Unknown export type %s', $type);
    }
}
elsif ($base_action eq $str_refresh) {
    $ret = $session->refreshLdap();
}
## parse action result
$ret = 1 if (not defined $ret);
if ($session->{'LDAP_ERROR_CODE'}) {
    ## LDAP direct error with description
    log_error_print "Error in LDAP access (", $session->{'LDAP_ERROR_CODE'}, "): ", $session->{'LDAP_ERROR_MSG'};
    log_error_print "Operation failed.";
    exit 1;
}
elsif ($session->{'ERROR_CODE'}) {
    ## posAdmin direct error with description
    #print STDERR "Error ",$session->{'ERROR_CODE'}," in session: ",$session->{'ERROR_MSG'},"\n";
    log_error_print "Error in session (", $session->{'ERROR_CODE'}, "): ", $session->{'ERROR_MSG'};
    log_error_print "Operation failed.";
    exit 1;
}
elsif ($ret == 1) {
    log_error_print "Operation failed.";
}
elsif ($ret != 0) {
    ## other error
    ## check for LDAP error
    log_error_print "General posAdmin error: ", $ret;
    my $str_specific = "";
    if (defined $session->{'SPECIFICACTION'}) {
        $str_specific = " --" . $session->{'SPECIFICACTION'};
    }
    log_warn_print "For more help, run '$0 --help --$base_action$str_specific'";
    log_error_print "Operation failed.";
}
else {    #passed all LDAP checks
    log_notice_print "Operation successful.";
}

exit $ret;

### END

sub getBaseAction
{
    my ($op_hash, $options) = @_;
    my @base_actions = keys %LDAP_ACTIONS;
    my $base_action  = undef;
    my $base_option  = undef;
    my $last_action  = undef;
    my $help         = 0;
    my $optionshelp  = 0;
    foreach my $op (@{$options}) {
        #print "Option: ",$op,"\n";
        if ($op =~ /^--([a-zA-Z0-9]+)$/) {
            my $opname = $1;
            if ($opname eq $str_help) {
                $help = 1;
                next;
            }
            if ($opname eq $sub_str_objectlist) {
                $optionshelp = 1;
                next;
            }
            my $check = undef;
            foreach my $op1 (@base_actions) {
                # check both in lower case so baseaction parsing is case insensitive
                if (lc $opname eq lc $op1) {
                    if ($base_action) {
                        #more than one baseaction error
                        log_error_print
                          " cannot have '--$base_action', '--$opname' base options in the same time, exiting...";
                        display_help($str_help);
                        exit 1;
                    }
                    $check       = 1;
                    $base_action = $op1;
                    last;
                }
            }
            if ($check) {
                next;
            }
            #$op is not a base action
            #print "Lastaction: ", $last_action,"\n";
            #print "baseaction: ", $base_action,"\n";
            $last_action = $opname;
            $op_hash->{$last_action} = undef;

        }
        else {
            if ($last_action) {
                $op_hash->{$last_action} = $op;
                $last_action = undef;
            }
            elsif ($base_action eq $str_remove) {
                # special remove rule for missing DN bnc#751589
                $op_hash->{'dn'} = $op;
            }
            else {    #two values, i.e. missing --key
                (my $ok, my $opguess) = syntaxcheck($op);
                #my $opguess=findclose($op);
                if (defined $opguess) {
                    log_error_print " unrecognized value/option: '$op' (maybe you meant  $opguess ?), exiting...";
                }
                else {
                    log_error_print " unrecognized value/option: '$op', exiting...";
                }
                display_help($str_help);
                exit 1;
            }
        }
    }    #of main option parsing cycle


    #help handling here
    if ($help) {
        log_info_printf("base action: '--$str_help'");
        if (defined $base_action) {
            # printf( " --$base_action'\n");
            #my @keyarray=keys(%{$op_hash});
            display_help($base_action, keys(%{$op_hash}));
        }
        elsif ($optionshelp) {
            display_help($str_help, $sub_str_objectlist);
        }
        else {
            display_help($str_help);
        }
        exit 0;
    }

    if (!defined $base_action) {
        #try to guess it
        my @guess = ();
        foreach my $opt (keys(%{$op_hash})) {
            (my $ok, my $opguess) = syntaxcheck($opt, \@base_actions);    # $ok==1 should not happen here
            if (defined $opguess) {
                @guess = (@guess, $opguess);
            }
        }
        if (scalar(@guess)) {
            my $guess_str = join(', or ', @guess);
            log_error_printf(" error, no base action found, maybe you meant  $guess_str ?");
        }
        else {
            log_warn_printf(" no base action found, exiting...");
            display_help($str_help);
        }
        exit 1;
    }


    log_info_printf(" base action: --$base_action");
    return $base_action;
} ## end sub getBaseAction

sub display_help
{
    my ($action, @specific_items) = @_;

    if (!defined $action) {
        print STDERR " unknown error.. :\n";    #should not happen
        $action = $str_help;
    }
    if ($action eq $str_help) {
        if (!@specific_items) {                 #general help
            print STDERR "\n\n";
            print STDERR "General help :\n\n";
            print STDERR " usage: 1. $0 <options> <base action> <option1> value1 ... <optionN> valueN>\n";
            #print STDERR "        2. $0 <base action> --help\n";
            print STDERR "\n";
            print STDERR "base action:\n\n";
            print STDERR "  --$str_query:                    query the database\n";
            print STDERR "  --$str_add:                      add an item to the database\n";
            print STDERR "  --$str_remove:                   delete an item from the database\n";
            print STDERR "  --$str_modify:                   modify an item in the database\n";
            print STDERR "  --$str_generate:                 generate Offline Installation File for branch\n";
            print STDERR "                              server deployement\n";
            print STDERR
              "  --$str_refresh                   update internal SLEPOS data after editing LDAP with external tools\n";
            print STDERR "  --$str_check                  validates SLEPOS LDAP data\n";
            print STDERR "  --$str_checkXML               validates SLEPOS XML against SLEPOS LDAP data\n";
            print STDERR "  --$str_checkLDIF              validates LDIF against SLEPOS LDAP data\n";
            print STDERR "  --$str_import                    imports SLEPOS XML or LDIF into LDAP\n";
            print STDERR "  --$str_export                    exports SLEPOS XML of LDIF from LDAP\n";
            print STDERR "  --$str_help:                     display this help\n";
            print STDERR "\n";
            print STDERR "options:\n\n";
            log_print_usage();
            print STDERR "  --user:                     username\n";
            print STDERR "  --password:                 password\n";
            print STDERR "  --base, --DN, and other base action specific options\n";
            print STDERR "\n\n";
            print STDERR "call '$0 --$str_help' <base action>  to get an extended help for a specific action\n";
            # print STDERR "  to get an extended help for a specific action\n";
            print STDERR "\n";
            print STDERR "\n";
        }
        else {
            #help on ldap structure
            #elsif (
            #if ($action eq $str_add || $action eq $str_modify || $action eq $str_query) {
            print STDERR "<option list> = <option1> value1 ... <optionN> valueN\n";
            print STDERR "\n";
            print STDERR "option list:\n\n";

            my $struct = \%LDAP_STRUCT;
            if (@specific_items == 1 && $specific_items[0] eq $sub_str_objectlist) {    #show all items info
                @specific_items = sort(keys(%{$struct}));
            }
            my @context = keys(%{$struct});
            foreach my $specific_item (@specific_items) {                               #show items info
                (my $ok, my $guess) = syntaxcheck($specific_item, \@context);
                if (!$ok) {
                    if ($guess) {
                        print STDERR "  --", $specific_item, ":\t unknown option (maybe you meant  ", $guess, " ?)\n";
                    }
                    else {
                        print STDERR "  --", $specific_item, ":\t unknown option\n";
                    }
                }
                else {
                    print STDERR "  --", $specific_item, ":\t";
                    print STDERR "$struct->{$specific_item}{'DESC'}\n";
                    print STDERR "          must attributes:\n";
                    for my $att (@{ $struct->{$specific_item}{'attributes'}{'must'} }) {
                        my $helpstr = $struct->{$specific_item}{$str_query}{$att};
                        printf(STDERR "\t\t%-25s   %-16s\n", $att . ":", $helpstr) if defined $helpstr;
                    }
                    print STDERR "          may attributes:\n";
                    for my $att (@{ $struct->{$specific_item}{'attributes'}{'may'} }) {
                        my $helpstr = $struct->{$specific_item}{$str_query}{$att};
                        printf(STDERR "\t\t%-25s   %-16s\n", $att . ":", $helpstr) if defined $helpstr;
                    }
                }
                printf STDERR "\n";
            }
        }

    }
    else {    #specific help. for a given action
        my $specstr = join(', --', @specific_items);
        if ($specstr) { $specstr = " --" . $specstr; }
        print STDERR "\n\n";
        print STDERR "Specific help:\n (", $action, $specstr, ") \n\n";
        my $str_optlist = "<option list>";
        if ($action eq $str_remove) { $str_optlist = "" }
        print STDERR "usage: $0 <options> --", $action, " ", $str_optlist, "\n";
        print STDERR "\n";
        print STDERR "options:\n";
        my $options = "";    #general

        for (my $i = 0; $i < @{ $LDAP_GENERAL_OPTIONS{'value'} }; $i++) {
            $options = $options .
              sprintf("  --%-30s%s\n",
                @{ $LDAP_GENERAL_OPTIONS{'value'} }[$i] . ":",
                @{ $LDAP_GENERAL_OPTIONS{'description'} }[$i]);
        }
        if ($options) {
            print STDERR "\ngeneral options:\n";
            print STDERR $options;
        }
        $options = "";       #must
        for (my $i = 0; $i < @{ $LDAP_ACTIONS{$action}{'value'}{'must'} }; $i++) {
            $options = $options .
              sprintf("  --%-30s%s\n",
                @{ $LDAP_ACTIONS{$action}{'value'}{'must'} }[$i] . ":",
                @{ $LDAP_ACTIONS{$action}{'description'}{'must'} }[$i]);
        }
        if ($options) {
            print STDERR "\nmust options:\n";
            print STDERR $options;
        }
        $options = "";       #may
        for (my $i = 0; $i < @{ $LDAP_ACTIONS{$action}{'value'}{'may'} }; $i++) {
            $options = $options .
              sprintf("  --%-30s%s\n",
                @{ $LDAP_ACTIONS{$action}{'value'}{'may'} }[$i] . ":",
                @{ $LDAP_ACTIONS{$action}{'description'}{'may'} }[$i]);
        }
        if ($options) {
            print STDERR "\nmay options:\n";
            print STDERR $options;
        }
        if ($str_optlist) {
            print STDERR "\n";
            print STDERR
" call '$0 --$str_help --$sub_str_objectlist' to get $str_optlist help for all LDAP objects and their values\n";
        }
        print STDERR "\n";
        print STDERR "\n";
    }
    print STDERR "For more help, see the SLEPOS manual.\n";


} ## end sub display_help

sub guesscheckmsg
{
    # tests if the checked option '$opt' is included in the specified set '$context_ref'
    # '$paramtype' is just description to fit in the resulting message
    my ($opt, $context_ref, $paramtype) = @_;
    my $res = '';
    (my $ok, my $guess_str, my $guess_ref) = syntaxcheck($opt, $context_ref);
    if (!$ok && defined $guess_ref) {
        my $is, my $s;
        if   (scalar @$guess_ref > 1) { $is = "are";  $s = "s"; }
        else                          { $is = "is a"; $s = ""; }
        $res = "maybe you meant $guess_str, which $is $paramtype$s ?";
    }
    else {
        # we have a match, return it to solve case sensitivity
        $res = $$guess_ref[0];
    }
    return ($ok, $res);
} ## end sub guesscheckmsg


sub check_options
{
    my ($baseaction, $opt_ref) = @_;
    my %ophash = %{$opt_ref};
    if ($baseaction) {
        delete($ophash{$baseaction});

        my $generalOptions = $LDAP_GENERAL_OPTIONS{'value'};
        my $mustOptions    = $LDAP_ACTIONS{$baseaction}{'value'}{'must'};
        my $mayOptions     = $LDAP_ACTIONS{$baseaction}{'value'}{'may'};
        my $act_item;
        my $act_item_str = '';
        my $error        = 0;
        if ($baseaction =~ /$str_add|$str_modify|$str_query/) {
            # we need special handling for add&modify ops, query may have act_item
            # only one adding, modifying objectClass at one time is allowed
            foreach my $item (@LDAP_ADD_ACTIONS) {
                if (exists($opt_ref->{$item})) {
                    if (defined $act_item) {
                        log_error_print("error, more than one item to $baseaction found ($act_item, $item..)");
                        log_error_print("For more help, run '$0 --help --$baseaction'");
                        exit 1;
                    }
                    $act_item = $item;
                }
            }
            if ($act_item) {
                log_info_print("item : '$act_item'");
                delete($ophash{$act_item});
                $act_item_str = " --$act_item";
                # enhance must and may options with object specific options
                $mustOptions = [@$mustOptions, @{ $LDAP_STRUCT{$act_item}{'attributes'}{'must'} }];
                $mayOptions  = [@$mayOptions,  @{ $LDAP_STRUCT{$act_item}{'attributes'}{'may'} }];
            }
            ## not even main item found
            elsif ($baseaction ne $str_query) {    #no item to add/modify/remove..., try to guess it
                my @guess       = ();
                my @unknown_opt = ();
                foreach my $opt (keys(%ophash)) {
                    (my $ok, my $opguess) = syntaxcheck($opt, \@LDAP_ADD_ACTIONS);   # $ok==1 really should not happen..
                    if (defined $opguess) {
                        @guess       = (@guess,       $opguess);
                        @unknown_opt = (@unknown_opt, "'--$opt'");
                    }
                }
                if (scalar(@guess)) {
                    my $guess_str       = join(', or ', @guess);
                    my $unknown_opt_str = join(', or ', @unknown_opt);
                    log_error_print("error, no item to $baseaction found, maybe you meant  $guess_str");
                    log_error_print("                                  instead of $unknown_opt_str ?");
                }
                else {
                    log_error_print("no item to $baseaction found !");
                }
                log_error_print("For more help, run '$0 --help --$baseaction'");
                exit 1;
            }
        }

        foreach my $opt (keys(%ophash)) {
            my $ok;
            my $newOpt;
            my ($guessGeneral, $guessMust, $guessMay);
            ($ok, $guessGeneral) = guesscheckmsg($opt, $generalOptions, "base option");
            $newOpt = $guessGeneral if $ok;
            unless ($ok) {
                ($ok, my $guessMust) = guesscheckmsg($opt, $mustOptions, "must-attribute");
                $newOpt = $guessMust if $ok;
            }
            unless ($ok) {
                ($ok, my $guessMay) = guesscheckmsg($opt, $mayOptions, "may-attribute");
                $newOpt = $guessMay if $ok;
            }

            unless ($ok) {    #$opt is neither base option, must-attribute or must-attribute
                log_error_printf('error, unknown option or%s attribute : "%s"', $act_item_str, $opt);
                log_error_print($guessGeneral) if $guessGeneral;
                log_error_print($guessMust)    if $guessMust;
                log_error_print($guessMay)     if $guessMay;
                $error = 1;
            }
            elsif ($opt ne $newOpt) {
                # update original attribute name. This fixes case insensitivity
                $opt_ref->{$newOpt} = $opt_ref->{$opt};
                delete $opt_ref->{$opt};
            }
        }
        #test baseaction must parameters, rest are tested in POS::Admin after defaults are added
        foreach (@{ $LDAP_ACTIONS{$baseaction}{'value'}{'must'} }) {
            if (!defined $ophash{$_}) {
                log_error_print("error, missing must-attribute '$_'");
                $error = 1;
            }
        }

        if ($error) {
            log_info_print("For more help, run '$0 --help --$baseaction'");
            exit 1;
        }
    }

    else {
        log_error_print("unknown error, no base action specified !");    #should not happen...
        exit 1;
    }
} ## end sub check_options


sub initGlob
{
    # initializes option lists for syntax checking
    my @all_substruct_must = ();
    my @all_substruct_may  = ();
    my @all_main_opts      = (
        keys %LDAP_ACTIONS,
        @LDAP_ADD_ACTIONS, @LDAP_FILE_ATTRIBUTES, @LDAP_LOCAL_FILE, @{ $LDAP_GENERAL_OPTIONS{'value'} },
    );
    for my $action_options (keys(%LDAP_ACTIONS)) {    #unlike bellow, here may and must always defined
        @all_main_opts = (@all_main_opts, @{ $LDAP_ACTIONS{$action_options}{'value'}{'must'} });
        @all_main_opts = (@all_main_opts, @{ $LDAP_ACTIONS{$action_options}{'value'}{'may'} });
    }
    for my $option (keys(%LDAP_STRUCT)) {
        my $tmp = $LDAP_STRUCT{$option}{'attributes'}{'must'};
        if (defined $tmp) {
            @all_substruct_must = (@all_substruct_must, @{$tmp});
        }
        $tmp = $LDAP_STRUCT{$option}{'attributes'}{'may'};
        if (defined $tmp) {
            @all_substruct_may = (@all_substruct_may, @{$tmp});
        }
    }
    @all_opts = (@all_main_opts, @all_substruct_must, @all_substruct_may);
} ## end sub initGlob


sub syntaxcheck
{
    #find if '--[option]' exists in LDAP specification
    my ($unknown_op, $context) = @_;    #parameter option to test, #ref to array in which to testt
    my %op_guess      = ();
    my @op_guess      = ();
    my $op_guess      = undef;
    my $op_guess_aref = undef;
    my $check         = 0;

    if (!defined $unknown_op) { die "nothing to guess!!"; }
    #$unknown_op=~s/^-+//ig;
    $unknown_op =~ s/^\W+//ig;
    $unknown_op =~ s/\W+$//ig;

    if (!defined $context) {            #prepare option list
        $context = \@all_opts;
    }

    # my $do_bgtest=1;
    foreach my $option (@{$context}) {
        #exact, case insensitive match
        if (lc $unknown_op eq lc $option) {
            $check = 1;
            $op_guess{$option} = 1;
            last;
        }
    }

    #even more vague match (up to one changed letter, one superfluous or one missing (unless it is the last one),
    #                       in which case the next test (beginning same test) should be more appropriate           )
    if (!$check) {
        foreach my $option (@{$context}) {
            for (my $i = 0; $i < length($unknown_op); $i++) {
                my $xop = $unknown_op;
                $xop =~ s@^(.{$i}).@$1\\w?@i;    #make a new regular expression by overwritting '\w?'
                if ($option =~ /^$xop$/ig) {
                    $op_guess{$option} += 2;
                }
            }
            for (my $i = 0; $i < length($unknown_op) + 1; $i++) {
                my $xip = $unknown_op;
                $xip =~ s@^(.{$i})@$1\\w?@i;     #make a new regular expression by inserting '\w?'
                if ($option =~ /^$xip$/ig) {
                    $op_guess{$option} += 2;
                }
            }
        }

        #even more vague match 2 (beginning is the same)
        ##only if no match yet
        foreach my $option (@{$context}) {
            if ($option =~ /^$unknown_op/ig) {
                $op_guess{$option} += 2;
            }
        }
    }    #of guesses

    #prepare output
    @op_guess = sort(keys(%op_guess));
    if (scalar(@op_guess) > 0) {
        $op_guess_aref = \@op_guess;
        $op_guess = "--" . join(', --', @op_guess) . "";
    }

    return ($check, $op_guess, $op_guess_aref);    # 0 bad, 1 good ; formated output ; ref to guesses hash
} ## end sub syntaxcheck


sub filter_params
{
    my ($op_hash) = @_;
    if (exists($op_hash->{'scService'}) && $base_action eq 'add' && $op_hash->{'scServiceName'} eq "dhcp") {
        ## defaults for dhcp server
        if (not defined $op_hash->{'scDhcpDynLeaseTime'}) {
            $op_hash->{'scDhcpDynLeaseTime'} = 300;
        }
        if (not defined $op_hash->{'scDhcpFixedLeaseTime'}) {
            $op_hash->{'scDhcpFixedLeaseTime'} = 14400;
        }
    }
} ## end sub filter_params


END {
    $pid->delete() if defined $pid && $$ == $pid->get_pid() && $$ == $pid->get_pidf();
} ## end END

sub break
{
    #main reason is to invoke END block in the ctrl+C evend to clear the pid file
    log_warn_print("ctrl+C detected!");
    exit(1);
} ## end sub break
