#!/usr/bin/perl -w

=pod

=head1 NAME

posGetIP - find addresses for branch server clients

=head1 SYNOPSIS

posGetIP [--ip] [--mac]

=head1 OPTIONS

=over

=item --ip

print IP addresses. Enabled by default; disable with --noip

=item --mac

print MAC addresses

=back

=head1 DESCRIPTION

Finds all addresses for POS terminals that are managed by the local
branch server. This tool must be run on the branch server. Output is
the list of addresses, one line each.

Both IP and MAC addresses can be listed. Default is to list the IP
addresses.

It finds its server base by looking at the IP addresses that are
configured on the local machines.

=head1 EXAMPLES

 D201:~ # posGetIP
 192.168.99.19
 192.168.99.18
 192.168.99.17
 192.168.99.16
 192.168.99.15
 192.168.99.14
 192.168.99.13
 192.168.99.12
 192.168.99.11


 D201:~ # ./posGetIP --mac --noip
 00:09:6B:24:25:C7
 00:02:55:23:F3:93
 00:06:29:E3:02:E6
 00:04:AC:AC:B3:B0
 00:09:6B:27:FF:55
 00:09:6B:27:FF:4D
 00:09:6B:3B:01:07
 00:02:55:E8:FA:C9
 00:04:AC:D5:02:1E

 D201:~ # ./posGetIP --mac       
 192.168.99.19   00:09:6B:24:25:C7
 192.168.99.18   00:02:55:23:F3:93
 192.168.99.17   00:06:29:E3:02:E6
 192.168.99.16   00:04:AC:AC:B3:B0
 192.168.99.15   00:09:6B:27:FF:55
 192.168.99.14   00:09:6B:27:FF:4D
 192.168.99.13   00:09:6B:3B:01:07
 192.168.99.12   00:02:55:E8:FA:C9
 192.168.99.11   00:04:AC:D5:02:1E

=head1 CONFIGURATION

F<.../opt/SLES/POS/branchserver.conf> is used to find the LDAP
connection information.

=head1 AUTHOR

Till Franke <franke@suse.com>

(C) SuSE 2003

=cut


use POS::Config;
use POS::Syslog;
use POS::LdapUtil;
use strict;
use POS::LDAP;
use Getopt::Long;

log_parse_argv();

my $OPT_MAC   = 0;
my $OPT_IP    = 1;
my $arguments = {};
$arguments->{'mac'} = $OPT_MAC;
$arguments->{'ip'}  = $OPT_IP;
GetOptions($arguments, ("mac!", "ip!"));
$OPT_MAC = $arguments->{'mac'};
$OPT_IP  = $arguments->{'ip'};


my $config = new POS::Config("/etc/SLEPOS/branchserver.conf");
my $ldap   = POS::LDAP->new(
    'CONFIG'     => $config,
    'NEEDCAP'    => 'ro',
    'LDAPBINDDN' => $config->{'POS_ADMIN_DN'},
    'LDAPBINDPW' => $config->{'POS_ADMIN_PASSWORD'}
);

my @IPs;
my $IP;
for $IP (`/sbin/ip addr | grep "inet "`) {
    $IP =~ /\s+inet\s+([0-9.]+)\/.*$/;
    $IP = $1;
    push @IPs, $IP;
}

my $result = undef;
for $IP (@IPs) {
    $result = $ldap->search(
        base   => $config->{'ALL_LDAPBASE'},
        scope  => 'sub',
        filter => "(&(ipHostNumber=$IP)(objectClass=scNetworkCard))",
        attrs  => ['dn']
    );
    if ($result->count == 1) {
        last;
    }
}
if (!$result) {
    die("cannot find my IP address in LDAP - exiting");
}
my $entry = $result->pop_entry;
my @a = split(/,/, $entry->dn);
shift @a;
my $result_loc = undef;
while (@a) {
    #print join(",", @a),"\n";
    $result_loc =
      $ldap->search(base => join(",", @a), scope => 'base', filter => '(objectClass=scLocation)', attrs => ['dn']);
    shift @a;
    if ($result_loc->count == 1) {
        last;
    }
}
if (!$result_loc) {
    die("cannot find my location in LDAP - exiting");
}
$entry = $result_loc->pop_entry;
my $result_IPs = $ldap->search(
    base   => $entry->dn,
    scope  => 'sub',
    filter => '(objectClass=scWorkstation)',
    attrs  => ['ipHostNumber', 'macAddress']
);
while ($entry = $result_IPs->pop_entry) {
    $OPT_IP && print $entry->get_value('ipHostNumber');
    ($OPT_IP && $OPT_MAC) && print "\t";
    $OPT_MAC && print $entry->get_value('macAddress');
    print "\n";
}


